#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <math.h>
#include <sys/time.h>
#include "System.h"
#include "Utility.h"
#include "SNPData.h"
#include "Simulate.h"
#include "RelationshipUtils.h"
#include "CountHaplosUtils.h"

string getRelStr(int mrca, int genA, int genB){\
  stringstream stream;
  stream.clear();
  stream.str("");
  stream << mrca << "_";
  if(genA > -1) stream << genA << "_";
  else stream << "n1_";
  if(genB > -1) stream << genB;
  else stream << "n1";
  return stream.str();
}

/*
 * Read the posterior probabilities for a certain pair of 
 * individuals and assumed relationship, discretize them based
 * on the given cutoffs and compute the allele frequencies.
 * 
 * The posterior filenames must have the format:
 * <postPref>chr<chrom num><postSuf>_X_X_X.postibd
 * and they contain the columns 
 * prob of 1 IBD
 * prob of 2 IBD (optional)
 * prob that phaseA = 1
 * prob that phaseB = 1
 * prob that corrA = 1 (optional)
 * prob that corrB = 1 (optional)
 *
 * The bin filenames must have the format:
 * <binPref>[Cleared_]chrom<chrom num>.snpprofiledata.bin and
 * <binPref>Pruned_chrom<chrom num>.snpprofiledata.bin.
 * <restoreBinPref>[Cleared_]chrom<chrom num>.snpprofiledata.bin and
 *
 * The pair filenames must have the format:
 * <pairPref>chrom*
 *
 * The output filenames will have the format:
 * <outPref>w<snpsBefore>_i<ibdCutoff>_n<nonIBDcutoff>_c<cutoff>_X_X_X.postcounts
 *
 * Cutoffs:
 * - If prob of 1 IBD > ibdCutoff, IBD = 1
 * - If prob of 2 IBD > ibdCutoff, IBD = 2 
 * (hopefully only one of the above will hold if ibdCutoff > 0.5.
 * - If prob of 1 IBD < nonIBDcutoff AND prob of 2 IBD < nonIBDcutoff, IBD = 0.
 * - Otherwise, IBD = -1 (unknown).
 * - If prob of phaseA > cutoff, phaseA = 1. Otherwise phaseA = 0;
 * similarly for phaseB, corrA, corrB. 
 * Cutoffs are NOT in log-space.
 */
int Main(vector<string> args){
  int numArgs = 13;
  AssertMsg(args.size() == numArgs,
	    "Usage: CountHaplosPostMain restoreBinPref binPref pairPref postPref postSuf configPath outPref snpsBefore ibdCutoff nonIBDcutoff cutoff cleared");
   
  bool cleared;
  int snpsBefore, snpsAfter;
  double ibdCutoff, nonIBDcutoff, cutoff, expIBDcutoff, expNonIBDcutoff, expCutoff;
  string restoreBinPref = args[1], binPref = args[2], pairPref = args[3], 
    postPref = args[4], postSuf = args[5], configPath = args[6], 
    outPref = args[7];

  stringstream stream(args[8]);
  stream >> snpsBefore;
  snpsAfter = snpsBefore;

  stream.clear();
  stream.str(args[9]);
  stream >> expIBDcutoff;
  ibdCutoff = log(expIBDcutoff);

  stream.clear();
  stream.str(args[10]);
  stream >> expNonIBDcutoff;
  nonIBDcutoff = log(expNonIBDcutoff);

  stream.clear();
  stream.str(args[11]);
  stream >> expCutoff;
  cutoff = log(expCutoff);

  stream.clear();
  stream.str(args[12]);
  stream >> cleared;

  cerr << "Using cutoffs " << expIBDcutoff << " (" << ibdCutoff << "), " 
       << expNonIBDcutoff << " (" << nonIBDcutoff << "), "
       << expCutoff << " (" << cutoff << ")" << endl;

  vector< vector<double> > errorFunction;
  errorFunction.resize(4);
  for(int i = 0; i < 4; ++i){
    errorFunction[i].resize(4, 0.003);
    errorFunction[i][i] = 0.991;
  }
  double pseudocount = 0.01;

  timeval tv;
  double start, end;

  //Read the relationships whose posteriors we'll consider.
  ifstream configFile(configPath.c_str());
  vector<int> mrcas, gensA, gensB;
  readConfig(configFile, mrcas, gensA, gensB);
  configFile.close();
  
  //Write the headers of all the files.
  for(int r = 0; r < mrcas.size(); ++r){
    if(mrcas[r] <= 0){
      cerr << "Unrelated not supported.";
      exit(1);
    }
    string relStr = getRelStr(mrcas[r], gensA[r], gensB[r]);

    stream.clear();
    stream.str("");
    stream << outPref << "w" << snpsBefore << "_i" << expIBDcutoff << "_n" 
	   << expNonIBDcutoff << "_c" << expCutoff 
	   << "_" << relStr << ".postcounts2";

    ofstream outfile(stream.str().c_str());
    

    outfile << "#prevIBD\tIBD\tidx\tchr\t"
	    << "startSNP\tendSNP\t"
	    << "mappedStartSNP\tmappedEndSNP\t"
	    << "blockGenomStart\tblockGenomEnd\t"
	    << "blockGenetStart\tblockGenetEnd\t"
	    << "freqA\tfreqB\t"
	    << "lastPos\tmappedLastPos\tgenomLastPos\tgenetLastPos\n";
    /*
    outfile << "#prevIBD\tIBD\tidx\tchr\t"
	    << "startSNP\tendSNP\t"
	    << "mappedStartSNP\tmappedEndSNP\t"
	    << "blockGenomStart\tblockGenomEnd\t"
	    << "blockGenetStart\tblockGenetEnd\tfreqIndepA\tfreqIndepB\t"
	    << "freqA\tfreqB\tfreqNonIbdA\tfreqNonIbdB\t"
	    << "lastPos\tmappedLastPos\tgenomLastPos\tgenetLastPos\n";
    */
    outfile.flush();
    outfile.close();
  }

  for(int chrom = 1; chrom <= 22; ++chrom){
    cerr << "Starting chromosome " << chrom << endl;

    //Read the full snpprofiledata.
    stream.clear(); stream.str("");
    stream << binPref;
    if(cleared == 1)
      stream << "Cleared_";
    stream << "chrom" << chrom << ".snpprofiledata.bin";
    SNPProfile profile, restrProfile;
    SNPData data, restrData;
    ReadBinSNPProfileData(stream.str(), profile, data);
    //cerr << "Read full bin " << stream.str() << ", size " << data.seq_len << endl;
    //Compute letter freq based on the full snpprofiledata.
    vector< vector<double> > letterFreq;
    computeLetterFreq(data, pseudocount, letterFreq, 4);
    
    stream.clear();
    stream.str("");
    stream << restoreBinPref;
    stream << "Pruned";
    if(cleared == 1)
      stream << "_Cleared";
    stream << "_chrom" << chrom << "_select";
    ifstream selectFile(stream.str().c_str());
    vector<int> select(data.seq_len, 0);
    ReadBin(selectFile, select);
    selectFile.close();
    //cerr << "Read select from " << stream.str() << ", size " << select.size() << endl;
    
    int restrictSize = 0;
    for(int i = 0; i < select.size(); ++i)
      restrictSize += select[i];
    vector<int> posMap(restrictSize, 0);
    int pos = 0;
    for(int i = 0; i < select.size(); ++i){
      if(select[i] == 1)
	posMap[pos++] = i;
    }    
    
    //Read the data (full) that was used to create the individuals.
    //Read the individuals and add them to "the other" data.
    SNPProfile restoreProfile;
    SNPData restoreData;
    stream.clear(); stream.str("");
    stream << restoreBinPref;
    if(cleared == 1)
      stream << "Cleared_"; 
    stream << "chrom" << chrom << ".snpprofiledata.bin";
    //As in LikelihoodTest3, this is not the best way to test for equality of paths.
    if(restoreBinPref != binPref)
      ReadBinSNPProfileData(stream.str(), restoreProfile, restoreData);      
    //cerr << "Read restore bin " << stream.str() << ", size " << data.seq_len << endl;
    
    stream.clear();
    stream.str("");
    stream << pairPref << "chrom" << chrom;
    if(cleared == 1)
      stream << "_cleared";
    ifstream pairFile(stream.str().c_str());
    Simulate::Diplo indivA, indivB;
    vector<int> relIndices(2, 0);
    if(restoreBinPref != binPref){
      relIndices[0] = restoreIndividual(restoreData, data, indivA, pairFile);
      relIndices[1] = restoreIndividual(restoreData, data, indivB, pairFile);
    }
    else{
      relIndices[0] = restoreIndividual(data, data, indivA, pairFile);
      relIndices[1] = restoreIndividual(data, data, indivB, pairFile);
    }
    pairFile.close();
    //cerr << "Restored individuals from " << stream.str() << endl;

    for(int r = 0; r < mrcas.size(); ++r){
      gettimeofday(&tv, NULL);
      start = tv.tv_sec + (tv.tv_usec/1000000.0);
      string relStr = getRelStr(mrcas[r], gensA[r], gensB[r]);
      stream.clear();
      stream.str("");
      stream << outPref << "w" << snpsBefore << "_i" << expIBDcutoff << "_n" 
	     << expNonIBDcutoff << "_c" << expCutoff 
	     << "_" << relStr << ".postcounts2";
      
      ofstream outfile(stream.str().c_str(), ios_base::app);    

      bool hasIBD2 = false;
      if(mrcas[r] == 2 && gensA[r] == 0 && gensB[r] == 0)
	hasIBD2 = true;
      bool hasCorr = true;
      //Uncomment these to handle the old format (that didn't always have corr)
      //if((mrcas[r] == 2 && gensA[r] > 0 && gensB[r] > 0) ||
      // (mrcas[r] == 1 && gensA[r] >= 0 && gensB[r] >= 0))
      // hasCorr = false;
      
      //Read the posteriors and discretize them.
      stream.clear();
      stream.str("");
      stream << postPref << "chr" << chrom << postSuf << "_" << relStr << ".postibd";
      //cerr << "Opening " << stream.str() << endl;
      ifstream postFile(stream.str().c_str());
      //stream << ".bin";
      //ifstream postBinFile(stream.str().c_str());
      string str;
      int num;
      vector< vector<int> > indicators(posMap.size(), vector<int>(5, 0));
      
      //There should be one line for each SNP in the restricted data.
      for(int i = 0; i < posMap.size(); ++i){
	postFile >> str;
	double postIBD1 = 0.0, postIBD2 = 0.0, postPhaseA = 0.0, 
	  postPhaseB = 0.0, postCorrA = 0.0, postCorrB = 0.0;
	if(postFile.eof()){
	  cerr << "Expected " << posMap.size() << " lines in posteriors file, found " 
	       << (i - 1) << endl;
	  exit(1);
	}
	stream.clear();
	stream.str(str);
	stream >> postIBD1;
	if(hasIBD2){
	  postFile >> str;
	  stream.clear();
	  stream.str(str);
	  stream >> postIBD2;
	}
	postFile >> str;
	stream.clear();
	stream.str(str);
	stream >> postPhaseA;
	
	postFile >> str;
	stream.clear();
	stream.str(str);
	stream >> postPhaseB;
	
	if(hasCorr){
	  postFile >> str;
	  stream.clear();
	  stream.str(str);
	  stream >> postCorrA;
	  double denom = postIBD1;
	  if(hasIBD2) 
	    denom = computeLogSumOfTwo(postIBD1, postIBD2);
	  postCorrA -= denom;
	  
	  postFile >> str;
	  stream.clear();
	  stream.str(str);
	  stream >> postCorrB;  
	  postCorrB -= denom;
	}
	else{
	  postCorrA = postPhaseA;
	  postCorrB = postPhaseB;
	}
	
	if(postIBD1 > ibdCutoff)
	  indicators[i][0] = 1;
	else if(hasIBD2 && postIBD2 > ibdCutoff)
	  indicators[i][0] = 2;
	else if(postIBD1 < nonIBDcutoff && (!hasIBD2 || (hasIBD2 && postIBD2 < nonIBDcutoff)))
	  indicators[i][0] = 0;
	else
	  indicators[i][0] = -1;
	
	indicators[i][1] = (int)(postPhaseA > cutoff);
	indicators[i][2] = (int)(postPhaseB > cutoff);
	indicators[i][3] = (int)(postCorrA > cutoff);
	indicators[i][4] = (int)(postCorrB > cutoff);
	//if(i < 10)
	//cerr << "Line " << i << ": " << postIBD1 << ", " << postPhaseA << ", " 
	//   << postPhaseB << ", " << postCorrA << ", " << postCorrB << endl;
      }
      postFile.close();
      /*else{
	//cerr << "In here" << endl;
	vector< vector<double> > contents(posMap.size(), vector<double>(6, 0.0));
	ReadBin(postBinFile, contents);
	postBinFile.close();
	//cerr << "Read bin post" << endl;
	Assert(contents.size() == posMap.size());
	for(int i = 0; i < posMap.size(); ++i){
	  int idx = 0;
	  double postIBD1 = contents[i][idx++];
	  double postIBD2 = 0.0, postPhaseA = 0.0, 
	    postPhaseB = 0.0, postCorrA = 0.0, postCorrB = 0.0;
	  if(hasIBD2){
	    postIBD2 = contents[i][idx++];
	  }
	  postPhaseA = contents[i][idx++];
	  postPhaseB = contents[i][idx++];
	  if(hasCorr){
	    postCorrA = contents[i][idx++];
	    postCorrB = contents[i][idx++];
	    double denom = postIBD1;
	    if(hasIBD2) 
	      denom = computeLogSumOfTwo(postIBD1, postIBD2);
	    postCorrA -= denom;
	    postCorrB -= denom;
	  }
	  else{
	    postCorrA = postPhaseA;
	    postCorrB = postPhaseB;
	  }
	  
	  if(postIBD1 > ibdCutoff)
	    indicators[i][0] = 1;
	  else if(hasIBD2 && postIBD2 > ibdCutoff)
	    indicators[i][0] = 2;
	  else if(postIBD1 < nonIBDcutoff && (!hasIBD2 || (hasIBD2 && postIBD2 < nonIBDcutoff)))
	    indicators[i][0] = 0;
	  else
	    indicators[i][0] = -1;
	  
	  indicators[i][1] = (int)(postPhaseA > cutoff);
	  indicators[i][2] = (int)(postPhaseB > cutoff);
	  indicators[i][3] = (int)(postCorrA > cutoff);
	  indicators[i][4] = (int)(postCorrB > cutoff);
	}
      }
      */
      int totalTrans = countAllelePairsPost(data, posMap, data.num_haplos,
					    mrcas[r], gensA[r], gensB[r],
					    relIndices[0], relIndices[1], chrom, indicators,
					    letterFreq, errorFunction, snpsBefore, snpsAfter, outfile);
      outfile.flush();
      outfile.close(); 
      gettimeofday(&tv, NULL);
      end = tv.tv_sec + (tv.tv_usec/1000000.0);
      cerr << "Rel " << relStr << " (" << totalTrans << " trans, " << (end - start) << " sec)" << endl;
    }
  }
}
