#include "CountHaplosUtils.h"

/*
 * Return value:
 * the log-probability of observing the given allele at the given position
 * taking sequencing errors into account.
 */
double logFreqError(vector< vector<double> > &letterFreq,
		    vector< vector<double> > &errorFunction, 
		    int allele, int pos){
  Assert(pos >= 0 && pos < letterFreq.size());
  Assert(allele >= 0 && allele < errorFunction.size());

  double prob = 0.0;
  for(int c = 0; c < errorFunction.size(); ++c)
    prob += errorFunction[c][allele] * letterFreq[pos][c];
  return correctLog(prob);
}

/*
 * Return value:
 * the log-probability of observing the given sequence of alleles
 * at the given position, assuming that the alleles at each position
 * are independent of each other and taking into account sequencing 
 * errors.
 */
double haploLogFreqIndep(vector< vector<double> > &letterFreq,
			 vector< vector<double> > &errorFunction,
			 vector<int> &alleles, int startPos){
  Assert(startPos >= 0);
  Assert(startPos + alleles.size() - 1 < letterFreq.size()); 

  double logProb = 0.0;
  for(int i = 0; i < alleles.size(); ++i){
    logProb += logFreqError(letterFreq, errorFunction, 
			    alleles[i], startPos + i);
  }
  return logProb;
}

/*
 * Return value:
 * the log-probability of observing the given sequence of alleles
 * at the given position, assuming that the alleles at each position
 * are independent of each other.
 */
double haploLogFreqIndepNoError(vector< vector<double> > &letterFreq,
				vector<int> &alleles, int startPos){
  Assert(startPos >= 0);
  Assert(startPos + alleles.size() - 1 < letterFreq.size()); 

  double logProb = 0.0;
  for(int i = 0; i < alleles.size(); ++i){
    logProb += correctLog(letterFreq[startPos + i][alleles[i]]);
  }
  return logProb;
}

/*
 * Return value:
 * the log-frequency of the given sequence of alleles,
 * starting at the given position.
 * Each haplotype h in the data object is assigned a weight
 * Pr(h is transformed into the given sequence of alleles by
 * the error function).
 * The value returned is (log of the sum of weights for all
 * haplotypes)
 *
 * Only the first numHaplos haplotypes of the SNPData object
 * are considered (this can be used if we don't want to 
 * consider newly added haplotypes at the end of the SNPData).
*/
double haploLogFreq(SNPData &data, int numHaplos,
		    vector< vector<double> > &errorFunction,
		    vector<int> &alleles, int startPos, int idx){
  Assert(startPos >= 0);
  Assert(startPos + alleles.size() - 1 < data.seq_len); 
  Assert(numHaplos >= 0 && numHaplos <= data.num_haplos);
  
  /* No normalizaton needed
  double maxLogProb = 0.0;
  for(int i = 0; i < alleles.size(); ++i)
    maxLogProb += correctLog(errorFunction[alleles[i]][alleles[i]]);
  //log(N * max_prob) = log(N) + log(max_prob)
  maxLogProb += correctLog(numHaplos);
  */

  vector<double> logs;
  for(int h = 0; h < numHaplos; ++h){
    if(h == idx) continue;
    double logProb = 0.0;
    for(int i = 0; i < alleles.size(); ++i)
      logProb += correctLog(errorFunction[getAllele(data, h, startPos + i)][alleles[i]]);
    logs.push_back(logProb);
  }
  return computeLogSum(logs) - correctLog(numHaplos);// - maxLogProb; (no normalization needed)
}

/*
 * Return value:
 * the log-frequency of the given sequence of alleles,
 * starting at the given position.
 * Each haplotype h in the data object is assigned a weight
 * Pr(h is transformed into the given sequence of alleles by
 * the error function).
 * The value returned is (log of the sum of weights for all
 * haplotypes)
 *
 * Only the haplotypes of the SNPData object in the haplos vector
 * are considered.
*/
double haploLogFreq(SNPData &data, vector<int> &haplos,
		    vector<int> &exclude,
		    vector< vector<double> > &errorFunction,
		    vector<int> &alleles, int startPos, int idx){
  Assert(startPos >= 0);
  Assert(startPos + alleles.size() - 1 < data.seq_len); 
  
  /* No normalizaton needed
  double maxLogProb = 0.0;
  for(int i = 0; i < alleles.size(); ++i)
    maxLogProb += correctLog(errorFunction[alleles[i]][alleles[i]]);
  //log(N * max_prob) = log(N) + log(max_prob)
  maxLogProb += correctLog(numHaplos);
  */

  vector<double> logs;
  for(int h = 0; h < haplos.size(); ++h){
    bool ok = (haplos[h] != idx);
    for(int ex = 0; ex < exclude.size(); ++ex){
      if(haplos[h] == exclude[ex]){
	ok = false;
	break;
      }
    }
    if(ok){
      double logProb = 0.0;
      for(int i = 0; i < alleles.size(); ++i){
	logProb += correctLog(errorFunction[getAllele(data, haplos[h], startPos + i)][alleles[i]]);
      }
      logs.push_back(logProb);
    }
  }
  return computeLogSum(logs) - correctLog(haplos.size() - exclude.size());// - maxLogProb; (no normalization needed)
}

/*
 * Return value:
 * the log-frequency the given sequence of alleles,
 * starting at the given position.
 *
 * Only the first numHaplos haplotypes of the SNPData object
 * are considered (this can be used if we don't want to 
 * consider newly added haplotypes at the end of the SNPData).
*/
double haploLogFreqNoError(SNPData &data, int numHaplos,
			   vector<int> &alleles, int startPos){
  Assert(startPos >= 0);
  Assert(startPos + alleles.size() - 1 < data.seq_len); 
  Assert(numHaplos >= 0 && numHaplos <= data.num_haplos);

  int count = 0;
  for(int h = 0; h < numHaplos; ++h){
    int isSame = 1;
    for(int i = 0; i < alleles.size(); ++i){
      if((int)getAllele(data, h, startPos + i) != alleles[i]){
	isSame = 0;
	break;
      }
    }
    count += isSame;
  }
  return correctLog((float)count / numHaplos);// - maxLogProb; (no normalization needed)
}

/*
 * prevIBD: IBD status before transition
 * IBD: IBD status after transition
 * idx: either 1 or 2 indicating whether this is the 1st or 2nd
 * line for this transition (some transitions have 2 lines)
 * chr: chromosome of the transition
 * data: SNPData object from which haplotypes will be read
 * numHaplos: consider only the first numHaplos haplotypes to compute freq
 * letterFreq, errorFunction as usual
 * allelesA, allelesB: the haplotypes of A and B in the region of interest
 * startPos: starting position for the window in allelesA
 * nonIBDstart: position within the window where the nonIBD part starts
 * nonIBDend: similar
 * blockStart: starting position of the block ending with the current transition
 * transPos: (approximate) position of the current transition. This is 
 * different from the starting position of the current window.
 */
void printFreq(int prevIBD, int IBD, int idx,
	       int chr, SNPData &data, vector<int> &posMap, 
	       int numHaplos,
	       vector< vector<double> > &letterFreq,
	       vector< vector<double> > &errorFunction,
	       vector<int> &allelesA, vector<int> &allelesB, 
	       int blockStart, int winStart, int lastPos, int transPos,
	       int nonIBDstart, int nonIBDend, ostream &os, int idxA, int idxB){

  //Used in the old format.
  /*
  int nonIBDlen = nonIBDend - nonIBDstart + 1;
  vector<int> nonIBDallelesA(nonIBDlen, 0), nonIBDallelesB(nonIBDlen, 0);

  for(int j = 0; j < nonIBDlen; ++j){
    nonIBDallelesA[j] = allelesA[j + nonIBDstart - winStart];
    nonIBDallelesB[j] = allelesB[j + nonIBDstart - winStart];
  }
  */

  int mappedStart = posMap[blockStart];
  int mappedTransPos = posMap[transPos];
  int mappedLastPos = posMap[lastPos];


  /* 
   * NEW FORMAT
   * Information written:
   * 1,2: the IBD status before and after the transition.
   * 3: idx: some (few) transitions have 2 lines (see above).
   * 4: the chromosome of the transition
   * 5,6: the index of the first SNP of the block out of which we are 
   * transitioning and the index of the SNP right after this transition
   * 7,8: like 5,6 but mapped to the unpruned data
   * 9,10: genomic positions for 5,6
   * 11,12: genetic positions for 5,6
   * 13,14: freq of the haplotypes of A and B
   * 15-18: the end of the block out of which we are transitioning without 
   * including the "uncertain" part, that is, the last certain SNP of the 
   * block out of which we are transitioning. position in pruned data,
   * position in the unpruned data, genomic and genetic positions.
   */

  os << prevIBD << "\t" << IBD << "\t" << idx << "\t"
     << chr << "\t" 
     << blockStart << "\t" << transPos << "\t"
     << mappedStart << "\t" << mappedTransPos << "\t" 
     << (int)data.locus_genomic_pos[mappedStart] << "\t"
     << (int)data.locus_genomic_pos[mappedTransPos] << "\t" 
     << data.locus_genetic_pos[mappedStart] << "\t"
     << data.locus_genetic_pos[mappedTransPos] << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, allelesA, winStart, idxA) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, allelesB, winStart, idxB) << "\t"
     << lastPos << "\t" << mappedLastPos << "\t" 
     << (int)data.locus_genomic_pos[mappedLastPos] << "\t" 
     << data.locus_genetic_pos[mappedLastPos] << endl;

  /* 
   * OLD FORMAT
   * Information written:
   * 1,2: the IBD status before and after the transition.
   * 3: idx: some (few) transitions have 2 lines (see above).
   * 4: the chromosome of the transition
   * 5,6: the index of the first SNP of the block in the pruned data, and the 
   * index of the SNP right after the transition.
   * 7,8: the index of the first and last SNP of the block in the unpruned data
   * 9,10: the start and end genomic position of the block
   * 11,12: the start and end genetic position of the block
   * 13,14: freq of the haplotypes of A and B assuming positions are independent
   * 15,16: freq of the haplotypes of A and B
   * 17,18: freq of the haplotypes of A and B in the non-IBD part
   * 19-22: similar to transPos, but this position is the end of the block without 
   * including the "uncertain" part.
   */
  /*
  os << prevIBD << "\t" << IBD << "\t" << idx << "\t"
     << chr << "\t" 
     << blockStart << "\t" << transPos << "\t"
     << mappedStart << "\t" << mappedTransPos << "\t" 
     << (int)data.locus_genomic_pos[mappedStart] << "\t"
     << (int)data.locus_genomic_pos[mappedTransPos] << "\t" 
     << data.locus_genetic_pos[mappedStart] << "\t"
     << data.locus_genetic_pos[mappedTransPos] << "\t"
     << haploLogFreqIndep(letterFreq, errorFunction, allelesA, winStart) << "\t"
     << haploLogFreqIndep(letterFreq, errorFunction, allelesB, winStart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, allelesA, winStart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, allelesB, winStart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, nonIBDallelesA, nonIBDstart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, nonIBDallelesB, nonIBDstart) << "\t"
     << lastPos << "\t" << mappedLastPos << "\t" 
     << (int)data.locus_genomic_pos[mappedLastPos] << "\t" 
     << data.locus_genetic_pos[mappedLastPos] << endl;
  */
}


/*
 * prevIBD: IBD status before transition
 * IBD: IBD status after transition
 * idx: either 1 or 2 indicating whether this is the 1st or 2nd
 * line for this transition (some transitions have 2 lines)
 * chr: chromosome of the transition
 * data: SNPData object from which haplotypes will be read
 * numHaplos: consider only the first numHaplos haplotypes to compute freq
 * letterFreq, errorFunction as usual
 * allelesA, allelesB: the haplotypes of A and B in the region of interest
 * startPos: starting position for the window in allelesA
 * nonIBDstart: position within the window where the nonIBD part starts
 * nonIBDend: similar
 * blockStart: starting position of the block ending with the current transition
 * transPos: (approximate) position of the current transition. This is 
 * different from the starting position of the current window.
 */
void printFreq(int prevIBD, int IBD, int idx,
	       int chr, SNPData &data, vector<int> &posMap, 
	       vector<int> &haplos,
	       vector< vector<double> > &letterFreq,
	       vector< vector<double> > &errorFunction,
	       vector<int> &allelesA, vector<int> &allelesB, 
	       int blockStart, int winStart, int lastPos, int transPos,
	       int nonIBDstart, int nonIBDend, ostream &os, int idxA, int idxB,
	       vector<int> &excludeHaplosA, vector<int> &excludeHaplosB){

  //Used in the old format.
  /*
  int nonIBDlen = nonIBDend - nonIBDstart + 1;
  vector<int> nonIBDallelesA(nonIBDlen, 0), nonIBDallelesB(nonIBDlen, 0);

  for(int j = 0; j < nonIBDlen; ++j){
    nonIBDallelesA[j] = allelesA[j + nonIBDstart - winStart];
    nonIBDallelesB[j] = allelesB[j + nonIBDstart - winStart];
  }
  */

  int mappedStart = posMap[blockStart];
  int mappedTransPos = posMap[transPos];
  int mappedLastPos = posMap[lastPos];


  /* 
   * NEW FORMAT
   * Information written:
   * 1,2: the IBD status before and after the transition.
   * 3: idx: some (few) transitions have 2 lines (see above).
   * 4: the chromosome of the transition
   * 5,6: the index of the first SNP of the block out of which we are 
   * transitioning and the index of the SNP right after this transition
   * 7,8: like 5,6 but mapped to the unpruned data
   * 9,10: genomic positions for 5,6
   * 11,12: genetic positions for 5,6
   * 13,14: freq of the haplotypes of A and B
   * 15-18: the end of the block out of which we are transitioning without 
   * including the "uncertain" part, that is, the last certain SNP of the 
   * block out of which we are transitioning. position in pruned data,
   * position in the unpruned data, genomic and genetic positions.
   */

  os << prevIBD << "\t" << IBD << "\t" << idx << "\t"
     << chr << "\t" 
     << blockStart << "\t" << transPos << "\t"
     << mappedStart << "\t" << mappedTransPos << "\t" 
     << (int)data.locus_genomic_pos[mappedStart] << "\t"
     << (int)data.locus_genomic_pos[mappedTransPos] << "\t" 
     << data.locus_genetic_pos[mappedStart] << "\t"
     << data.locus_genetic_pos[mappedTransPos] << "\t"
     << haploLogFreq(data, haplos, excludeHaplosA, 
		     errorFunction, allelesA, winStart, idxA) << "\t"
     << haploLogFreq(data, haplos, excludeHaplosB, 
		     errorFunction, allelesB, winStart, idxB) << "\t"
     << lastPos << "\t" << mappedLastPos << "\t" 
     << (int)data.locus_genomic_pos[mappedLastPos] << "\t" 
     << data.locus_genetic_pos[mappedLastPos] << endl;

  /* 
   * OLD FORMAT
   * Information written:
   * 1,2: the IBD status before and after the transition.
   * 3: idx: some (few) transitions have 2 lines (see above).
   * 4: the chromosome of the transition
   * 5,6: the index of the first SNP of the block in the pruned data, and the 
   * index of the SNP right after the transition.
   * 7,8: the index of the first and last SNP of the block in the unpruned data
   * 9,10: the start and end genomic position of the block
   * 11,12: the start and end genetic position of the block
   * 13,14: freq of the haplotypes of A and B assuming positions are independent
   * 15,16: freq of the haplotypes of A and B
   * 17,18: freq of the haplotypes of A and B in the non-IBD part
   * 19-22: similar to transPos, but this position is the end of the block without 
   * including the "uncertain" part.
   */
  /*
  os << prevIBD << "\t" << IBD << "\t" << idx << "\t"
     << chr << "\t" 
     << blockStart << "\t" << transPos << "\t"
     << mappedStart << "\t" << mappedTransPos << "\t" 
     << (int)data.locus_genomic_pos[mappedStart] << "\t"
     << (int)data.locus_genomic_pos[mappedTransPos] << "\t" 
     << data.locus_genetic_pos[mappedStart] << "\t"
     << data.locus_genetic_pos[mappedTransPos] << "\t"
     << haploLogFreqIndep(letterFreq, errorFunction, allelesA, winStart) << "\t"
     << haploLogFreqIndep(letterFreq, errorFunction, allelesB, winStart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, allelesA, winStart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, allelesB, winStart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, nonIBDallelesA, nonIBDstart) << "\t"
     << haploLogFreq(data, numHaplos, errorFunction, nonIBDallelesB, nonIBDstart) << "\t"
     << lastPos << "\t" << mappedLastPos << "\t" 
     << (int)data.locus_genomic_pos[mappedLastPos] << "\t" 
     << data.locus_genetic_pos[mappedLastPos] << endl;
  */
}

/*
 * Finds the transitions based on the IBD states in indicators, 
 * and writes them in os, using the format described in printFreq.
 * indicators: m x 5 matrix, where m is the number of SNPs
 * columns are IBD state (0, 1, 2, or -1 for unknown),
 * phaseA, phaseB, corrA, corrB.
 * To simplify things I assume there are no phasing errors.
 * Return value:
 * number of transitions found
 */
int countAllelePairsPost(SNPData &data, vector<int> &posMap, int numHaplos,
			 int mrcas, int genA, int genB,
			 int idxA, int idxB, int chr, vector< vector<int> > &indicators,
			 vector< vector<double> > &letterFreq,
			 vector< vector<double> > &errorFunction,
			 int snpsBefore, int snpsAfter, ostream &os){
  Assert(snpsBefore > 0 && snpsAfter > 0);
  Assert(posMap.size() == indicators.size());
  int haploIdxA = data.diplos[idxA].haplos[0], haploIdxB  = data.diplos[idxB].haplos[0];
  int totalTrans = 0;
  
  //Find the first position for which we can determine tha IBD state.
  int pos = 0;
  int lastPos = -1;
  while(pos < indicators.size()){
    if(indicators[pos][0] > -1){
      lastPos = pos;
      break;
    }
    pos++;
  }
  //If we can't determine the IBD status for any position, we return 0.
  if(lastPos == -1)
    return 0;
  
  int firstKnown = lastPos;
  int blockStart = 0;
  //During the loop, lastPos will be the last position encountered, for 
  //which the IBD status is not -1. blockStart is the starting position
  //of the last IBD or non-IBD block.
  //                  i 
  //????11111111??????00000
  //    b   |  l         |
  //        --------------
  //b: blockStart
  //l: lastPos
  //underlined region: window for counting haplotypes,
  //assuming snpsBefore = snpsAfter = 3, and pruned and 
  //unpruned data are the same.
  for(int i = firstKnown + 1; i < indicators.size(); ++i){
    //if(i % 1000 == 0)
    //cerr << "countAllelePairsPost: pos " << i << endl;
    int IBD = indicators[i][0];
    //If we don't know the IBD status of this pos, 
    //or if the phasing is uncertain, we do nothing.
    if(IBD == -1 || (indicators[i][1] != indicators[lastPos][1]) ||
       (indicators[i][2] != indicators[lastPos][2])) 
      continue;

    int corrA = indicators[i][3];
    int corrB = indicators[i][4];
    int prevIBD = indicators[lastPos][0];
    int prevCorrA = indicators[lastPos][3];
    int prevCorrB = indicators[lastPos][4];
    
    int winStart = max(0, posMap[lastPos] - snpsBefore);
    //Don't let the window extend beyond the previous SNP in the pruned
    //data, since we know that consecutive loci in the pruned data
    //are uncorrelated (or not strongly correlated).
    if(lastPos > 0)
      winStart = max(winStart, posMap[lastPos - 1] + 1);
    int winEnd = min(posMap[i] + snpsAfter, data.seq_len - 1);
    if(i < indicators.size() - 1)
      winEnd = min(winEnd, posMap[i + 1] - 1);
    int winLen = winEnd - winStart + 1;

    if(IBD != prevIBD){
      totalTrans++;
      //I must get the correlated allele. I can only get this information from
      //the IBD part of the transition. That is, if position i - 1 is IBD and
      //position i is not, then I look at the correlation at position i - 1.
      //If position i - 1 is not IBD and position i is IBD, I look at the
      //correlation at position i.
      int choiceA = IBD? corrA: prevCorrA;
      int choiceB = IBD? corrB: prevCorrB;
      
      int nonIBDstart = IBD > prevIBD? winStart: posMap[i];
      int nonIBDend = IBD > prevIBD? posMap[lastPos]: winEnd;
      
      if(IBD + prevIBD == 1 || IBD + prevIBD == 3){
	//transition between 0 and 1 or between 2 and 1
	if(IBD + prevIBD == 3){
	  //This looks like:
	  //A: --------
	  //   ********
	  //B: --------
	  //   ^^^^****
	  //Here prevIBD = 1 and IBD = 2. We only consider the
	  //allele containing the transition (bottom allele of both
	  //individuals in this example).
	  choiceA = prevIBD == 1? 1 - prevCorrA: 1 - corrA;
	  choiceB = prevIBD == 1? 1 - prevCorrB: 1 - corrB;
	}
	vector<int> allelesA(winLen, 0), allelesB(winLen, 0);
	
	for(int j = 0; j < winLen; ++j){
	  allelesA[j] = getAllele(data, haploIdxA + choiceA, winStart + j);
	  allelesB[j] = getAllele(data, haploIdxB + choiceB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, numHaplos, letterFreq, errorFunction,
		  allelesA, allelesB, blockStart, winStart, lastPos, i, 
		  nonIBDstart, nonIBDend, os, haploIdxA + choiceA, haploIdxB + choiceB);
	blockStart = i;
      }
      else{
	//transition between 2 and 0.
	//In this case there are in fact 2 transitions, so we increment totalTrans again.
	totalTrans++;
	vector<int> allelesA0(winLen, 0), allelesB0(winLen, 0);
	vector<int> allelesA1(winLen, 0), allelesB1(winLen, 0);
      
	for(int j = 0; j < winLen; ++j){
	  allelesA0[j] = getAllele(data, haploIdxA + choiceA, winStart + j);
	  allelesB0[j] = getAllele(data, haploIdxB + choiceB, winStart + j);
	  allelesA1[j] = getAllele(data, haploIdxA + 1 - choiceA, winStart + j);
	  allelesB1[j] = getAllele(data, haploIdxB + 1 - choiceB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, numHaplos, letterFreq, errorFunction,
		  allelesA0, allelesB0, blockStart, winStart, lastPos, i,
		  nonIBDstart, nonIBDend, os, haploIdxA + choiceA, haploIdxB + choiceB);
	printFreq(prevIBD, IBD, 2, chr, data, posMap, numHaplos, letterFreq, errorFunction,
		  allelesA1, allelesB1, blockStart, winStart, lastPos, i,
		  nonIBDstart, nonIBDend, os, haploIdxA + 1 - choiceA, haploIdxB + 1 - choiceB);
	blockStart = i;
      }
    }
    else if(IBD == 1 && prevIBD == 1 && (corrA != prevCorrA || corrB != prevCorrB)){
      totalTrans++;
      
      if(((genA == -1 || genB == -1) && (corrA == prevCorrA || corrB == prevCorrB)) ||
	 (mrcas == 2 && (corrA == prevCorrA || corrB == prevCorrB) && 
	  ((genA == 0 && genB > 0) || (genA > 0 && genB == 0)))){
	//This can arise when one individual is an ancestor of the other.
	//For ex. if A is B's grandparent, then it is possible to have corrA != prevCorrA.
	//CorrB must be equal to prevCorrB in this case. 
	//A: ---------
	//   +++++++++
	//B: ----+++++
	//We match the corrB-th allele of B, with one of the alleles of A.
	//It doesn't really matter which one.
	   //This can also arise in the 2 0 X cases where X > 0.
	vector<int> allelesA(winLen, 0), allelesB(winLen, 0);
	
	for(int j = 0; j < winLen; ++j){
	  allelesA[j] = getAllele(data, haploIdxA + prevCorrA, winStart + j);
	  allelesB[j] = getAllele(data, haploIdxB + prevCorrB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, numHaplos, letterFreq, errorFunction,
		  allelesA, allelesB, blockStart, winStart, lastPos, i,
		  posMap[i], winEnd, os, haploIdxA + prevCorrA, haploIdxB + prevCorrB);
	blockStart = i;
      }
      else if(mrcas == 2 && genA == 0 && genB == 0){
	//In this case corrA != prevCorrA and corrB != prevCorrB. 
	//This can arise in the full-sib case, although it is unlikely.
	//A: --------
	//   ********
	//B: ----^^^^
	//   ++++****
	//We match prevCorrA with prevCorrB and corrA with corrB.
	//There are in fact two transitions, so we increment totalTrans again.
	totalTrans++;
	vector<int> allelesA0(winLen, 0), allelesB0(winLen, 0);
	vector<int> allelesA1(winLen, 0), allelesB1(winLen, 0);
      
	for(int j = 0; j < winLen; ++j){
	  allelesA0[j] = getAllele(data, haploIdxA + prevCorrA, winStart + j);
	  allelesB0[j] = getAllele(data, haploIdxB + prevCorrB, winStart + j);
	  allelesA1[j] = getAllele(data, haploIdxA + corrA, winStart + j);
	  allelesB1[j] = getAllele(data, haploIdxB + corrB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, numHaplos, letterFreq, errorFunction,
		  allelesA0, allelesB0, blockStart, winStart, lastPos, i,
		  posMap[i], winEnd, os, haploIdxA + prevCorrA, haploIdxB + prevCorrB);
	printFreq(prevIBD, IBD, 2, chr, data, posMap, numHaplos, letterFreq, errorFunction,
		  allelesA1, allelesB1, blockStart, winStart, lastPos, i,
		  winStart, posMap[lastPos], os, haploIdxA + corrA, haploIdxB + corrB);
	blockStart = i;
      }
    }
    lastPos = i;
  }
  return totalTrans;
}

int countAllelePairsPost(SNPData &data, vector<int> &posMap, vector<int> &haplos,
			 vector< vector<int> > &excludeHaplos,
			 int mrcas, int genA, int genB,
			 int idxA, int idxB, int chr, vector< vector<int> > &indicators,
			 vector< vector<double> > &letterFreq,
			 vector< vector<double> > &errorFunction,
			 int snpsBefore, int snpsAfter, ostream &os){
  Assert(snpsBefore > 0 && snpsAfter > 0);
  Assert(posMap.size() == indicators.size());
  int haploIdxA = data.diplos[idxA].haplos[0], haploIdxB  = data.diplos[idxB].haplos[0];
  int totalTrans = 0;
  
  //Find the first position for which we can determine tha IBD state.
  int pos = 0;
  int lastPos = -1;
  while(pos < indicators.size()){
    if(indicators[pos][0] > -1){
      lastPos = pos;
      break;
    }
    pos++;
  }
  //If we can't determine the IBD status for any position, we return 0.
  if(lastPos == -1)
    return 0;
  
  int firstKnown = lastPos;
  int blockStart = 0;
  //During the loop, lastPos will be the last position encountered, for 
  //which the IBD status is not -1. blockStart is the starting position
  //of the last IBD or non-IBD block.
  //                  i 
  //????11111111??????00000
  //    b   |  l         |
  //        --------------
  //b: blockStart
  //l: lastPos
  //underlined region: window for counting haplotypes,
  //assuming snpsBefore = snpsAfter = 3, and pruned and 
  //unpruned data are the same.
  for(int i = firstKnown + 1; i < indicators.size(); ++i){
    //if(i % 1000 == 0)
    //cerr << "countAllelePairsPost: pos " << i << endl;
    int IBD = indicators[i][0];
    //If we don't know the IBD status of this pos, 
    //or if the phasing is uncertain, we do nothing.
    if(IBD == -1 || (indicators[i][1] != indicators[lastPos][1]) ||
       (indicators[i][2] != indicators[lastPos][2])) 
      continue;

    int corrA = indicators[i][3];
    int corrB = indicators[i][4];
    int prevIBD = indicators[lastPos][0];
    int prevCorrA = indicators[lastPos][3];
    int prevCorrB = indicators[lastPos][4];
    
    int winStart = max(0, posMap[lastPos] - snpsBefore);
    //Don't let the window extend beyond the previous SNP in the pruned
    //data, since we know that consecutive loci in the pruned data
    //are uncorrelated (or not strongly correlated).
    if(lastPos > 0)
      winStart = max(winStart, posMap[lastPos - 1] + 1);
    int winEnd = min(posMap[i] + snpsAfter, data.seq_len - 1);
    if(i < indicators.size() - 1)
      winEnd = min(winEnd, posMap[i + 1] - 1);
    int winLen = winEnd - winStart + 1;

    if(IBD != prevIBD){
      totalTrans++;
      //I must get the correlated allele. I can only get this information from
      //the IBD part of the transition. That is, if position i - 1 is IBD and
      //position i is not, then I look at the correlation at position i - 1.
      //If position i - 1 is not IBD and position i is IBD, I look at the
      //correlation at position i.
      int choiceA = IBD? corrA: prevCorrA;
      int choiceB = IBD? corrB: prevCorrB;
      
      int nonIBDstart = IBD > prevIBD? winStart: posMap[i];
      int nonIBDend = IBD > prevIBD? posMap[lastPos]: winEnd;
      
      if(IBD + prevIBD == 1 || IBD + prevIBD == 3){
	//transition between 0 and 1 or between 2 and 1
	if(IBD + prevIBD == 3){
	  //This looks like:
	  //A: --------
	  //   ********
	  //B: --------
	  //   ^^^^****
	  //Here prevIBD = 1 and IBD = 2. We only consider the
	  //allele containing the transition (bottom allele of both
	  //individuals in this example).
	  choiceA = prevIBD == 1? 1 - prevCorrA: 1 - corrA;
	  choiceB = prevIBD == 1? 1 - prevCorrB: 1 - corrB;
	}
	vector<int> allelesA(winLen, 0), allelesB(winLen, 0);
	
	for(int j = 0; j < winLen; ++j){
	  allelesA[j] = getAllele(data, haploIdxA + choiceA, winStart + j);
	  allelesB[j] = getAllele(data, haploIdxB + choiceB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, haplos, letterFreq, errorFunction,
		  allelesA, allelesB, blockStart, winStart, lastPos, i, 
		  nonIBDstart, nonIBDend, os, haploIdxA + choiceA, haploIdxB + choiceB,
		  excludeHaplos[idxA], excludeHaplos[idxB]);
	blockStart = i;
      }
      else{
	//transition between 2 and 0.
	//In this case there are in fact 2 transitions, so we increment totalTrans again.
	totalTrans++;
	vector<int> allelesA0(winLen, 0), allelesB0(winLen, 0);
	vector<int> allelesA1(winLen, 0), allelesB1(winLen, 0);
      
	for(int j = 0; j < winLen; ++j){
	  allelesA0[j] = getAllele(data, haploIdxA + choiceA, winStart + j);
	  allelesB0[j] = getAllele(data, haploIdxB + choiceB, winStart + j);
	  allelesA1[j] = getAllele(data, haploIdxA + 1 - choiceA, winStart + j);
	  allelesB1[j] = getAllele(data, haploIdxB + 1 - choiceB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, haplos, letterFreq, errorFunction,
		  allelesA0, allelesB0, blockStart, winStart, lastPos, i,
		  nonIBDstart, nonIBDend, os, haploIdxA + choiceA, haploIdxB + choiceB,
		  excludeHaplos[idxA], excludeHaplos[idxB]);
	printFreq(prevIBD, IBD, 2, chr, data, posMap, haplos, letterFreq, errorFunction,
		  allelesA1, allelesB1, blockStart, winStart, lastPos, i,
		  nonIBDstart, nonIBDend, os, haploIdxA + 1 - choiceA, haploIdxB + 1 - choiceB,
		  excludeHaplos[idxA], excludeHaplos[idxB]);
	blockStart = i;
      }
    }
    else if(IBD == 1 && prevIBD == 1 && (corrA != prevCorrA || corrB != prevCorrB)){
      totalTrans++;

      if(((genA == -1 || genB == -1) && (corrA == prevCorrA || corrB == prevCorrB)) ||
	 (mrcas == 2 && (corrA == prevCorrA || corrB == prevCorrB) && 
	  ((genA == 0 && genB > 0) || (genA > 0 && genB == 0)))){      
	//This can arise when one individual is an ancestor of the other.
	//For ex. if A is B's grandparent, then it is possible to have corrA != prevCorrA.
	//CorrB must be equal to prevCorrB in this case. 
	//A: ---------
	//   +++++++++
	//B: ----+++++
	//We match the corrB-th allele of B, with one of the alleles of A.
	//It doesn't really matter which one.
	vector<int> allelesA(winLen, 0), allelesB(winLen, 0);
	
	for(int j = 0; j < winLen; ++j){
	  allelesA[j] = getAllele(data, haploIdxA + prevCorrA, winStart + j);
	  allelesB[j] = getAllele(data, haploIdxB + prevCorrB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, haplos, letterFreq, errorFunction,
		  allelesA, allelesB, blockStart, winStart, lastPos, i,
		  posMap[i], winEnd, os, haploIdxA + prevCorrA, haploIdxB + prevCorrB,
		  excludeHaplos[idxA], excludeHaplos[idxB]);
	blockStart = i;
      }
      else if(mrcas == 2 && genA == 0 && genB == 0){
	//In this case corrA != prevCorrA and corrB != prevCorrB. 
	//This can arise in the full-sib case, although it is unlikely.
	//A: --------
	//   ********
	//B: ----^^^^
	//   ++++****
	//We match prevCorrA with prevCorrB and corrA with corrB.
	//There are in fact two transitions, so we increment totalTrans again.
	totalTrans++;
	vector<int> allelesA0(winLen, 0), allelesB0(winLen, 0);
	vector<int> allelesA1(winLen, 0), allelesB1(winLen, 0);
      
	for(int j = 0; j < winLen; ++j){
	  allelesA0[j] = getAllele(data, haploIdxA + prevCorrA, winStart + j);
	  allelesB0[j] = getAllele(data, haploIdxB + prevCorrB, winStart + j);
	  allelesA1[j] = getAllele(data, haploIdxA + corrA, winStart + j);
	  allelesB1[j] = getAllele(data, haploIdxB + corrB, winStart + j);
	}
	printFreq(prevIBD, IBD, 1, chr, data, posMap, haplos, letterFreq, errorFunction,
		  allelesA0, allelesB0, blockStart, winStart, lastPos, i,
		  posMap[i], winEnd, os, haploIdxA + prevCorrA, haploIdxB + prevCorrB,
		  excludeHaplos[idxA], excludeHaplos[idxB]);
	printFreq(prevIBD, IBD, 2, chr, data, posMap, haplos, letterFreq, errorFunction,
		  allelesA1, allelesB1, blockStart, winStart, lastPos, i,
		  winStart, posMap[lastPos], os, haploIdxA + corrA, haploIdxB + corrB,
		  excludeHaplos[idxA], excludeHaplos[idxB]);
	blockStart = i;
      }
    }
    lastPos = i;
  }
  return totalTrans;
}

int readConfig(istream &configFile, vector<int> &mrcas,
		vector<int> &gensA, vector<int> &gensB){
  string str;
  int maxGen = 2;
  while(!configFile.eof()){
    str = "";
    configFile >> str;
    if(str.size() > 0 && str.at(0) == 'u'){
      mrcas.push_back(0);
      gensA.push_back(0);
      gensB.push_back(0);
    }
    else if(str.size() > 0 && (str.at(0) == '1' || str.at(0) == '2')){
      stringstream stream(str);
      int num;
      stream >> num;
      mrcas.push_back(num);
      configFile >> str;
      stream.clear(); 
      stream.str(str);
      stream >> num;
      gensA.push_back(num);
      if(num > maxGen)
	maxGen = num;
      configFile >> str;
      stream.clear();
      stream.str(str);
      stream >> num;
      gensB.push_back(num);
      if(num > maxGen)
	maxGen = num;
    }
  }
  return maxGen;
}
