#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <math.h>
#include <sys/time.h>
#include "System.h"
#include "Utility.h"
#include "SNPData.h"
#include "Simulate.h"
#include "RelationshipModel.h"
#include "RelationshipUtils.h"

/*
 * Computes likelihoods and posteriors for a given pair.
 * Arguments:
 * - configFile: lines of the form <MRCAs>\t<gensA>\t<gensB> or "unrel".
 * - restoreBinPath: path to snpprofiledata.bin used to restore the individuals.
 * - binPath: path to snpprofiledata.bin used to compute likelihoods.
 * That is, the individuals are assumed to have been created using the file at
 * restoreBinPath, but they will be added to the binPath and the likelihoods will
 * be computed based on binPath.
 * - pairPath: path to file with stored individuals.
 * - outPath: prefix (including path) of the output files.
 * - outSuf: suffix of output likelihood files. 
 * Likelihoods will be stored in <outPath>_<outSuf>.logProb.
 * Posteriors for relationship X_Y_Z will be stored in <outPath>_X_Y_Z.postibd.
 * post: whether to compute posteriors (1) or just likelihoods (0).
 */
int Main(vector<string> args){
  AssertMsg(args.size() == 9,
	    "Usage: LikelihoodTest3 configFile restoreBinPath binPath pairPath outPath outSuf post switchErr");
  
  string configPath = args[1], restoreBinPath = args[2],
    binPath = args[3], pairPath = args[4], outpath = args[5], outSuf = args[6];
  ifstream pairFile(pairPath.c_str());
  ifstream configFile(configPath.c_str());
  ofstream outfile((outpath + '_' + outSuf + ".logProb").c_str());
  stringstream input(args[7]);
  bool doPost;
  input >> doPost;
  double switchErr = 0.0;
  input.clear();
  input.str(args[8]);
  input >> switchErr;
  
  timeval tv;
  double start, end;

  gettimeofday(&tv, NULL);
  start = tv.tv_sec + (tv.tv_usec/1000000.0);  
  //Read SNPProfile and SNPData.
  SNPProfile profile, restoreProfile;
  SNPData data, restoreData;
  //OK, this is not the best way to compare paths but I'm lazy right now.
  if(restoreBinPath != binPath)
    ReadBinSNPProfileData(restoreBinPath, restoreProfile, restoreData);
  ReadBinSNPProfileData(binPath, profile, data);
  data.SetAllDiploSwitch(SNPData::switch_encode(switchErr));

  //Read stored individuals
  Simulate::Diplo indivA, indivB;
  vector<int> relIndices(2, 0);
  if(restoreBinPath != binPath){
    relIndices[0] = restoreIndividual(restoreData, data, indivA, pairFile);
    relIndices[1] = restoreIndividual(restoreData, data, indivB, pairFile);
  }
  else{
    relIndices[0] = restoreIndividual(data, data, indivA, pairFile);
    relIndices[1] = restoreIndividual(data, data, indivB, pairFile);
  }
  pairFile.close();
  gettimeofday(&tv, NULL);
  end = tv.tv_sec + (tv.tv_usec/1000000.0);
  cerr << "Read individuals from " << pairPath << " and stored them at indices " 
       << relIndices[0] << " and " << relIndices[1] << " ("
       << (end - start) << " sec)" << endl;

  //Read the configuration file into vectors.
  int maxGen = 2;
  vector<int> mrcas, gensA, gensB;
  string str;
  while(!configFile.eof()){
    str = "";
    configFile >> str;
    if(str.size() > 0 && str.at(0) == 'u'){
      mrcas.push_back(0);
      gensA.push_back(0);
      gensB.push_back(0);
    }
    else if(str.size() > 0 && (str.at(0) == '1' || str.at(0) == '2')){
      stringstream stream(str);
      int num;
      stream >> num;
      mrcas.push_back(num);
      configFile >> str;
      stream.clear(); 
      stream.str(str);
      stream >> num;
      gensA.push_back(num);
      if(num > maxGen)
	maxGen = num;
      configFile >> str;
      stream.clear();
      stream.str(str);
      stream >> num;
      gensB.push_back(num);
      if(num > maxGen)
	maxGen = num;
    }
  }
  configFile.close();

  //Create the data-structures required for the models.
  gettimeofday(&tv, NULL);
  start = tv.tv_sec + (tv.tv_usec/1000000.0);  
  vector< vector<double> > errorFunction;
  errorFunction.resize(4);
  for(int i = 0; i < 4; ++i){
    errorFunction[i].resize(4, 0.003);
    errorFunction[i][i] = 0.991;
  }
  double pseudocount = 0.01;
  vector< vector<double> > letterFreq;
  computeLetterFreq(data, pseudocount, letterFreq, 4);
  Assert(errorFunction.size() == letterFreq[0].size());

  vector< vector<int> > pascalTri;
  computePascalTri(pascalTri, maxGen);
  vector< vector<double> > logRecomb, logNoRecomb;
  computeRecomb(data, logRecomb, logNoRecomb, maxGen);
  gettimeofday(&tv, NULL);
  end = tv.tv_sec + (tv.tv_usec/1000000.0);

  //Set the phasing error
  data.SetAllDiploSwitch(SNPData::switch_encode(switchErr));
  cerr << "Switch error will be set to " << switchErr << endl;
  cerr << "Created initial data structures (" << (end - start) << " sec)" << endl;

  //Write the output file headers.
  outfile << "#err\t";
  for(int i = 0; i < errorFunction[0].size() - 1; ++i)
    outfile << errorFunction[0][i] << ",";
  outfile << errorFunction[0][errorFunction[0].size() - 1] << endl;

  outfile << "#pseudocount\t " << pseudocount << endl;

  outfile << "#";

  for(int i = 0; i < mrcas.size(); ++i){
    if(mrcas[i] > 0){
      stringstream stream("");
      stream << mrcas[i] << "_";
      if(gensA[i] > -1) stream << gensA[i] << "_";
      else stream << "n1_";
      if(gensB[i] > -1) stream << gensB[i];
      else stream << "n1";      
      outfile << stream.str();
    }
    else
      outfile << "unrel";
    if(i < mrcas.size() - 1)
      outfile << "\t";
  }
  outfile << endl;
  
  vector< vector<double> > forwardProbs, backwardProbs, posteriors;
  //vector<string> statesStr;

  //Compute likelihoods, forward, and backward probabilities.
  for(int i = 0; i < mrcas.size(); ++i){
    gettimeofday(&tv, NULL);
    start = tv.tv_sec + (tv.tv_usec/1000000.0);  
    double seqProb;
    if(mrcas[i] == 2){
      TwoMRCAModel model(data, relIndices[0], relIndices[1], 
			 letterFreq, errorFunction, 
			 pascalTri, logRecomb,
			 logNoRecomb, gensA[i], gensB[i]);
      seqProb = model.factorialForward(forwardProbs);
      if(doPost){
	model.factorialBackward(backwardProbs);
	//model.getBinaryStates(statesStr);
	model.simplePosteriorIBD(forwardProbs, backwardProbs, seqProb, posteriors);
      }
    }
    else if(mrcas[i] == 1){
      OneMRCAModel2 model(data, relIndices[0], relIndices[1], 
			  letterFreq, errorFunction, 
			  pascalTri, logRecomb,
			  logNoRecomb, gensA[i], gensB[i]);
      seqProb = model.factorialForward(forwardProbs);
      if(doPost){
	model.factorialBackward(backwardProbs);
	//model.getBinaryStates(statesStr);
	model.simplePosteriorIBD(forwardProbs, backwardProbs, seqProb, posteriors);
      }
    }
    else{
      UnrelatedModel unrelated(data, relIndices[0], relIndices[1],
			       letterFreq, errorFunction);
      seqProb = unrelated.logLikelihood();
    }     
    outfile << seqProb;
    //Write the likelihood
    if(i < mrcas.size() - 1)
      outfile << "\t";
    //Write the forward and backward probabilities in separate files.
    if(mrcas[i] > 0 && doPost){
      stringstream stream("");
      stream << outpath << "_" << mrcas[i] << "_";
      if(gensA[i] > -1) stream << gensA[i] << "_";
      else stream << "n1_";
      if(gensB[i] > -1) stream << gensB[i];
      else stream << "n1"; 
      stream << ".postibd";

      /*
      gzostream posteriorFile(stream.str());
      for(int k = 0; k < posteriors.size(); ++k){
	for(int m = 0; m < posteriors[k].size(); ++m){
	  stringstream helper("");
	  helper << posteriors[k][m];
	  if(m < posteriors[k].size() - 1)
	    helper << "\t";
	  Write(posteriorFile, helper.str());
	}
	Write(posteriorFile, "\n");
	posteriorFile.flush();
      }
      */

      
      ofstream posteriorFile(stream.str().c_str());
      //WriteBin(posteriorFile, posteriors);
      printGenericMatrix(posteriors, posteriorFile);
      posteriorFile.flush();
      posteriorFile.close();
      //stringstream command("");
      //command << "gzip " << stream.str();
      //system(command.str().c_str());

      /*
      stringstream stream("");
      stream << outpath << "_" << mrcas[i] << "_";
      if(gensA[i] > -1) stream << gensA[i] << "_";
      else stream << "n1_";
      if(gensB[i] > -1) stream << gensB[i];
      else stream << "n1"; 
      stringstream stream2("");
      stream2 << stream.str() << ".back";
      stream << ".for";

      ofstream posteriorFile(stream.str().c_str());
      posteriorFile << "#";
      for(int i = 0; i < statesStr.size(); ++i){
	posteriorFile << statesStr[i];
	if(i < statesStr.size() - 1)
	  posteriorFile << "\t";
      }
      posteriorFile << endl;
      printGenericMatrix(forwardProbs, posteriorFile);
      posteriorFile.close();
      ofstream posteriorFile2(stream2.str().c_str());
      posteriorFile2 << "#";
      for(int i = 0; i < statesStr.size(); ++i){
	posteriorFile2 << statesStr[i];
	if(i < statesStr.size() - 1)
	  posteriorFile2 << "\t";
      }
      posteriorFile2 << endl;
      printGenericMatrix(backwardProbs, posteriorFile2);
      posteriorFile2.close();
      */
    }
    gettimeofday(&tv, NULL);
    end = tv.tv_sec + (tv.tv_usec/1000000.0);
    cerr << "Computed probability for " << mrcas[i] << " " 
	 << gensA[i] << " " << gensB[i] 
	 << " (" << (end - start) << " sec)" << endl;

  }
  outfile << endl;
  outfile.flush();
  outfile.close();
}
