#include "RelationshipModel.h"

OneMRCAState2::OneMRCAState2(bool inSharedAncestral, 
			     bool inPhaseA, bool inPhaseB, 
			     bool inDonationsA, bool inDonationsB) : 
  RelationshipHMMState::RelationshipHMMState(), 
  sharedAncestral(inSharedAncestral), phaseA(inPhaseA), phaseB(inPhaseB),
  donationsA(inDonationsA), donationsB(inDonationsB), idx(-1),
  variables(NUM_VAR_1MRCA){
  //I won't update the variables until I need them.
  //updateVariables();
}

OneMRCAState2::OneMRCAState2(const OneMRCAState2 &copy) : 
  RelationshipHMMState::RelationshipHMMState(static_cast<RelationshipHMMState>(copy)),
  sharedAncestral(copy.sharedAncestral), phaseA(copy.phaseA), phaseB(copy.phaseB),
  donationsA(copy.donationsA), donationsB(copy.donationsB), 
  idx(copy.idx), variables(NUM_VAR_1MRCA){
  //I won't update the variables until I need them.
  //updateVariables();
}

OneMRCAState2& OneMRCAState2::operator=(const OneMRCAState2 &that){
  if(this != &that){
    RelationshipHMMState::operator=(that);
    sharedAncestral = that.sharedAncestral;
    phaseA = that.phaseA;
    phaseB = that.phaseB;
    donationsA = that.donationsA;
    donationsB = that.donationsB;
    idx = that.idx;
    variables.resize(NUM_VAR_1MRCA);
    updateVariables();
  }
  return *this;
}

OneMRCAState2::~OneMRCAState2(){
  OneMRCAState2::clearObj();
}

void OneMRCAState2::updateVariables(){
  variables[SHARED_ANC_IDX] = &sharedAncestral;
  variables[PHASEA_IDX] = &phaseA;
  variables[PHASEB_IDX] = &phaseB;
  variables[DONA_IDX] = &donationsA;
  variables[DONB_IDX] = &donationsB;
}

string OneMRCAState2::getBinaryString(){
  stringstream stream("");
  for(int i = 0; i < NUM_VAR_1MRCA; ++i)
    stream << (int)(*variables[i]);
  return stream.str();
}

void OneMRCAState2::print(ostream &os) const{
  os << "State " << idx << " (sharedAnc = " << sharedAncestral << ", ";
  os << "phaseA = " << phaseA << ", ";
  os << "phaseB = " << phaseB << ", ";
  os << "donA = " << donationsA << ", ";
  os << "donB = " << donationsB << ", ";
  os << "numCorr = " << numCorr << ", ";
  os << "corrA = " << corrA << ", ";
  os << "corrB = " << corrB << ", ";
  os << "logInitPr = " << logInitialProb << ")" << endl;
}

OneMRCAModel2::OneMRCAModel2(SNPData &inData, int inIdxA, int inIdxB,
			     vector< vector<double> > &inLetterFreq, 
			     vector< vector<double> > &inErrorFunction,
			     vector< vector<int> > &inPascalTri,
			     vector< vector<double> > &inLogRecomb,
			     vector< vector<double> > &inLogNoRecomb, 
			     int inNumGensA, int inNumGensB) :
  RelationshipHMM::RelationshipHMM(inData, inIdxA, inIdxB, inLetterFreq, 
				   inErrorFunction, inPascalTri,
				   inLogRecomb, inLogNoRecomb,
				   inNumGensA, inNumGensB),
  maxValues(NUM_VAR_1MRCA, 1){
  validateNumGens(numGensA, numGensB);
  maxValues[DONA_IDX] = (int)(numGensA > 0);
  maxValues[DONB_IDX] = (int)(numGensB > 0);
  initStates();
}

/*
 * Creates shallow copies of the pointers (e.g. data), but deep
 * copies of the non-pointer vectors (e.g. logRecomb). 
 */
OneMRCAModel2::OneMRCAModel2(const OneMRCAModel2 &copy) : 
  RelationshipHMM::RelationshipHMM(static_cast<RelationshipHMM>(copy)),
  maxValues(copy.maxValues), states(copy.states){
  //cerr << "In OneMRCAModel copy constructor" << endl;
  validateNumGens(numGensA, numGensB);
}

OneMRCAModel2::~OneMRCAModel2(){
  OneMRCAModel2::clearObj();
}

/*
 * Like the copy constructor, the assignment operator returns shallow 
 * copies of the pointers (e.g. data), but deep
 * copies of the non-pointer vectors (e.g. logRecomb). 
 */
OneMRCAModel2& OneMRCAModel2::operator=(const OneMRCAModel2 &that){
  if(this != &that){
    RelationshipHMM::operator=(that);
    maxValues = that.maxValues;
    states = that.states;
  }
  return *this;
}


/*
 * Return value:
 * An index (or ID) for the given state.
 * The states for a given level will be assigned consecutive integer
 * indices starting from 0.
 * The value returned by this function is also the index of the state
 * in the model's "states" vector. 
 * Since each variable is a bool (i.e. a bit), the state index
 * is in fact the decimal corresponding to the binary number
 * represented by the state's variables.
 */
int OneMRCAModel2::getStateIdx(const OneMRCAState2 &state){
  int pos = 0;
  int idx = 0;
  for(int var = 0; var < NUM_VAR_1MRCA; ++var){
    //This will be either 0 (if that variable has the value 0)
    //or 2^pos (if that variable has the value 1).
    idx += (1 << pos) * (*(state.variables[var]));
    pos += maxValues[var];
  }
  return idx;
}

/*
 * Returns the index of the state that is similar to the 
 * given state, but has the value of the "var" variable
 * reversed.
 * var is the index of the variable (from 0 to NUM_VAR_1MRCA - 1)
 */
int OneMRCAModel2::getInverseStateIdx(const OneMRCAState2& state, int var){
  int val = *(state.variables[var]);
  //The position of the var-th variable in the binary representation
  //is not fixed, but depends on how many of the previous variables
  //are redundant.
  int sumMax = 0;
  for(int v = 0; v < var; ++v)
    sumMax += maxValues[v];
  int multiplier = 1 << sumMax;
  //mult * (1 - val) - mult * val = mult - mult * val - mult * val = 
  //mult - 2 * mult * val = mult * (1 - 2 * val) = mult * (1 - (val << 1))
  return state.idx + multiplier * (1 - (val << 1));
}

/*
 * Return value:
 * the model's total number of states.
 */
int OneMRCAModel2::getNumStates(){
  int sumMax = 0;
  for(int var = 0; var < NUM_VAR_1MRCA; ++var)
    sumMax += maxValues[var];
  return 1 << sumMax;
}

void OneMRCAModel2::getBinaryStates(vector<string> &binStr){
  binStr.clear();
  //I know that the states vector has state with index idx
  //at the idx-th position.
  for(int i = 0; i < states.size(); ++i){
    binStr.push_back(states[i].getBinaryString());
  }
}

/*
static void validateNumGens(int inNumGensA, int inNumGensB){
  AssertMsg(inNumGensA >= -1 && inNumGensB >= -1 && inNumGensA + inNumGensB >= -1,
	    "Invalid number of generations for the 1-MRCA model.");
}
*/

/*
 * Initializes the model's vector of states. It also computes the 
 * derived variables of the states.
 */
void OneMRCAModel2::initStates(){
  states.clear(); 
  states.resize(getNumStates());
  for(int sharedAncestral = 0; sharedAncestral < 2; ++sharedAncestral){
    for(int phaseA = 0; phaseA < 2; ++phaseA){
      for(int phaseB = 0; phaseB < 2; ++phaseB){
	for(int donA = 0; donA <= maxValues[DONA_IDX]; ++donA){
	  for(int donB = 0; donB <= maxValues[DONB_IDX]; ++donB){
	    OneMRCAState2 state((bool)sharedAncestral, (bool)phaseA,
				(bool)phaseB, (bool)donA, (bool)donB);
	    state.updateVariables();
	    computeCorrelation(state);
	    computeLogInitialProb(state);
	    int stateIdx = getStateIdx(state);
	    state.idx = stateIdx;
	    states[stateIdx] = state;
	    //state.print(cerr);
	    //cerr << *(state.variables[0]) << endl;
	  }
	}
      }
    }
  }
}

/*
 * Computes the variables numCorr, corrA, and corrB of the given state,
 * based on the state's "core" variables.
 */
void OneMRCAModel2::computeCorrelation(OneMRCAState2 &state){
  AssertMsg(!state.donationsA || numGensA > 0, "Invalid state");
  AssertMsg(!state.donationsB || numGensB > 0, "Invalid state");
  state.numCorr = 0;
  //Default value for corr variables when there are no correlated alleles is 0.
  state.corrA = 0;
  state.corrB = 0;
  if(!state.donationsA && !state.donationsB && 
     (numGensA == -1 || numGensB == -1 || state.sharedAncestral == 1)){
    state.numCorr = 1;
    state.corrA = state.phaseA;
    state.corrB = state.phaseB;
    if(numGensA == -1 && state.sharedAncestral)
      state.corrA = 1 - state.phaseA;
    else if(numGensB == -1 && state.sharedAncestral)
      state.corrB = 1 - state.phaseB;
  }
}

/*
 * Computes the log of the initial probability of the state.
 * The probability can be computed as:
 * P(this state is the initial state) = 
 * P(phaseA)*P(phaseB)*P(sharedAncestral)*P(donationsA)*P(donationsB) 
 * The probabilities P(donationsA) and P(donationsB) depend on the 
 * number of generations in the lineage of A and B respectively.
 * 
 * The probabilities are computed in log space.
 */
void OneMRCAModel2::computeLogInitialProb(OneMRCAState2 &state){
  AssertMsg((!state.donationsA || numGensA > 0), "Invalid state.");
  AssertMsg((!state.donationsB || numGensB > 0), "Invalid state.");

  //logProbA = log(P(donationsA)), and similarly for logProbB.
  //If numGensA <= 0, and assuming that the tests above did not fail,
  //donationsA will be false with probability 1, so logDonA = log(1.0) = 0.0.
  double logDonA = 0.0, logDonB = 0.0, logPhaseA = -log(2.0), logPhaseB = -log(2.0);

  if(numGensA > 0){
    //The probability of no donations in numGensA generations is (1/2)^numGensA.
    logDonA = (!state.donationsA)? log(2.0) * (-numGensA): log(1.0 - pow(2.0, -numGensA)); 
  }
  if(numGensB > 0)
    logDonB = (!state.donationsB)? log(2.0) * (-numGensB): log(1.0 - pow(2.0, -numGensB)); 
  
  //A phasing error at the very first position of the chromosome 
  //doesn't make sense, so I assign (almost) zero probability
  //to these initial states.
  if(numGensA == -1)
    logPhaseA = state.phaseA? correctLog(ZERO): correctLog(1.0 - ZERO);
  else if(numGensB == -1)
    logPhaseB = state.phaseB? correctLog(ZERO): correctLog(1.0 - ZERO);
  //The extra -log(2.0) is for sharedAncestral.
  state.logInitialProb = -log(2.0) + logPhaseA + logPhaseB + logDonA + logDonB;
}

void OneMRCAModel2::print(ostream &os){
  os << "OneMRCAModel: numGensA = " << numGensA << ", numGensB = " << 
    numGensB << ", idxA = " << idxA << ", idxB = " << idxB << endl;
  os << "Model's pascal triangle:" << endl;
  printGenericMatrix(*pascalTri, os);
  os << "Model's 10 first log-recomb probabilities:" << endl;
  printGenericMatrix(*logRecomb, os, 10);
  os << "Model's 10 first log-no-recomb probabilities:" << endl;
  printGenericMatrix(*logNoRecomb, os, 10);

  os << "Model's states (" << states.size() << "): " << endl;
  for(int i = 0; i < states.size(); ++i){
    os << i << "\t";
    states[i].print(os);
  }
}

/*
 * Return value:
 * the log-probability of transitioning from state currentState at position pos
 * to state nextState at position pos + 1, under the current model.
 */
double OneMRCAModel2::transitionProb(const OneMRCAState2& currentState, 
				     const OneMRCAState2& nextState, 
				     int pos, bool debug){
  AssertMsg(pos >= 0 && pos < (*data).seq_len - 1, "pos out of data range");
  AssertMsg(currentState.numCorr > -1, "uninitialized state");
  AssertMsg(nextState.numCorr > -1, "uninitialized state");

  if((*data).locus_chrom[pos] != (*data).locus_chrom[pos + 1]){
    if(debug)
      cerr << "In transitionProb: transitioning between states, prob = "
	   << nextState.logInitialProb << endl;
    return nextState.logInitialProb;//(nextState.logInitialProb < log(ZERO))? log(ZERO): nextState.logInitialProb;
  }

  double logProb = 0.0;
  
  //The probability of a switch between positions pos and pos + 1 in individual A.
  double switchProb = SNPData::switch_decode((*data).diplos[idxA].switches[pos]);
  if(debug){
    cerr << "In transitionProb: switchesA = " << (int)(*data).diplos[idxA].switches[pos] << 
      ", switchProbA = " << switchProb;
  }
  logProb += (currentState.phaseA == nextState.phaseA)? 
    correctLog(1.0 - switchProb): correctLog(switchProb);

  switchProb = SNPData::switch_decode((*data).diplos[idxB].switches[pos]);
  if(debug){
    cerr << ", switchesB = " << (int)(*data).diplos[idxB].switches[pos] << 
      ", switchProbB = " << switchProb << ", ";
  }
  logProb += (currentState.phaseB == nextState.phaseB)? 
    correctLog(1.0 - switchProb): correctLog(switchProb);
  if(debug) cerr << "logProb = " << logProb << endl;

  //There are two ways in which the immediate descendants of the MRCA (call them 
  //individual C and individual D) can share the same haplotype from the MRCA at 
  //both position pos and position pos + 1: Either no recombination occured
  //between pos and pos + 1 in both C and D (eg. both C and D got the maternal 
  //haplotype of the MRCA at pos and at pos + 1), or a recombination occured
  //in both C and D (eg. both C and D got the maternal haplotype of the MRCA at pos
  //and the paternal haplotype of the MRCA at pos + 1).
  if(debug){
    cerr << "In transitionProb: logRecomb[pos][1] = " <<  (*logRecomb)[pos][1] << 
      ", logRecomb[pos][2] = " << (*logRecomb)[pos][2] << endl;
    cerr << "In transitionProb: logNoRecomb[pos][1] = " <<  (*logNoRecomb)[pos][1] << 
      ", logNoRecomb[pos][2] = " << (*logNoRecomb)[pos][2] << endl;
  }
  if(numGensA > -1 && numGensB > -1){
    if(currentState.sharedAncestral == nextState.sharedAncestral){
      //See Durbin et al. p. 78...
      logProb += computeLogSumOfTwo((*logNoRecomb)[pos][2], (*logRecomb)[pos][2]);
    }
    else
      logProb += log(2.0) + (*logRecomb)[pos][1] + (*logNoRecomb)[pos][1];
  }
  else{
    //There's an implicit assumption here that at most one of numGensA, numGensB is -1.
    //If, say, numGensA is -1, then sharedAncestral shows which allele of A was passed
    //down. So, the sharedAncestral variable changes value between pos and pos + 1,
    //iff there was a recombination between pos and pos + 1 on A.
    logProb += (currentState.sharedAncestral == nextState.sharedAncestral)?
      (*logNoRecomb)[pos][1]: (*logRecomb)[pos][1];
  }
  if(debug)
    cerr << "In transitionProb: logProb with sharedAncestral = " << logProb << endl;

  logProb += donationsTransitionProb(currentState.donationsA, 
				     nextState.donationsA, 
				     pos, numGensA, debug);
  logProb += donationsTransitionProb(currentState.donationsB, 
				     nextState.donationsB, 
				     pos, numGensB, debug);
  if(debug)
    cerr << "In transitionProb: Final logProb = " << logProb << endl; 
  return logProb;//((logProb < log(ZERO))? log(ZERO): logProb);
}

void OneMRCAModel2::ancestralTransProbs(int pos, vector< vector<double> > &probs){
  if(numGensA > -1 && numGensB > -1){
    probs[0][0] = probs[1][1] = computeLogSumOfTwo((*logNoRecomb)[pos][2], (*logRecomb)[pos][2]);
    probs[0][1] = probs[1][0] = log(2.0) + (*logRecomb)[pos][1] + (*logNoRecomb)[pos][1];
  }
  else{
    //There's an implicit assumption here that at most one of numGensA, numGensB is -1.
    //If, say, numGensA is -1, then sharedAncestral shows which allele of A was passed
    //down. So, the sharedAncestral variable changes value between pos and pos + 1,
    //iff there was a recombination between pos and pos + 1 on A.
    probs[0][0] = probs[1][1] = (*logNoRecomb)[pos][1];
    probs[0][1] = probs[1][0] = (*logRecomb)[pos][1];
  }
}


/*
 * Fills in the matrix probs with the transition probabilities for all
 * the model variables. The first dimension is always the variable
 * index. probs[idx] is a 2x2 matrix with the transition probabilities
 * between 0 and 1 for the variable with index idx.
 */
void OneMRCAModel2::computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs){
  donationsTransProbs(numGensA, pos, probs[DONA_IDX]);
  donationsTransProbs(numGensB, pos, probs[DONB_IDX]);
  switchProbs(idxA, pos, probs[PHASEA_IDX]);
  switchProbs(idxB, pos, probs[PHASEB_IDX]);
  ancestralTransProbs(pos, probs[SHARED_ANC_IDX]);
}

double OneMRCAModel2::forward(){
  vector<double> forwardProbs(states.size());
  vector<double> nextForwardProbs(states.size());
  for(int k = 0; k < states.size(); ++k){
    forwardProbs[k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);    
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	transProb += transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED_ANC_IDX][states[s].sharedAncestral][states[k].sharedAncestral];
	logs.push_back(forwardProbs[s] + transProb);
      }
      nextForwardProbs[k] = emissionProb(states[k], i) + computeLogSum(logs);
    }
    forwardProbs.swap(nextForwardProbs);
  }
  return computeLogSum(forwardProbs);
}

double OneMRCAModel2::forward(vector< vector<double> > &forwardProbs){
  forwardProbs.clear();
  forwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));

  for(int k = 0; k < states.size(); ++k){
    //P(starting at state k) * P(emitting first set of symbols from state k)
    forwardProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED_ANC_IDX][states[s].sharedAncestral][states[k].sharedAncestral];
	transProb += transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	logs.push_back(forwardProbs[i - 1][s] + transProb);
	//logs.push_back(forwardProbs[i - 1][s] + transitionProb(states[s], states[k], i - 1));
      }
      forwardProbs[i][k] = emissionProb(states[k], i) + computeLogSum(logs);
    }
  }
  return computeLogSum(forwardProbs.back());
}

/*
 * Similar to forward(vector< vector<double> > &), but using the
 * forward algorithm for factorial HMMs.
 */
double OneMRCAModel2::factorialForward(vector< vector<double> > &forwardProbs){
  forwardProbs.clear();
  forwardProbs.resize((*data).seq_len);
  forwardProbs[0].resize(states.size(), 0.0);

  for(int k = 0; k < states.size(); ++k){
    //states[k].print(cerr);
    //cerr << *(states[k].variables[SHARED_ANC_IDX]) << endl;
    //cerr << getInverseStateIdx(states[k], 0) << endl;
    //P(starting at state k) * P(emitting first set of symbols from state k)
    forwardProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }

  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    forwardProbs[i].resize(states.size(), 0.0);
    computeAllTransProbs(i - 1, transProbs);
    vector<double> prevForwardProbs(forwardProbs[i - 1]);
    for(int var = 0; var < NUM_VAR_1MRCA; ++var){
      for(int k = 0; k < states.size(); ++k){
	int val = *(states[k].variables[var]);
	int zeroIdx = (val == 0)? states[k].idx: getInverseStateIdx(states[k], var);
	int oneIdx = (val == 1)? states[k].idx: getInverseStateIdx(states[k], var);
	forwardProbs[i][k] = (maxValues[var] == 0)? prevForwardProbs[k]:
	  computeLogSumOfTwo(transProbs[var][0][val] + prevForwardProbs[zeroIdx],
			     transProbs[var][1][val] + prevForwardProbs[oneIdx]);
      }
      swap(forwardProbs[i], prevForwardProbs);
    }
    for(int k = 0; k < states.size(); ++k){
      forwardProbs[i][k] = emissionProb(states[k], i) + prevForwardProbs[k];
    }
  }
  return computeLogSum(forwardProbs.back());
}

/*
 * Staightforward implementation of the backward algorithm in log-space.
 * Computes log-backward probabilities for the current model,
 * and returns them in a matrix M, such that M[k][i] is 
 * log(b_k(i)) (see Durbin et al. p. 59)
 */
double OneMRCAModel2::backward(vector< vector<double> > &backwardProbs){
  backwardProbs.clear();
  backwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));

  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    computeAllTransProbs(i, transProbs);
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[PHASEA_IDX][states[k].phaseA][states[s].phaseA];
	transProb += transProbs[PHASEB_IDX][states[k].phaseB][states[s].phaseB];
	transProb += transProbs[SHARED_ANC_IDX][states[k].sharedAncestral][states[s].sharedAncestral];
	transProb += transProbs[DONA_IDX][states[k].donationsA][states[s].donationsA];
	transProb += transProbs[DONB_IDX][states[k].donationsB][states[s].donationsB];
	logs.push_back(backwardProbs[i + 1][s] + transProb + emissionProb(states[s], i + 1));
	//logs.push_back(transitionProb(states[k], states[s], i) + 
	//       emissionProb(states[s], i + 1) + backwardProbs[s][i + 1]);
      }
      backwardProbs[i][k] = computeLogSum(logs);
    }
  }
  
  vector<double> firstCol(states.size(), 0.0);
  for(int k = 0; k < states.size(); ++k){
    firstCol[k] = backwardProbs[0][k] + emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  return computeLogSum(firstCol);
}

double OneMRCAModel2::factorialBackward(vector< vector<double> > &backwardProbs){
  backwardProbs.clear();
  backwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));
  
  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  vector<double> prevBackwardProbs(states.size(), 0.0);
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    for(int k = 0; k < states.size(); ++k){
      prevBackwardProbs[k] = emissionProb(states[k], i + 1) + backwardProbs[i + 1][k];
    }
    computeAllTransProbs(i, transProbs);
    for(int var = NUM_VAR_1MRCA - 1; var >= 0; --var){
      for(int k = 0; k < states.size(); ++k){
	int val = *(states[k].variables[var]);
	int zeroIdx = (val == 0)? states[k].idx: getInverseStateIdx(states[k], var);
	int oneIdx = (val == 1)? states[k].idx: getInverseStateIdx(states[k], var);
	backwardProbs[i][k] = (maxValues[var] == 0)? prevBackwardProbs[k]:
	  computeLogSumOfTwo(transProbs[var][val][0] + prevBackwardProbs[zeroIdx],
			     transProbs[var][val][1] + prevBackwardProbs[oneIdx]);
      }
      if(var > 0)
	swap(backwardProbs[i], prevBackwardProbs);
    }
  }
  
  for(int k = 0; k < states.size(); ++k){
    prevBackwardProbs[k] = backwardProbs[0][k] + emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  return computeLogSum(prevBackwardProbs);
}

double OneMRCAModel2::viterbi(vector<int> &stateIdx){
  vector< vector<double> > viterbiProbs((*data).seq_len, vector<double>(states.size(), 0.0));
  vector< vector<int> > maxStates((*data).seq_len, vector<int>(states.size(), -1));
  stateIdx.resize((*data).seq_len);

  for(int k = 0; k < states.size(); ++k){
    //P(starting at state k) * P(emitting first set of symbols from state k)
    viterbiProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);
    for(int k = 0; k < states.size(); ++k){
      double maxVal = correctLog(0.0);
      int maxState = -1;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED_ANC_IDX][states[s].sharedAncestral][states[k].sharedAncestral];
	transProb += transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	transProb += viterbiProbs[i - 1][s];
	if(transProb > maxVal){
	  maxVal = transProb;
	  maxState = s;
	}
      }
      viterbiProbs[i][k] = emissionProb(states[k], i) + maxVal;
      maxStates[i][k] = maxState;
    }
  }
  
  vector<double> lastProbs = viterbiProbs.back();
  double maxVal = lastProbs[0];
  int maxState = 0;
  for(int k = 1; k < states.size(); ++k){
    if(lastProbs[k] > maxVal){
      maxVal = lastProbs[k];
      maxState = k;
    }
  }
  stateIdx[(*data).seq_len - 1] = maxState;
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    stateIdx[i] = maxStates[i + 1][stateIdx[i + 1]];
  }
  return maxVal;
}

/*
 * Similar to forward(vector< vector<double> > &), but using the
 * forward algorithm for factorial HMMs.
 */
/*
double OneMRCAModel2::factorialForward(vector< vector<double> > &forwardProbs){
  forwardProbs.clear();
  forwardProbs.resize((*data).seq_len);
  forwardProbs[0].resize(states.size(), 0.0);

  for(int k = 0; k < states.size(); ++k){
    //states[k].print(cerr);
    //cerr << *(states[k].variables[SHARED_ANC_IDX]) << endl;
    //cerr << getInverseStateIdx(states[k], 0) << endl;
    //P(starting at state k) * P(emitting first set of symbols from state k)
    forwardProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }

  vector< vector< vector<double> > > transProbs(NUM_VAR_1MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    forwardProbs[i].resize(states.size(), 0.0);
    computeAllTransProbs(i - 1, transProbs);
    vector<double> prevForwardProbs(forwardProbs[i - 1]);
    for(int var = 0; var < NUM_VAR_1MRCA; ++var){
      for(int k = 0; k < states.size(); ++k){
	int val = *(states[k].variables[var]);
	int zeroIdx = (val == 0)? states[k].idx: getInverseStateIdx(states[k], var);
	int oneIdx = (val == 1)? states[k].idx: getInverseStateIdx(states[k], var);
	forwardProbs[i][k] = (maxValues[var] == 0)? prevForwardProbs[k]:
	  computeLogSumOfTwo(transProbs[var][0][val] + prevForwardProbs[zeroIdx],
			     transProbs[var][1][val] + prevForwardProbs[oneIdx]);
      }
      swap(forwardProbs[i], prevForwardProbs);
    }
    for(int k = 0; k < states.size(); ++k){
      forwardProbs[i][k] = emissionProb(states[k], i) + prevForwardProbs[k];
    }
  }
  return computeLogSum(forwardProbs.back());
}
*/

void OneMRCAModel2::simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
				       vector< vector<double> > &backwardProbs,
				       double seqProb, vector<double> &posteriorLogProb){
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len, 0.0);
  for(int i = 0; i < (*data).seq_len; ++i){
    vector<double> logs;
    for(int k = 0; k < states.size(); ++k){
      AssertMsg(states[k].numCorr > -1, 
		"Attempt to compute posterior IBD probabilities with uninitialized states.");
      if(states[k].numCorr > 0)
	logs.push_back(forwardProbs[k][i] + backwardProbs[k][i] - seqProb);
    }
    posteriorLogProb[i] = logs.empty()? correctLog(0.0): computeLogSum(logs);
  }
}

/*
 * After this method returns, posteriorLogProbs.size() will be equal to the
 * number of SNPs. posteriorLogProb[i] will be a vector of size 4.
 * posteriorLogProb[i][0] will be the posterior probability that numCorr at 
 * position i is 1.
 * posteriorLogProb[i][1] will be the posterior probability that numCorr at
 * position i is 2 (this is meaningless in most cases).
 * posteriorLogProb[i][2] will be the posterior probability that corrA at
 * position i is 1 (i.e. the correlated allele of A is 1. Notice that 
 * when numCorr is 0, that is, when there are no correlated alleles, 
 * corrA and corrB are always 0).
 * posteriorLogProb[i][3] will be the posterior probability that corrB at
 * position i is 1.
 */
void OneMRCAModel2::simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
				       vector< vector<double> > &backwardProbs,
				       double seqProb,
				       vector< vector<double> > &posteriorLogProb){
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len);
  bool computeCorr = true;
  for(int i = 0; i < (*data).seq_len; ++i){
    posteriorLogProb[i].resize(5, 0.0);
    vector<double> IBDlogs1, phaseAlogs, phaseBlogs, corrAlogs, corrBlogs;
    for(int k = 0; k < states.size(); ++k){
      //Posterior probability of that state.
      double stateProb = forwardProbs[i][k] + backwardProbs[i][k] - seqProb;
      if(states[k].numCorr == 1) IBDlogs1.push_back(stateProb);     
      if(states[k].phaseA == 1) phaseAlogs.push_back(stateProb);
      if(states[k].phaseB == 1) phaseBlogs.push_back(stateProb);
      if(states[k].numCorr > 0 && states[k].corrA == 1) corrAlogs.push_back(stateProb);
      if(states[k].numCorr > 0 && states[k].corrB == 1) corrBlogs.push_back(stateProb);
    }
    int idx = 0;
    posteriorLogProb[i][idx++] = IBDlogs1.empty()? correctLog(0.0): computeLogSum(IBDlogs1);
    posteriorLogProb[i][idx++] = phaseAlogs.empty()? correctLog(0.0): computeLogSum(phaseAlogs);
    posteriorLogProb[i][idx++] = phaseBlogs.empty()? correctLog(0.0): computeLogSum(phaseBlogs);
    posteriorLogProb[i][idx++] = corrAlogs.empty()? correctLog(0.0): computeLogSum(corrAlogs);
    posteriorLogProb[i][idx++] = corrBlogs.empty()? correctLog(0.0): computeLogSum(corrBlogs);
  }
}

void OneMRCAModel2::simpleDiscretePosteriorIBD(vector< vector<double> > &forwardProbs, 
					       vector< vector<double> > &backwardProbs,
					       double seqProb, double nonIBDcutoff, double IBDcutoff, 
					       vector< vector< double> > &posteriorLogProb){
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len);
  //double logNonIBDcutoff = correctLog(nonIBDcutoff);
  //double logIBDcutoff = correctLog(IBDcutoff);
  for(int i = 0; i < (*data).seq_len; ++i){
    posteriorLogProb[i].resize(3, 0.0);
    vector<double> IBDlogs1, IBDlogs2, corrAlogs, corrBlogs;
    for(int k = 0; k < states.size(); ++k){
      AssertMsg(states[k].numCorr > -1, 
		"Attempt to compute posterior IBD probabilities with uninitialized states.");
      double stateProb = forwardProbs[k][i] + backwardProbs[k][i] - seqProb;
      if(states[k].corrA == 1) corrAlogs.push_back(stateProb);
      if(states[k].corrB == 1) corrBlogs.push_back(stateProb);
      if(states[k].numCorr > 0) IBDlogs1.push_back(stateProb);
      //if(states[k].numCorr == 1) IBDlogs1.push_back(stateProb);
      //else if(states[k].numCorr == 2) IBDlogs2.push_back(stateProb);
    }
    double postProbIBD1 = IBDlogs1.empty()? correctLog(0.0): computeLogSum(IBDlogs1);
    posteriorLogProb[i][0] = postProbIBD1;
    //double postProbIBD2 = IBDlogs2.empty()? correctLog(0.0): computeLogSum(IBDlogs2);
    //if(postProbIBD1 > logIBDcutoff) posteriorLogProb[i][0] = 1;
    //else if(postProbIBD2 > logIBDcutoff) posteriorLogProb[i][0] = 2;
    //else if(postProbIBD1 < logNonIBDcutoff && postProbIBD2 < logNonIBDcutoff)
    //posteriorLogProb[i][0] = 0;
    //else posteriorLogProb[i][0] = -1;
    posteriorLogProb[i][1] = corrAlogs.empty()? correctLog(0.0): computeLogSum(corrAlogs);
    posteriorLogProb[i][2] = corrBlogs.empty()? correctLog(0.0): computeLogSum(corrBlogs);
  }
  /*
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len);
  double logNonIBDcutoff = correctLog(nonIBDcutoff);
  double logIBDcutoff = correctLog(IBDcutoff);
  for(int i = 0; i < (*data).seq_len; ++i){
    posteriorLogProb[i].resize(3, 0.0);
    vector<double> IBDlogs1, IBDlogs2, corrAlogs, corrBlogs;
    for(int k = 0; k < states.size(); ++k){
      AssertMsg(states[k].numCorr > -1, 
		"Attempt to compute posterior IBD probabilities with uninitialized states.");
      double stateProb = forwardProbs[k][i] + backwardProbs[k][i] - seqProb;
      if(states[k].corrA == 1) corrAlogs.push_back(stateProb);
      if(states[k].corrB == 1) corrBlogs.push_back(stateProb);
      if(states[k].numCorr == 1) IBDlogs1.push_back(stateProb);
      else if(states[k].numCorr == 2) IBDlogs2.push_back(stateProb);
    }
    double postProbIBD1 = IBDlogs1.empty()? correctLog(0.0): computeLogSum(IBDlogs1);
    double postProbIBD2 = IBDlogs2.empty()? correctLog(0.0): computeLogSum(IBDlogs2);
    if(postProbIBD1 > logIBDcutoff) posteriorLogProb[i][0] = 1;
    else if(postProbIBD2 > logIBDcutoff) posteriorLogProb[i][0] = 2;
    else if(postProbIBD1 < logNonIBDcutoff && postProbIBD2 < logNonIBDcutoff)
      posteriorLogProb[i][0] = 0;
    else posteriorLogProb[i][0] = -1;
    posteriorLogProb[i][1] = corrAlogs.empty()? correctLog(0.0): computeLogSum(corrAlogs);
    posteriorLogProb[i][2] = corrBlogs.empty()? correctLog(0.0): computeLogSum(corrBlogs);
  }
  */
}

/*
 * Creates naive IBD blocks based on the given posterior log-probabilities 
 * of IBD. All positions with log-probability > log(0.5) are considered IBD.
 * Consecutive IBD or non-IBD positions are "packed" into Segments.
 * The "left" of the Segment is the starting position of the corresponding 
 * IBD or non-IBD block, the right side is the ending position + 1.
 * Return value:
 * A vector of pairs <Segment, int> where Segment is an IBD or non-IBD block
 * and int is an indicator of whether this block is IBD or not.
 */
vector< pair<Segment, int> > OneMRCAModel2::posteriorIBDblocks(vector<double> posteriorIBDLogProbs){
  double logHalf = log(0.5);
  int lastStatus = -1;
  int lastEnd = 0;
  vector< pair<Segment, int> > blocks;
  for(int i = 0; i < (*data).seq_len; ++i){
    int status = (int)(posteriorIBDLogProbs[i] > logHalf);
    if((status != lastStatus) && (lastStatus > -1)){
      //Following the convention for the haplo segments, the
      //block extends from Segment.left() to Segment.right() - 1.
      Segment seg(lastEnd, i);
      blocks.push_back(pair<Segment, int>(seg, lastStatus));
      lastEnd = i;
    }
    lastStatus = status;
  }
  if(lastStatus > -1){
    Segment seg(lastEnd, (*data).seq_len);
    blocks.push_back(pair<Segment, int>(seg, lastStatus));
  }
  return blocks;
}

/*
 * After this method returns, posteriorLogProbs[k][i] will be the posterior
 * probability of the k-th state at the i-th position.
 */
void OneMRCAModel2::computePostLogProb(vector< vector<double> > &forwardProbs,
				       vector< vector<double> > &backwardProbs,
				       vector< vector<double> > &posteriorLogProbs){
  vector<double> lastColForward;
  for(int k = 0; k < states.size(); ++k){
    lastColForward.push_back(forwardProbs[k].back());
  }
  double seqLogProb = computeLogSum(lastColForward);

  posteriorLogProbs.clear();
  posteriorLogProbs.resize(states.size());
  for(int k = 0; k < states.size(); ++k){
    posteriorLogProbs[k].resize((*data).seq_len, 0.0);
    for(int i = 0; i < (*data).seq_len; ++i)
      posteriorLogProbs[k][i] = forwardProbs[k][i] + backwardProbs[k][i] - seqLogProb;
  }
}
