#include "RelationshipModel.h"

RelationshipHMM::RelationshipHMM(SNPData &inData, int inIdxA, int inIdxB,
				 vector< vector<double> > &inLetterFreq, 
				 vector< vector<double> > &inErrorFunction,
				 vector< vector<int> > &inPascalTri,
				 vector< vector<double> > &inLogRecomb,
				 vector< vector<double> > &inLogNoRecomb,
				 int inNumGensA, int inNumGensB) : 
  RelationshipModel::RelationshipModel(inData, inIdxA, inIdxB, inLetterFreq, inErrorFunction),
  pascalTri(&inPascalTri), logRecomb(&inLogRecomb), logNoRecomb(&inLogNoRecomb),
  numGensA(inNumGensA), numGensB(inNumGensB){
  //cerr << "In RelationshipHMM::RelationshipHMM(1)" << endl;
}

RelationshipHMM::RelationshipHMM(const RelationshipHMM &copy) : 
  RelationshipModel::RelationshipModel(*(copy.data), copy.idxA, copy.idxB, *(copy.letterFreq), *(copy.errorFunction)),
  numGensA(copy.numGensA), numGensB(copy.numGensB), logRecomb(copy.logRecomb), logNoRecomb(copy.logNoRecomb), 
  pascalTri(copy.pascalTri){}

RelationshipHMM::~RelationshipHMM(){
  RelationshipHMM::clearObj();
}

RelationshipHMM& RelationshipHMM::operator=(const RelationshipHMM &that){
  if(this != &that){
    RelationshipModel::operator=(that);
    numGensA = that.numGensA;
    numGensB = that.numGensB;
    pascalTri = that.pascalTri;
    logRecomb = that.logRecomb;
    logNoRecomb = that.logNoRecomb;
  }
  return *this;
}

/*
 * Return value:
 * the log-probability of transitioning from currentDonations at position pos to 
 * nextDonations at position pos + 1, where currentDonations and nextDonations are 
 * either true or false indicating whether there was a non-lineage donation in
 * numGens generations or not.
 */
double RelationshipHMM::donationsTransitionProb(bool currentDonations, bool nextDonations, 
						int pos, int numGens, bool debug){
  if(numGens < 1){
    AssertMsg(!currentDonations && !nextDonations, 
	      "Invalid value for the number of non-lineage donations.");
    //If nunGens < 1, then we can only transition from 0 donations to
    //0 donations with probability 1 (i.e. log-prob 0).
    return 0.0;
  }
  double donationProb = 0.0;
  if(currentDonations){
    vector<double> logs;
    //If there were i donations at position pos and 0 donations at position pos + 1,
    //then there were exactly i recombinations between pos and pos + 1.
    //i is at least 1 (since currentDonations is true) and at most numGens.
    for(int i = 1; i <= numGens; ++i){
      logs.push_back(log((*pascalTri)[numGens][i]) + 
		     (*logRecomb)[pos][i] + (*logNoRecomb)[pos][numGens - i]);
      if(debug){
	cerr << "In donationsTransitionProb: log-prob for " <<
	  i << " donations " << logs.back() << endl;
      }
    }
    donationProb = computeLogSum(logs) - log(pow(2.0, numGens) - 1.0);
  }
  else{
    donationProb = (*logNoRecomb)[pos][numGens];
  }
  return nextDonations? correctLog(1.0 - exp(donationProb)): donationProb;
} 

void RelationshipHMM::donationsTransProbs(int numGens, int pos, vector< vector<double> > &probs){
  if(numGens < 1){
    //Only the first position of the array will be used, we
    //don't care about the rest.
    probs[0][0] = probs[0][1] = probs[1][0] = probs[1][1] = 0.0;
  }
  else{
    double noDon2noDon = donationsTransitionProb(false, false, pos, numGens, false);
    probs[0][0] = noDon2noDon; //from no donations to no donations
    probs[0][1] = correctLog(1.0 - exp(noDon2noDon)); //from no donations to donations
    double don2noDon = donationsTransitionProb(true, false, pos, numGens, false);
    probs[1][0] = don2noDon; //from donations to no donations
    probs[1][1] = correctLog(1.0 - exp(don2noDon)); //from donations to donations
  }
}

void RelationshipHMM::switchProbs(int diplo, int pos, vector< vector<double> > &probs){
  double switchProb = SNPData::switch_decode((*data).diplos[diplo].switches[pos]);
  probs[0][0] = probs[1][1] = correctLog(1.0 - switchProb);
  probs[0][1] = probs[1][0] = correctLog(switchProb);
}

/*
 * This function assumes that the emission probability depends only on
 * the values of the variables numCorr, corrA, and corrB of the state.
 *
 * Return value:
 * the log probability of emitting the observed alleles at position pos of A and B.
 * Ensures that the smallest value returned is log(ZERO).
 */
double RelationshipHMM::emissionProb(const RelationshipHMMState &state, int pos, bool debug){
  AssertMsg(state.corrA > -1 && state.corrB > -1 && state.numCorr > -1, 
	    "emissionProb called for a state with undefined correlation variables.");

  //Indices in the haplos vector of the SNPData object.
  int haploAIdx = 2 * idxA;
  int haploBIdx = 2 * idxB;

  if(debug){
    cerr << "In emissionProb: letter frequencies at pos " << pos << ":";
    for(int letter = 0; letter < (*errorFunction).size(); ++letter)
      cerr << " " << (*letterFreq)[pos][letter];
    cerr << endl;
    cerr << "In emissionProb: indiv A alleles: " << 
      (int)getAllele(*data, haploAIdx, pos) << ", " <<
      (int)getAllele(*data, haploAIdx + 1, pos) << endl;
    cerr << "In emissionProb: indiv B alleles: " << 
      (int)getAllele(*data, haploBIdx, pos) << ", " <<
      (int)getAllele(*data, haploBIdx + 1, pos) << endl;
  }
  if(state.numCorr > 0){
    double probCorr1 = 0.0, probCorr2 = 0.0, probUncorr1 = 0.0, probUncorr2 = 0.0;
    for(int letter = 0; letter < (*errorFunction).size(); ++letter){
      //Compute sum(P(c in population) * P(c turned into whatever A has in the 
      //shared allele) * P(c turned into whatever B has in the shared allele)),
      //where the sum is taken over all possible letters.
      probCorr1 += (*letterFreq)[pos][letter] * 
	(*errorFunction)[letter][getAllele(*data, haploAIdx + state.corrA, pos)] * 
	(*errorFunction)[letter][getAllele(*data, haploBIdx + state.corrB, pos)];
      if(state.numCorr > 1){
	//If A and B share both alleles, do the same for the other shared allele.
	probCorr2 += (*letterFreq)[pos][letter] * 
	  (*errorFunction)[letter][getAllele(*data, haploAIdx + 1 - state.corrA, pos)] * 
	  (*errorFunction)[letter][getAllele(*data, haploBIdx + 1 - state.corrB, pos)];
      }
      else{
	//If the other allele is not shared compute
	//sum(P(c in population) * P(c turned into whatever A has there)) and
	//sum(P(c in population) * P(c turned into whatever B has there))
	probUncorr1 += (*letterFreq)[pos][letter] * 
	  (*errorFunction)[letter][getAllele(*data, haploAIdx + 1 - state.corrA, pos)];
	probUncorr2 += (*letterFreq)[pos][letter] * 
	  (*errorFunction)[letter][getAllele(*data, haploBIdx + 1 - state.corrB, pos)];
      }
    }
    if(state.numCorr > 1){
      if(debug){
	cerr << "In emissionProb: probCorr1 = " << probCorr1 << ", probCorr2 = " << probCorr2 << endl;
      } 
      return correctLog(probCorr1) + correctLog(probCorr2);
    }
    else{
      if(debug){
	cerr << "In emissionProb: probCorr1 = " << probCorr1 << 
	  ", probUncorr1 = " << probUncorr1 << ", probUncorr2 = " << probUncorr2 << endl;
      } 
      return correctLog(probCorr1) + correctLog(probUncorr1) + correctLog(probUncorr2);
    }
  }
  else{
    double probA1 = 0.0, probA2 = 0.0, probB1 = 0.0, probB2 = 0.0;
    for(int letter = 0; letter < (*errorFunction).size(); ++letter){
      probA1 += (*letterFreq)[pos][letter] * 
	(*errorFunction)[letter][getAllele(*data, haploAIdx, pos)];
      probA2 += (*letterFreq)[pos][letter] *
	(*errorFunction)[letter][getAllele(*data, haploAIdx + 1, pos)];
      probB1 += (*letterFreq)[pos][letter] * 
	(*errorFunction)[letter][getAllele(*data, haploBIdx, pos)];
      probB2 += (*letterFreq)[pos][letter] *
	(*errorFunction)[letter][getAllele(*data, haploBIdx + 1, pos)];
    }
    if(debug){
      cerr << "In emissionProb: probA1 = " << probA1 << ", probA2 = " << probA2 <<
	", probB1 = " << probB1 << ", probB2 = " << probB2 << endl;
    } 
    return correctLog(probA1) + correctLog(probA2) + correctLog(probB1) + correctLog(probB2);
  }
} 


