#ifndef RELATIONSHIPMODEL_H
#define RELATIONSHIPMODEL_H

#include <stdio.h>
#include <stdlib.h>
#include "System.h"
#include "Utility.h"
#include "SNPData.h"
#include "RelationshipUtils.h"

#define NUM_VAR_1MRCA 5
#define NUM_VAR_2MRCA 8

#define DONA_IDX 0
#define DONB_IDX 1
#define PHASEA_IDX 2
#define PHASEB_IDX 3
#define SHARED_ANC_IDX 4

#define SHARED0_IDX 4
#define SHARED1_IDX 5
#define PASSEDA_IDX 6
#define PASSEDB_IDX 7

/*
 * "Generic" model for the relationship between two individuals
 * (referred to as individual A and individual B).
 * Subclasses can restrict the types of relationships (e.g. to allow
 * only relationships with 1 MRCA).
 *
 * **** Note ****
 * Instances of this class are NOT owners of the pointer members 
 * (such as data), i.e., they SHOULD NOT delete these pointers.
 * The copy constructor and assigment operators will create SHALLOW
 * copies of the pointers, so many members of the class can point
 * to the same object (e.g. to the same SNPData). 
 */
class RelationshipModel{
 protected:
  //The SNPData object contains all the genotypes and haplotypes of 
  //the modeled population, including the two individuals under consideration.
  SNPData *data;

  //Indices of the two individuals (or diplos) in "data". 
  int idxA, idxB;

  //errorFunction[c][b] is the probability that a sequencing error will turn
  //letter c in the true sequence to letter b.
  vector< vector<double> > *errorFunction;

  //letterFreq[i][c] is the frequency of letter c at position i in the
  //population.
  vector< vector<double> > *letterFreq;

  void init(SNPData &data, int inIdxA, int inIdxB,
	    vector< vector<double> > &inLetterFreq,
	    vector< vector<double> > &inErrorFunction);
  void init(SNPData *data, int inIdxA, int inIdxB,
	    vector< vector<double> > *inLetterFreq,
	    vector< vector<double> > *inErrorFunction);
 public:
  RelationshipModel(){}
  RelationshipModel(SNPData &inData, int inIdxA, int inIdxB,
		    vector< vector<double> > &inLetterFreq,
		    vector< vector<double> > &inErrorFunction);
  RelationshipModel(const RelationshipModel &copy);
  ~RelationshipModel(){}
  RelationshipModel& operator=(const RelationshipModel &that);
};

/* 
 * "Generic" class for modeling two unrelated individuals.
 */
class UnrelatedModel: public RelationshipModel{
 public:
  UnrelatedModel(SNPData &inData, int inIdxA, int inIdxB,
		 vector< vector<double> > &inLetterFreq,
		 vector< vector<double> > &inErrorFunction);
  double logLikelihood();
};

/*
 * Class for the states of HMMs for relationships. 
 * This is a "relatively abstract" class. Although not abstract in the formal
 * sense, it doesn't really specify what type of information a state holds,
 * rather it only contains variables that should be included in any state definition.
 */
class RelationshipHMMState{
 public:
  //Number of correlated alleles, and indicators of which allele of A is correlated
  //with what allele of B. These variables are not really part of the state definition,
  //but can be derived from the variables of the state.
  //corrA and corrB are meaningless when numCorr is 0.
  int numCorr, corrA, corrB;
  
  //Log of the probability that this state is the initial state. Again, this is a 
  //derived feature.
  double logInitialProb;
  
 RelationshipHMMState() : numCorr(-1), corrA(-1), corrB(-1), logInitialProb(correctLog(ZERO)){}
  RelationshipHMMState(const RelationshipHMMState &copy);
  RelationshipHMMState& operator=(const RelationshipHMMState &that);
  ~RelationshipHMMState(){}
};

/*
 * Superclass for "HMM templates" for modelling relationships.
 */
class RelationshipHMM : public RelationshipModel{
 protected:
  //Pascal's triangle used for computations of transition probabilities.
  //**** Note ****
  //logRecomb, logNoRecomb, and pascalTri are not owned by objects of this class
  //so they should NOT be deleted by objects of this class.
  vector< vector<int> > *pascalTri;

  //logRecomb[i][j] is the log probability that there are j independent
  //recombinations between positions i and i + 1, that is, 
  //log(P(recomb between i and i+1)^j).
  //logNoRecomb[i][j] is the log probability of j no-recombinations between positions
  //i and i + 1.
  //logRecomb[i][0] and logNoRecomb[i][0] are 0.0. 
  vector< vector<double> > *logRecomb, *logNoRecomb;

  //numGensA/B are the number of generations between the MRCA(s) and 
  //individuals A and B respectively.
  int numGensA, numGensB;

  void computeRecomb(int maxGens);
 public:
  RelationshipHMM(){}
  RelationshipHMM(SNPData &inData, int inIdxA, int inIdxB,
		  vector< vector<double> > &inLetterFreq, 
		  vector< vector<double> > &inErrorFunction,
		  vector< vector<int> > &inPascalTri,
		  vector< vector<double> > &inLogRecomb,
		  vector< vector<double> > &inLogNoRecomb,
		  int inNumGensA, int inNumGensB);
  RelationshipHMM(const RelationshipHMM &copy);
  ~RelationshipHMM();
  RelationshipHMM& operator=(const RelationshipHMM &that);

  int getNumGensA() const {return numGensA;}
  int getNumGensB() const {return numGensB;}
  vector< vector<double> >& getLogRecomb(){return *logRecomb;}
  vector< vector<double> >& getLogNoRecomb(){return *logNoRecomb;}

  double donationsTransitionProb(bool currentDonations, bool nextDonations, 
				 int pos, int numGens, bool debug = false);  
  void donationsTransProbs(int numGens, int pos, vector< vector<double> > &probs);
  void switchProbs(int diplo, int pos, vector< vector< double> > &probs);
  double emissionProb(const RelationshipHMMState &state, int pos, bool debug = false);
  virtual void clearObj(){
    /*
    pascalTri.clear();
    pascalTri.resize(0);
    logRecomb.clear();
    logRecomb.resize(0);
    logNoRecomb.clear();
    logNoRecomb.resize(0);
    cerr << "In RelationshipHMM::clearObj: Done!" << endl;
    */
  }
};

/*
 * Class for the states of the 1-MRCA HMMs (which include the case where
 * one of the two individuals is an ancestor of the other). 
 * A state encapsulates information about the alleles passed down from the 
 * MRCA to each of A and B at a specific position (or the alleles passed
 * down from individual A to individual B if A is an ancestor of B). 
 */
class OneMRCAState : public RelationshipHMMState{
 public:
  //sharedAncestral is 1 if the immediate descendants of the MRCA got the same allele
  //from the MRCA and 0 otherwise. 
  //BUT, if one of A or B is the MRCA, then sharedAncestral shows which allele of this
  //individual was passed to his/her descendant.
  bool sharedAncestral;
  
  //phaseA and phaseB model phasing (although this is not very accurate, read below). 
  //If phaseA is 0, then the 0-th haplotype of A is the one that comes from the
  //ancestor of A "within" the lineage from the MRCA, and the 1st haplotype of A
  //is the one that comes from "outside." 
  //For example, if phaseA is 0 and phaseB is 1, then A's 0-th haplotype may 
  //be IBD with B's 1st haplotype.
  //phaseA (and phaseB) can change from 0 to 1 and vice versa between
  //consecutive positions, if there's a phasing switch. So, in some sense
  //phaseA models the phasing of A.
  //Notice that if A is the ancestor of B, then sharedAncestral tells us
  //which allele of A is the one that might be IBD with B. So, in this case
  //phaseA will be 0 if there was no phasing error at this position, and 1
  //otherwise.
  bool phaseA, phaseB;

  //donationsA is true if there has been non-lineage donations in the lineage of A
  //after the immediate descendants of the MRCA. Similarly for donationsB.
  //For small pedigrees, these variables may be redundant.
  bool donationsA, donationsB;

  OneMRCAState(bool inSharedAncestral = false, 
	       bool inPhaseA = false, bool inPhaseB = false, 
	       bool inDonationsA = false, bool inDonationsB = false);
  ~OneMRCAState(){}
  void print(ostream &os = cout) const;
};

/*
 * "HMM template" for modelling relationships between two individuals 
 * A and B, where:
 * - A is an ancestor of B or vice versa, or
 * - A and B have exactly one MRCA.
 * If A is the ancestor of B, then numGensA will be -1. If B is the 
 * ancestor of A, then numGensB will be -1. Otherwise, both numGensA,
 * and numGensB will be > 0. At most one of numGensA,
 * numGensB can be -1. 
 */
class OneMRCAModel: public RelationshipHMM{
 public:
  OneMRCAModel(SNPData &inData, int inIdxA, int inIdxB, 
	       vector< vector<double> > &inLetterFreq, 
	       vector< vector<double> > &inErrorFunction,
	       vector< vector<int> > &inPascalTri,
	       vector< vector<double> > &inLogRecomb,
	       vector< vector<double> > &inLogNoRecomb, 
	       int inNumGensA, int inNumGensB);
  OneMRCAModel(const OneMRCAModel &copy);
  ~OneMRCAModel();
  OneMRCAModel& operator=(const OneMRCAModel &that);
  
  const vector<OneMRCAState>& getStates() const {return states;}
  void static validateNumGens(int numGensA, int numGensB){
    AssertMsg(numGensA >= -1 && numGensB >= -1 && numGensA + numGensB >= -1,
	      "Invalid number of generations for the 1-MRCA model.");
  }

  void print(ostream &os = cout);
  double transitionProb(const OneMRCAState& currentState, 
			const OneMRCAState& nextState, 
			int pos, bool debug = false);
  void ancestralTransProbs(int pos, vector< vector<double> > &probs);
  double forward();  
  double forward(vector< vector<double> > &forwardProbs);
  void backward(vector< vector<double> > &backwardProbs);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb, vector<double> &posteriorLogProb);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb,
			  vector< vector<double> > &posteriorLogProb);
  void simpleDiscretePosteriorIBD(vector< vector<double> > &forwardProbs, 
				  vector< vector<double> > &backwardProbs,
				  double seqProb, double nonIBDcutoff, double IBDcutoff, 
				  vector< vector< double> > &posteriorLogProb);
  vector< pair<Segment, int> > posteriorIBDblocks(vector<double> posteriorIBDLogProbs);
  void computePostLogProb(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  vector< vector<double> > &posteriorLogProb);
  
  virtual void clearObj(){
    states.clear();
    states.resize(0);
    RelationshipHMM::clearObj();
  }

 private:
  vector<OneMRCAState> states; //Set of states of the model.

  void initStates();
  void computeCorrelation(OneMRCAState &state);
  void computeLogInitialProb(OneMRCAState &state);
  void computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs);
};

/*
 * Class for the states of the 1-MRCA HMMs (which include the case where
 * one of the two individuals is an ancestor of the other). 
 * A state encapsulates information about the alleles passed down from the 
 * MRCA to each of A and B at a specific position (or the alleles passed
 * down from individual A to individual B if A is an ancestor of B). 
 */
class OneMRCAState2 : public RelationshipHMMState{
 public:
  //sharedAncestral is 1 if the immediate descendants of the MRCA got the same allele
  //from the MRCA and 0 otherwise. 
  //BUT, if one of A or B is the MRCA, then sharedAncestral shows which allele of this
  //individual was passed to his/her descendant.
  bool sharedAncestral;
  
  //phaseA and phaseB model phasing (although this is not very accurate, read below). 
  //If phaseA is 0, then the 0-th haplotype of A is the one that comes from the
  //ancestor of A "within" the lineage from the MRCA, and the 1st haplotype of A
  //is the one that comes from "outside." 
  //For example, if phaseA is 0 and phaseB is 1, then A's 0-th haplotype may 
  //be IBD with B's 1st haplotype.
  //phaseA (and phaseB) can change from 0 to 1 and vice versa between
  //consecutive positions, if there's a phasing switch. So, in some sense
  //phaseA models the phasing of A.
  //Notice that if A is the ancestor of B, then sharedAncestral tells us
  //which allele of A is the one that might be IBD with B. So, in this case
  //phaseA will be 0 if there was no phasing error at this position, and 1
  //otherwise.
  bool phaseA, phaseB;

  //donationsA is true if there has been non-lineage donations in the lineage of A
  //after the immediate descendants of the MRCA. Similarly for donationsB.
  //For small pedigrees, these variables may be redundant.
  bool donationsA, donationsB;

  int idx;

  vector<bool*> variables;

  OneMRCAState2(bool inSharedAncestral = false, 
		bool inPhaseA = false, bool inPhaseB = false, 
		bool inDonationsA = false, bool inDonationsB = false);
  OneMRCAState2(const OneMRCAState2 &copy);
  OneMRCAState2& operator=(const OneMRCAState2 &that);
  ~OneMRCAState2();
  void clearObj(){
    vector<bool*>().swap(variables);
  }
  void updateVariables();
  string getBinaryString();
  void print(ostream &os = cout) const;
};

/*
 * "HMM template" for modelling relationships between two individuals 
 * A and B, where:
 * - A is an ancestor of B or vice versa, or
 * - A and B have exactly one MRCA.
 * If A is the ancestor of B, then numGensA will be -1. If B is the 
 * ancestor of A, then numGensB will be -1. Otherwise, both numGensA,
 * and numGensB will be > 0. At most one of numGensA,
 * numGensB can be -1. 
 */
class OneMRCAModel2: public RelationshipHMM{
 public:
  OneMRCAModel2(SNPData &inData, int inIdxA, int inIdxB, 
		vector< vector<double> > &inLetterFreq, 
		vector< vector<double> > &inErrorFunction,
		vector< vector<int> > &inPascalTri,
		vector< vector<double> > &inLogRecomb,
		vector< vector<double> > &inLogNoRecomb, 
		int inNumGensA, int inNumGensB);
  OneMRCAModel2(const OneMRCAModel2 &copy);
  ~OneMRCAModel2();
  OneMRCAModel2& operator=(const OneMRCAModel2 &that);
  
  const vector<OneMRCAState2>& getStates() const {return states;}
  void getBinaryStates(vector<string> &binStr);
  void static validateNumGens(int numGensA, int numGensB){
    AssertMsg(numGensA >= -1 && numGensB >= -1 && numGensA + numGensB >= -1,
	      "Invalid number of generations for the 1-MRCA model.");
  }

  void print(ostream &os = cout);
  int getNumStates();
  double transitionProb(const OneMRCAState2& currentState, 
			const OneMRCAState2& nextState, 
			int pos, bool debug = false);
  void ancestralTransProbs(int pos, vector< vector<double> > &probs);
  double forward();  
  double forward(vector< vector<double> > &forwardProbs);
  double factorialForward(vector< vector<double> > &forwardProbs);
  double backward(vector< vector<double> > &backwardProbs);
  double factorialBackward(vector< vector<double> > &backwardProbs);
  double viterbi(vector<int> &stateIdx);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb, vector<double> &posteriorLogProb);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb,
			  vector< vector<double> > &posteriorLogProb);
  void simpleDiscretePosteriorIBD(vector< vector<double> > &forwardProbs, 
				  vector< vector<double> > &backwardProbs,
				  double seqProb, double nonIBDcutoff, double IBDcutoff, 
				  vector< vector< double> > &posteriorLogProb);
  vector< pair<Segment, int> > posteriorIBDblocks(vector<double> posteriorIBDLogProbs);
  void computePostLogProb(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  vector< vector<double> > &posteriorLogProb);
  
  virtual void clearObj(){
    vector<int>().swap(maxValues);
    vector<OneMRCAState2>().swap(states);
    //maxValues.clear();
    //maxValues.resize(0);
    //states.clear();
    //states.resize(0);
    RelationshipHMM::clearObj();
  }

 private:
  vector<int> maxValues;
  vector<OneMRCAState2> states; //Set of states of the model.

  void initStates();
  void computeCorrelation(OneMRCAState2 &state);
  void computeLogInitialProb(OneMRCAState2 &state);
  int getStateIdx(const OneMRCAState2 &state);
  //int getStateIdx(bool sharedAncestral, bool phaseA,
  //	  bool phaseB, bool donA, bool donB);
  int getInverseStateIdx(const OneMRCAState2& state, int var);
  void computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs);
};

/*
 * "HMM template" for modelling relationships between two individuals 
 * A and B, where:
 * - A is an ancestor of B or vice versa, or
 * - A and B have exactly one MRCA.
 * If A is the ancestor of B, then numGensA will be -1. If B is the 
 * ancestor of A, then numGensB will be -1. Otherwise, both numGensA,
 * and numGensB will be > 0. At most one of numGensA,
 * numGensB can be -1. 
 */
class OneMRCAModel2Phased: public RelationshipHMM{
 public:
  OneMRCAModel2Phased(SNPData &inData, int inIdxA, int inIdxB, 
		      vector< vector<double> > &inLetterFreq, 
		      vector< vector<double> > &inErrorFunction,
		      vector< vector<int> > &inPascalTri,
		      vector< vector<double> > &inLogRecomb,
		      vector< vector<double> > &inLogNoRecomb, 
		      int inNumGensA, int inNumGensB,
		      bool inTrueCorrA, bool inTrueCorrB);
  OneMRCAModel2Phased(const OneMRCAModel2Phased &copy);
  ~OneMRCAModel2Phased();
  OneMRCAModel2Phased& operator=(const OneMRCAModel2Phased &that);
  
  const vector<OneMRCAState2>& getStates() const {return states;}
  void getBinaryStates(vector<string> &binStr);
  void static validateNumGens(int numGensA, int numGensB){
    AssertMsg(numGensA >= -1 && numGensB >= -1 && numGensA + numGensB >= -1,
	      "Invalid number of generations for the 1-MRCA model.");
  }

  void print(ostream &os = cout);
  int getNumStates();
  double transitionProb(const OneMRCAState2& currentState, 
			const OneMRCAState2& nextState, 
			int pos, bool debug = false);
  void ancestralTransProbs(int pos, vector< vector<double> > &probs);
  double forward();  
  double forward(vector< vector<double> > &forwardProbs);
  double factorialForward(vector< vector<double> > &forwardProbs);
  double backward(vector< vector<double> > &backwardProbs);
  double factorialBackward(vector< vector<double> > &backwardProbs);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb, vector<double> &posteriorLogProb);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb,
			  vector< vector<double> > &posteriorLogProb);
  
  virtual void clearObj(){
    vector<int>().swap(maxValues);
    vector<OneMRCAState2>().swap(states);
    //maxValues.clear();
    //maxValues.resize(0);
    //states.clear();
    //states.resize(0);
    RelationshipHMM::clearObj();
  }

 private:
  vector<int> maxValues;
  vector<OneMRCAState2> states; //Set of states of the model.

  bool trueCorrA, trueCorrB;

  void initStates();
  void computeCorrelation(OneMRCAState2 &state);
  void computeLogInitialProb(OneMRCAState2 &state);
  int getStateIdx(const OneMRCAState2 &state);
  //int getStateIdx(bool sharedAncestral, bool phaseA,
  //	  bool phaseB, bool donA, bool donB);
  int getInverseStateIdx(const OneMRCAState2& state, int var);
  void computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs);
};


/*
 * Class for the states of the 2-MRCAs HMMs.
 * A state encapsulates information about the alleles passed down from the 
 * MRCAs to each of A and B at a specific position. 
 * In the following comments, let MRCA0 and MRCA1 be the two MRCAs, 
 * and C and D be the immediate descendants of the MRCAs in the 
 * lineage of A and B respectively.
 */
class TwoMRCAState : public RelationshipHMMState{
 public:
  //sharedMRCA0 is true if the immediate descendants of the MRCAs
  //got the same allele from MRCA0 (i.e. they both got the maternal
  //allele of MRCA0, or they both got the paternal allele of MRCA0).
  //Same for sharedMRCA1 and MRCA1.
  bool sharedMRCA0, sharedMRCA1;

  //passedA is 0 (false) if allele 0 of C (i.e. the allele C got from 
  //MRCA0) was passed to his/her descendant, and 1 (true) if allele 1 of 
  //C was passed to C's descendant.
  //Similarly for passedB and D.
  bool passedA, passedB;

  //donationsA will be true if there were non-lineage donations in
  //the lineage of A AFTER the descendant of C (since C is 
  //considered by sharedMRCA0/1 and C's descendant is considered by
  //passedA).
  //Similarly for donationsB.
  bool donationsA, donationsB;

  //phaseA, phaseB show which allele of A is IBD with which allele of B
  //(if some alleles are IBD).
  bool phaseA, phaseB;

  int idx;

  vector<bool*> variables;

  TwoMRCAState(bool inSharedMRCA0 = false, bool inSharedMRCA1 = false, 
	       bool inPassedA = false, bool inPassedB = false,
	       bool inDonationsA = false, bool inDonationsB = false,
	       bool inPhaseA = false, bool inPhaseB = false);
  TwoMRCAState(const TwoMRCAState &copy); 
  TwoMRCAState& operator=(const TwoMRCAState &that);
  ~TwoMRCAState();
  void clearObj(){
    vector<bool*>().swap(variables);
  }
  void updateVariables();
  string getBinaryString();
  void print(ostream &os = cout) const;
};

/*
 * "HMM template" for modelling relationships between two individuals 
 * A and B, such that A and B have 2 MRCAs.
 */
class TwoMRCAModel : public RelationshipHMM{
 public:
  TwoMRCAModel(SNPData &inData, int inIdxA, int inIdxB, 
	       vector< vector<double> > &inLetterFreq, 
	       vector< vector<double> > &inErrorFunction, 
	       vector< vector<int> > &inPascalTri,
	       vector< vector<double> > &inLogRecomb,
	       vector< vector<double> > &inLogNoRecomb,
	       int numGensA, int numGensB);
  TwoMRCAModel(const TwoMRCAModel &copy);
  ~TwoMRCAModel();
  TwoMRCAModel& operator=(const TwoMRCAModel &that);

  const vector<TwoMRCAState>& getStates() const {return states;}

  void print(ostream &os = cout);
  void getBinaryStates(vector<string> &binStr);
  double transitionProb(const TwoMRCAState& currentState, 
			const TwoMRCAState& nextState, 
			int pos, bool debug = false);
  void sharedMRCAprobs(int pos, vector< vector<double> > &probs);
  void passedProbs(int numGens, int pos, vector< vector<double> > &probs);
  double forward();
  double forward(vector< vector<double> > &forwardProbs);
  double factorialForward(vector< vector<double> > &forwardProbs);
  double backward(vector< vector<double> > &backwardProbs);
  double viterbi(vector<int> &stateIdx);
  double factorialBackward(vector< vector<double> > &backwardProbs);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb,
			  vector< vector<double> > &posteriorLogProb);
  void simplePosteriorIBDAll(vector< vector<double> > &forwardProbs, 
			     vector< vector<double> > &backwardProbs,
			     double seqProb,
			     vector< vector<double> > &posteriorLogProb);
  void simpleDiscretePosteriorIBD(vector< vector<double> > &forwardProbs, 
				  vector< vector<double> > &backwardProbs,
				  double seqProb, double nonIBDcutoff, double IBDcutoff, 
				  vector< vector< double> > &posteriorLogProb);
  void computePostLogProb(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  vector< vector<double> > &posteriorLogProb);
  
  virtual void clearObj(){
    vector<int>().swap(maxValues);
    vector<TwoMRCAState>().swap(states);
    RelationshipHMM::clearObj();
  }

 private:
  vector<int> maxValues;
  vector<TwoMRCAState> states; //Set of states of the model.

  int getStateIdx(const TwoMRCAState &state);
  int getInverseStateIdx(const TwoMRCAState& state, int var);
  int getNumStates();
  void initStates();
  void validateNumGens(int numGensA, int numGensB){
    AssertMsg(numGensA >= 0 && numGensB >= 0,
	      "Invalid number of generations for the 2-MRCA model.");
  }
  void computeCorrelation(TwoMRCAState &state);
  void validateState(const TwoMRCAState &state) const;
  void computeLogInitialProb(TwoMRCAState &state);
  bool isTransitionPossible(const TwoMRCAState& currentState, 
			    const TwoMRCAState& nextState, int pos);
  void computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs);
};


/*
 * "HMM template" for modelling relationships between two individuals 
 * A and B, such that A and B have 2 MRCAs.
 */
class TwoMRCAModelPhased : public RelationshipHMM{
 public:
  TwoMRCAModelPhased(SNPData &inData, int inIdxA, int inIdxB, 
		     vector< vector<double> > &inLetterFreq, 
		     vector< vector<double> > &inErrorFunction, 
		     vector< vector<int> > &inPascalTri,
		     vector< vector<double> > &inLogRecomb,
		     vector< vector<double> > &inLogNoRecomb,
		     int numGensA, int numGensB,
		     bool inTrueCorrA, bool inTrueCorrB);
  TwoMRCAModelPhased(const TwoMRCAModelPhased &copy);
  ~TwoMRCAModelPhased();
  TwoMRCAModelPhased& operator=(const TwoMRCAModelPhased &that);

  const vector<TwoMRCAState>& getStates() const {return states;}

  void print(ostream &os = cout);
  void getBinaryStates(vector<string> &binStr);
  double transitionProb(const TwoMRCAState& currentState, 
			const TwoMRCAState& nextState, 
			int pos, bool debug = false);
  void sharedMRCAprobs(int pos, vector< vector<double> > &probs);
  void passedProbs(int numGens, int pos, vector< vector<double> > &probs);
  double forward();
  double forward(vector< vector<double> > &forwardProbs);
  double factorialForward(vector< vector<double> > &forwardProbs);
  double backward(vector< vector<double> > &backwardProbs);
  double factorialBackward(vector< vector<double> > &backwardProbs);
  void simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  double seqProb,
			  vector< vector<double> > &posteriorLogProb);
  void simpleDiscretePosteriorIBD(vector< vector<double> > &forwardProbs, 
				  vector< vector<double> > &backwardProbs,
				  double seqProb, double nonIBDcutoff, double IBDcutoff, 
				  vector< vector< double> > &posteriorLogProb);
  void computePostLogProb(vector< vector<double> > &forwardProbs, 
			  vector< vector<double> > &backwardProbs,
			  vector< vector<double> > &posteriorLogProb);
  
  virtual void clearObj(){
    vector<int>().swap(maxValues);
    vector<TwoMRCAState>().swap(states);
    RelationshipHMM::clearObj();
  }

 private:
  vector<int> maxValues;
  vector<TwoMRCAState> states; //Set of states of the model.
  bool trueCorrA, trueCorrB;

  int getStateIdx(const TwoMRCAState &state);
  int getInverseStateIdx(const TwoMRCAState& state, int var);
  int getNumStates();
  void initStates();
  void validateNumGens(int numGensA, int numGensB){
    AssertMsg(numGensA >= 0 && numGensB >= 0,
	      "Invalid number of generations for the 2-MRCA model.");
  }
  void computeCorrelation(TwoMRCAState &state);
  void validateState(const TwoMRCAState &state) const;
  void computeLogInitialProb(TwoMRCAState &state);
  bool isTransitionPossible(const TwoMRCAState& currentState, 
			    const TwoMRCAState& nextState, int pos);
  void computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs);
};

/*
 * State of simple RelationshipHMMs that only encode the identity
 * by descent of each locus (such an HMM is not really an HMM, since 
 * it violates the Markov property in some cases, but whatever).
 */
/*
class IBDState : public RelationshipHMMState{
 public:
  //Whether each of the haplotypes is IBD.
  bool ibd0, ibd1;
  
  //phaseA and phaseB model phasing (although this is not very accurate, read below). 
  //If phaseA is 0, then the 0-th haplotype of A is the one that comes from the
  //ancestor of A "within" the lineage from the MRCA, and the 1st haplotype of A
  //is the one that comes from "outside." 
  //For example, if phaseA is 0 and phaseB is 1, then A's 0-th haplotype may 
  //be IBD with B's 1st haplotype.
  //phaseA (and phaseB) can change from 0 to 1 and vice versa between
  //consecutive positions, if there's a phasing switch. So, in some sense
  //phaseA models the phasing of A.
  //Notice that if A is the ancestor of B, then sharedAncestral tells us
  //which allele of A is the one that might be IBD with B. So, in this case
  //phaseA will be 0 if there was no phasing error at this position, and 1
  //otherwise.
  bool phaseA, phaseB;

  IBDState(bool inIbd0 = false, bool inIbd1 = false, 
	   bool inPhaseA = false, bool inPhaseB = false);
  ~IBDState(){}
  void print(ostream &os = cout) const;
};
*/
#endif
