#include "RelationshipUtils.h"
#include <sstream>
#include <math.h>
#include <time.h>

/*
Creates two individuals with either 1 or 2 most recent common ancestors.
Inputs:
- numMRCAs: number of most recent common ancestors (MRCAs). Valid values are 1 and 2.
- genA: number of generations that have passed between the MRCAs and 
the first individual to be created (also called conventionally individual A, or just A).
- genB: number of generations that have passed between the MRCAs and the second  
individual to be created (also called conventionally individual B, or just B).
Eg. If there are 2 MRCAs, and A is their child, then genA is 0. If A is their grandchild, 
then genA is 1. So, if A and B are siblings, then genA and genB are 0.
If A and B are half-siblings, then numMRCAs is 1 and genA and genB are 0. By convention,
if A is the parent of B, then numMRCAs is 1, genA is -1, and genB is 0.
Therefore, if numMRCAs is 2, genA and genB must be at least 0. If numMRCAs is 1, then
at most 1 of genA and genB can be -1, and the other must be at least 0.
- indivA and indivB: references to the new individuals
-maxIndiv: The code will only consider individuals in profile whose index is 1 to maxIndiv-1.
Return value:
A vector of length 2 with the indices of A and B in the SNPData object.
*/
vector<int> createRelatives(SNPProfile &profile, SNPData &data, 
							Simulate::Diplo &indivA, Simulate::Diplo &indivB,
							int numMRCAs, int genA, int genB, int maxIndiv, bool verbose, bool addToData){
	Assert((numMRCAs == 1 && genA >= -1 && genB >= -1 && genA + genB >= -1)
		   || (numMRCAs == 2 && genA >= 0 && genB >= 0));
	int numInitialIndiv = data.num_diplos;
	//We need at least 2 individuals to do any mating.
	Assert(maxIndiv > 1); Assert(numInitialIndiv >= maxIndiv); 
	double neededIndiv = 2*(pow(2,genA)) + 2*(pow(2,genB)) - numMRCAs; // This is the number of individuals we need to create the tree if there is no inbreeding.
	if(verbose)
	{
		cerr << "neededIndiv:" << neededIndiv << endl;
	}
	vector<int> returnIndices(2, 0);
	
	
	AssertMsg(neededIndiv<=maxIndiv, "Not enough individuals to avoid inbreeding");  //Replace it later with a more complex code which implements some inbreedings to create enough individuals
	//Generate a random permutation of the numbers 0, 1,...,maxIndiv-1,
	//which will be used to get a permutation of the individuals in the 
	//population.
	vector<int> perm;
	for(int i = 0; i < maxIndiv; i++){
		perm.push_back(i);
	}
	random_shuffle(perm.begin(), perm.end());
	
	
	int idx = 0; //This will be the index in the permutation vector.
	int mate0Idx = perm.at(idx++), mate1Idx = perm.at(idx++);
	Simulate::Diplo mate0(data, mate0Idx); 
	Simulate::Diplo mate1(data, mate1Idx); 
	Simulate::Diplo MRCA1 = mate0;
	Simulate::Diplo MRCA2 = mate1; // if numMRCAs=1 then MRCA2 will not be used. 
	Simulate::Diplo child;
	//These will be the indices of individuals A and B in the SNPData object.
	
	
	if(verbose)
	{
		cerr << "Generating lineage of A..." << endl;
	}
	//If genA == -1, then A is an ancestor of B, and A is the first individual pointed by perm.
	if(genA == -1){
		returnIndices[0] = MRCA1.idx;
		indivA = MRCA1;
	}
	
	else{
		vector<Simulate::Diplo> cLevelIndvs; //Individuals who are going to mate in this level
		cLevelIndvs.push_back(mate0); cLevelIndvs.push_back(mate1);
		for(idx = 2; idx < 2*(pow(2,genA)); idx++){
			cLevelIndvs.push_back(Simulate::Diplo(data, perm.at(idx)));
		} 
		for(int i = 0; i < genA + 1; i++){
			vector<Simulate::Diplo> children; // children resulting from mating cLevelIndvs individuals
			int j=0;
			while(j< cLevelIndvs.size()-1)
			{
				mate0 = cLevelIndvs.at(j);
				mate1 = cLevelIndvs.at(++j);
				//string mate0Label = data.diplos[mate0.idx].label;
				//string mate1Label = data.diplos[mate1.idx].label;
				//string childLabel = getChildLabel(mate0Label, mate1Label, data.num_diplos - numInitialIndiv + 1);
				
				if(i==genA && j==(cLevelIndvs.size()-1))
				{
					string childLabel = "indivA"; 
					mate(profile, data, mate0, mate1,child, childLabel, verbose, addToData);
				}
				else
				{
					string childLabel = "temp";
					mate(profile, data, mate0, mate1,child, childLabel, verbose, false);
					
				}
				children.push_back(child);
				j++;
			}
			random_shuffle(children.begin(), children.end());
			cLevelIndvs = children;
		}
		//A is the last child generated by the previous loop.
		if(addToData)
		  returnIndices[0] = data.num_diplos - 1;
		else
		  returnIndices[0]=-1;
		indivA = child;
	}
	//Reset mate0 to point to the first MRCA.
	mate0 = MRCA1;
	
	//If there is one MRCA, this will be mate0, and mate1 will be a new
	//random individual. If genA == -1, we don't need to get a new random
	//individual, since we never used mate1.
	if(numMRCAs == 1 && genA > -1){
		mate1Idx = perm.at(idx++);
		mate1 = Simulate::Diplo(data, mate1Idx);
	}
	else{
		//If there are 2 MRCAs, we must reset mate1 to point to the second MRCA.
		mate1 = MRCA2;
	}
	if(verbose)
	{
		cerr << "Generating lineage of B..." << endl;
	}
	if(genB == -1){
		returnIndices[1] = MRCA1.idx;
		indivB = MRCA1;
	}
	else{
		vector<Simulate::Diplo> cLevelIndvs; //Individuals who are going to mate in this level
		cLevelIndvs.push_back(mate0); cLevelIndvs.push_back(mate1);
		for(int i = 0; i < 2*(pow(2,genB))-2; i++){
			idx++;
			cLevelIndvs.push_back(Simulate::Diplo(data, perm.at(idx)));
		} 
		for(int i = 0; i < genB + 1; i++){
			vector<Simulate::Diplo> children; // children resulting from mating cLevelIndvs individuals
			int j=0;
			while(j< cLevelIndvs.size()-1)
			{
				mate0 = cLevelIndvs.at(j);
				mate1 = cLevelIndvs.at(++j);
				//string mate0Label = data.diplos[mate0.idx].label;
				//string mate1Label = data.diplos[mate1.idx].label;
				//string childLabel = getChildLabel(mate0Label, mate1Label, data.num_diplos - numInitialIndiv + 1);
				if(i==genB && j==(cLevelIndvs.size()-1))
				{
					string childLabel = "indivB";
					mate(profile, data, mate0, mate1,child, childLabel, verbose, addToData);
				}
				
				else
				{
					string childLabel = "temp";
					mate(profile, data, mate0, mate1,child, childLabel, verbose, false);
					
				}
				
				children.push_back(child);
				j++;
			}
			random_shuffle(children.begin(), children.end());
			cLevelIndvs = children;
		}
		//B is the last child generated by the previous loop.
		if(addToData)
		  returnIndices[1] = data.num_diplos - 1;
		else
		  returnIndices[1]=-1;
		indivB = child;		
		
	}
		//cerr << "final value of idx:" << idx << endl;
	return returnIndices; 
	
	
	
	
	/*vector < Simulate::Diplo > males, females;
	 for(int i = 0; i < numInitialIndividuals; i++){
	 Simulate::Diplo individual(data, i);
	 if(individual.gender == 0) females.push_back(individual);
	 else males.push_back(individual);
	 }
	 vector<int> malesShuffle, femalesShuffle;
	 vector<int>::iterator malesIt, femalesIt;
	 for(int i = 0; i < males.size(); i++) malesShuffle.push_back(i);
	 for(int i = 0; i < females.size(); i++) femalesShuffle.push_back(i);
	 random_shuffle(malesShuffle.begin(), malesShuffle.end());
	 random_shuffle(femalesShuffle.begin(), femalesShuffle.end());
	 malesIt = malesShuffle.begin();
	 femalesIt = femalesShuffle.begin();
	 Simulate::Diplo mrca0 = females[*femalesIt];
	 ++femalesIt;
	 Simulate::Diplo mrca1 = males[*malesIt];
	 ++malesIt;*/
}
/*
Creates two unrelated individuals.
Inputs:
- indivA and indivB: references to the new individuals
- maxIndiv: The code will only consider individuals in profile whose index is 1 to maxIndiv-1.
Return value:
A vector of length 2 with the indices of A and B in the SNPData object.
*/
vector<int> createUnrelated(SNPProfile &profile, SNPData &data, 
			    Simulate::Diplo &indivA, Simulate::Diplo &indivB,int maxIndiv, bool verbose){
			    
	int numIndiv = data.num_diplos;
	//We need at least 2 individuals.
	Assert(maxIndiv > 1); 
	Assert(maxIndiv<=numIndiv);
	srand(time(0));
	vector<int> returnIndices(2, 0);
	returnIndices[0] = rand()%maxIndiv;
	bool terminate = false;
	while(!terminate)
	{
	 returnIndices[1] = rand()%maxIndiv;
	 if(returnIndices[1] != returnIndices[0])
	 	terminate = true;
	}
	indivA= Simulate::Diplo(data, returnIndices[0]); 
	indivB= Simulate::Diplo(data, returnIndices[1]); 
	return returnIndices;	
}
/*
 * DOES NOT WORK. THROW AWAY?
 * Return value:
 * The Simulate::Diplo object corresponding to the SNPData::Diplo
 * object at the idx-th position of data.
 */
Simulate::Diplo getSimulateDiploAt(SNPData &data, int idx){
  Assert(idx >= 0 && idx < data.num_diplos);
  Simulate::Diplo diplo(data, idx);
  //The above call sets the haplos in the Diplo object to a set of 
  //brand new (un-recombined) haplotypes. To set the haplotypes of 
  //diplo correctly, we must grab the corresponding SNPData::Haplo
  //object and scan it to create the segments of the Simulate::Haplo 
  //object.
  int haplo0Idx = data.diplos[idx].haplos[0];
  int haplo1Idx = data.diplos[idx].haplos[1];
  setSimulateHaploSources(data.haplos[haplo0Idx].data, diplo.haplos[0]);
  setSimulateHaploSources(data.haplos[haplo1Idx].data, diplo.haplos[1]);
  return diplo;
}

/*
 * DOES NOT WORK. THROW AWAY?
 */
void setSimulateHaploSources(const vector<int8> haploData, Simulate::Haplo &haplo){
  haplo.sources.clear();
  if(haploData.empty()) return;
  int8 lastSource = haploData[0];
  int lastStart = 0;
  for(int i = 0; i < haploData.size(); ++i){
    if(haploData[i] != lastSource){
      haplo.sources.push_back(pair<Segment, int>(Segment(lastStart, i, true), lastSource));
      lastStart = i;
    }
    lastSource = haploData[i];
  }
  haplo.sources.push_back(pair<Segment, int>(Segment(lastStart, 
						     haploData.size(), 
						     true), lastSource));
}
  
/*
 * Returns a label for the child of the individuals at positions mate0Idx
 * and mate1Idx of data. The label has the form:
 * child<childNum>:mate0Label-mate1Label.
 */
string getChildLabel(string mate0Label, string mate1Label, int childNum){
	if(mate0Label.substr(0, 5).compare("child") == 0){
		int end = mate0Label.find(':');
		mate0Label = mate0Label.substr(0, end);
	}
	if(mate1Label.substr(0, 5).compare("child") == 0){
		int end = mate1Label.find(':');
		mate1Label = mate1Label.substr(0, end);
	}
	stringstream ss;
	ss << childNum;
	return "child" + ss.str() + ":" + mate0Label + "-" + mate1Label;
}

/*
 Prints the given segment.
*/
void printSegment(pair<Segment, int> &s){
  cerr << s.first.left() << "\t" << s.first.right() << "\t" << s.second << endl;
}

/*
 * For each haplotype of an individual, it prints the segments it consists of.
 * A segment is a part of the chromosome. It has a starting and an ending SNP
 * (i.e. endpoints are not given in terms of base-pairs, but in terms of SNPs).
 * It also has a label that basically shows from which haplotype of the original
 * population it comes.
 */
void printDiploSegments(Simulate::Diplo &diplo){
  cerr << "Haplotype 0\n";
  for(int i = 0; i < diplo.haplos[0].sources.size(); i++){
    cerr << "\t";
    printSegment(diplo.haplos[0].sources[i]);
  }
  cerr << "Haplotype 1\n";
  for(int i = 0; i < diplo.haplos[1].sources.size(); i++){
    cerr << "\t";
    printSegment(diplo.haplos[1].sources[i]);
  }
}

void printDiploSegments(SNPData &data, int idx){
  Simulate::Diplo temp(data, idx);
  printDiploSegments(temp);
}

/*
 Creates a child of the individuals of SNPData at the indices mate0Idx
 and mate1Idx. The child is stored in the input argument child.
 */
void mate(SNPProfile &profile, SNPData &data, Simulate::Diplo &mate0,
		  Simulate::Diplo &mate1, Simulate::Diplo &child, string childLabel, bool verbose, bool add){
	int mate0Idx = mate0.idx , mate1Idx = mate1.idx;
	if(verbose){
		cerr << "Mating Parent1:" << endl;
		printDiploSegments(mate0);
		cerr << "and Parent2:" <<  endl;
		printDiploSegments(mate1);
	}
	Simulate::MateDiplos(profile, mate0, mate1, child);
	//The previous call did not generate the haplotypes of child.
	vector < vector < int8 > > childHaplos(2);
	child.Generate(data, childHaplos[0], childHaplos[1]);
	child.pop = mate0.pop;
	if(add){
	  //Add the child to the SNPData object. Children are arbitrarily added to the 
		//population of mate0. The switch error probability is arbitrarily set.
		data.AddDiplo(childHaplos[0], childHaplos[1], 0.0, mate0.pop, childLabel);
		//This call is necessary, because the addition of the new haplos in data 
		//does not mean that the data object members (such as num_diplos) are updated.
		data.ComputeDerived();
		//set the id of child to be the same as its id in data(the newly generated child was added at the end of data)
		child.idx = data.num_diplos - 1;
	}
	if(verbose){
		cerr << "Child " << childLabel << endl;
		printDiploSegments(child);
	}
}

/*
 * Counts the number of SNPs that are IBD between two individuals and stores 
 * them in the array counts. At return, counts[0] will have the number of 
 * SNPs such that the one of the two alleles of the two individuals is IBD. 
 * counts[1] will have the number of SNPs for which both alleles of the 
 * two individuals are IBD (this will be 0 in most cases). 
 * Return value:
 * A vector M such that M[i] indicates how many alleles IBD there are at pos i.
 */

void trueIBDCounts(Simulate::Diplo &diplo1, Simulate::Diplo &diplo2, 
		   int counts[2], vector< vector<int> > &indicators){
  counts[0] = counts[1] = 0;
  indicators.clear();

  Simulate::Haplo haplos[4];
  haplos[0] = diplo1.haplos[0];
  haplos[1] = diplo1.haplos[1];
  haplos[2] = diplo2.haplos[0];
  haplos[3] = diplo2.haplos[1];
  
  //Index in the sources vectors of the Haplos (A "source" is a segment of 
  //consecutive loci that come from the same individual in the population. 
  //The sources are sorted by position). As we scan the SNPs, sourceIndex[j]
  //will point to the current source of haplos[j].
  int sourceIndex[4];
  sourceIndex[0] = sourceIndex[1] = sourceIndex[2] = sourceIndex[3] = 0;
  
  //Find the number of SNPs in each haplo. There should be the same 
  //number in all haplos.
  int numSNPs = -1;
  for(int i = 0; i < 4; ++i){
    Assert(!haplos[i].sources.empty());
    if(numSNPs < 0)
      numSNPs = haplos[i].sources.back().first.right();
    else
      Assert(haplos[i].sources.back().first.right() == numSNPs);
  }
  indicators.resize(numSNPs);
  
  //Iterate through all the positions...
  for(int i = 0; i < numSNPs; ++i){
    indicators[i].resize(3, 0);
    //For each Haplo, check whether we passed the end of the current segment.
    for(int j = 0; j < 4; ++j){
      //haplos[j].sources[sourceIndex[j]].first.right() is the end of the 
      //current segment (more precisely the end of the current segment + 1, 
      //or the beginning of the next segment).
      if(i == haplos[j].sources[sourceIndex[j]].first.right()){
	sourceIndex[j]++;
	Assert(sourceIndex[j] < haplos[j].sources.size());
      }
    }
    //Compare each of the Haplos of the first individual to each of the 
    //haplos of the second individual.
    int IBD = 0;
    for(int j = 0; j < 2; ++j){
      for(int k = 2; k < 4; ++k){
	if(haplos[j].sources[sourceIndex[j]].second ==
	   haplos[k].sources[sourceIndex[k]].second){
	  IBD++;
	  indicators[i][1] = j;
	  indicators[i][2] = k - 2;
	}
      }
    }
    indicators[i][0] = IBD;
    if(IBD > 0) counts[IBD - 1]++;
  }
}


/* For test purposes.
 * Finds the IBD blocks between the haplos of the two given individual. 
 * If there are two adjacent blocks coming from different sources, this method treats them as two blocks instead of one.
 * Return value:
 * A vector of ibd block sizes between haplos.
 * NIT_CausedBy is a vector of size 2. NIT_CausedBy[0] is the number of transitions from NonIBD to IBD which has caused by diplo1 (The left side of the IBD block is the start of a segment in diplo1 but is within a segment for diplo2) NIT_CausedBy[1] is the same for B.
 * INT_CausedBy is the same as NIT_CausedBy but for IBD to Non-IBD transitions. 
 */ 
vector<int> trueIBDBlocks(Simulate::Diplo &diplo1, Simulate::Diplo &diplo2, vector<int> & NIT_CausedBy, vector<int> & INT_CausedBy, int & NIT_tolTransitions, int & INT_tolTransitions){
	Simulate::Haplo haplos[4];
  haplos[0] = diplo1.haplos[0];
  haplos[1] = diplo1.haplos[1];
  haplos[2] = diplo2.haplos[0];
  haplos[3] = diplo2.haplos[1];

  //Index in the sources vectors of the Haplos (A "source" is a segment of 
  //consecutive loci that come from the same individual in the population. 
  //The sources are sorted by position). As we scan the SNPs, sourceIndex[j]
  //will point to the current source of haplos[j].
  int sourceIndex[4];
  sourceIndex[0] = sourceIndex[1] = sourceIndex[2] = sourceIndex[3] = 0;

  //Find the number of SNPs in each haplo. There should be the same 
  //number in all haplos.
  int numSNPs = -1;
  for(int i = 0; i < 4; ++i){
    Assert(!haplos[i].sources.empty());
    if(numSNPs < 0)
      numSNPs = haplos[i].sources.back().first.right();
    else
      Assert(haplos[i].sources.back().first.right() == numSNPs);
  }
  vector<int> ibdBlocks; // a vector of IBD block sizes
  vector< pair<int,int> > ibdBlocksBorders; // a vector of genetic sizes of IBD blocks
  int ibdBlockSizes[2][2]; // ibdBlockSizes[j][k] is the size of current ibd block between haplos[j] and haplos[k], 
  int ibdBlockLeftBorder[2][2];
  bool ibd[2][2]; // ibd[j][k] is true if the previous SNP of haplos[j] and haplos[k] were IBD
  //initialize blockSizes to be 0 and ibds to be false
  for(int i=0; i<2;i++)
    {
      for(int j=0;j<2;j++)
	{
	  ibdBlockSizes[i][j]=0;
	  ibdBlockLeftBorder[i][j]=0;
	  ibd[i][j]=false;
	}
    }
  NIT_tolTransitions=INT_tolTransitions=0;
  NIT_CausedBy.clear();  NIT_CausedBy.resize(2,0);
  INT_CausedBy.resize(2,0);
  //Iterate through all the positions...
  for(int i = 0; i < numSNPs; i++)
    {
      //For each Haplo, check whether we passed the end of the current segment.
      int segmentChanged[4];
      for(int j = 0; j < 4; j++){
	//haplos[j].sources[sourceIndex[j]].first.right() is the end of the 
	//current segment (more precisely the end of the current segment + 1, 
	//or the beginning of the next segment).
	segmentChanged[j]=0;
	if(i == haplos[j].sources[sourceIndex[j]].first.right()){
	  sourceIndex[j]++;
	  segmentChanged[j]=1;
	  Assert(sourceIndex[j] < haplos[j].sources.size());
	}
      }
      int IBD = 0;
      for(int j=0; j<2 ;j++)
	{
	  for(int k=2; k<4; k++)
	    {
	      if(haplos[j].sources[sourceIndex[j]].second == haplos[k].sources[sourceIndex[k]].second)   // If the current SNP is ibd
		{
		  IBD++;
		  // Check for nonIBD-IBD transitions and skip the left border of chromosome
		  if((!ibd[j][k-2])&& (i>0)){
		    NIT_tolTransitions++;
		    if(!(segmentChanged[j] && segmentChanged[k]))
		      {
			if(segmentChanged[j])
			  ++NIT_CausedBy[0];
			if(segmentChanged[k])
			  ++NIT_CausedBy[1];
		      }
		  }
		  if(!ibd[j][k-2]) // This is the left border of IBD block, record the SNP number
		    {
		      ibdBlockLeftBorder[j][k-2]=i;
		    }
		  ibdBlockSizes[j][k-2]++;
		  ibd[j][k-2]=true;
					
		}
	      else {
		if(ibd[j][k-2]) // If the previous SNP belonged to an ibd block then add the size to ibdBlocks vector
		  {
		    INT_tolTransitions++;		    
		    ibdBlocks.push_back(ibdBlockSizes[j][k-2]);
		    ibdBlocksBorders.push_back(pair<int,int>(ibdBlockLeftBorder[j][k-2],i));
		    ibdBlockSizes[j][k-2] = 0;
		    ibd[j][k-2]=false;
		    // This break in IBD block should have caused by a segment change in one of the haplos. Add it to INT_Caused by if
		    // it has not caused by a segment change in both of diplos.
		    if(!(segmentChanged[j] && segmentChanged[k]))
		      {
			if(segmentChanged[j])
			  ++INT_CausedBy[0];
			if(segmentChanged[k])
			  ++INT_CausedBy[1];
		      }
		  }
	      }

	    }
	}	}
  //add the last ibd block sizes to the vector
  for(int j=0; j<2;j++)
    {
      for(int k=0;k<2;k++)
	{
	  if(ibd[j][k])
	    {
	      ibdBlocks.push_back(ibdBlockSizes[j][k]);
	      ibdBlocksBorders.push_back(pair<int,int>(ibdBlockLeftBorder[j][k],numSNPs-1));
	    }
	}
    } 
	
  return ibdBlocks;
}

/*
 * Computes log(a + b) given log(a) and log(b).
 */
double computeLogSumOfTwo(double log1, double log2){
  //Avoid calling exp(large number), which will return inf.
  //Calling exp(small number) is ok, it will just return 0.
  if(log2 > log1){
    double badExp = 1.0 + exp(log1 - log2);
    return log2 + correctLog(badExp);
  }
  double badExp = 1.0 + exp(log2 - log1);
  return log1 + correctLog(badExp);
}

/*
 * Returns log(badNum) with some exceptions:
 * If badNum is less than ZERO, then log(ZERO) is returned (We don't want
 * to get log(0.0) = -inf).
 * If badNum is very close to 1.0, log(1.0 - ZERO) is returned.
 */
double correctLog(double badNum){
  stringstream str;
  str << badNum;
  AssertMsg(badNum >= 0.0, str.str());
  if(badNum <= ZERO) return log(ZERO);
  else if(badNum <= 1.0 && (1.0 - badNum <= ZERO)) return log(1.0 - ZERO);
  //else if(badNum >= INF) return log(1.0 - ZERO);
  return log(badNum);
}

/*
 * Computes log(sum(x(i))) given a vector of log(x(i)).
 * Arguments:
 * - logs: A vector of log(x(i)). Should be non-empty.
 * Return value:
 * log(sum(x(i))) 
 */
double computeLogSum(vector<double> logs){
  Assert(!logs.empty());
  double logSum = logs[0];
  for(int i = 1; i < logs.size(); ++i){
    if(logs[i] > logSum){
      double badExp = 1.0 + exp(logSum - logs[i]);
      logSum = logs[i] + correctLog(badExp);
    }
    else{
      double badExp = 1.0 + exp(logs[i] - logSum);
      logSum += correctLog(badExp);
    }
  }
  return logSum;
}

/*
 * Computes a matrix of binomial coefficients, that is, of the numbers 
 * (n choose k) for all n = 1,...,max. 
 * The matrix is arranged (and computed) as a pascal's triange. 
 * Return value:
 * a two-dimensional vector p, such that p[n][k] = (n choose k).
 * p is arranged like a lower diagonal matrix; p[n] has n + 1 positions.
 */
void computePascalTri(vector< vector<int> > &pascalTri, int max){
  pascalTri.resize(max + 1);
  pascalTri[0].resize(1, 1);
  pascalTri[1].resize(2, 1);
  for(int i = 2; i <= max; ++i){
    pascalTri[i].resize(i + 1, 1);
    //returnVector[i][0] and returnVector[i][i] are (i choose 0) and
    //(i choose i) respectively, so they will be left 1.0.
    for(int j = 1; j < i; ++j){
      pascalTri[i][j] = pascalTri[i - 1][j - 1] + pascalTri[i - 1][j];
    }
  }
}

/*
 * Computes the matrices logRecomb and logNoRecomb.
 * These will be (num SNPs - 1)x(maxGens + 1) matrices.
 */
void computeRecomb(SNPData &data, 
		   vector< vector< double> > &logRecomb,
		   vector< vector< double> > &logNoRecomb, 
		   int maxGens){  
  logRecomb.clear();
  logNoRecomb.clear();
  logRecomb.resize(data.seq_len - 1);
  logNoRecomb.resize(data.seq_len - 1);
  for(int i = 0; i < data.seq_len - 1; ++i){
    double probRecomb = -log(2.0), probNoRecomb = -log(2.0);
    if(data.locus_chrom[i + 1] == data.locus_chrom[i]){
      double dist = data.locus_genetic_pos[i + 1] - data.locus_genetic_pos[i];
      //P(recomb) = 1/2 * (1 - exp(-2dist)) per the Haldane process.
      probRecomb += correctLog(1.0 - exp(-2.0 * dist));
      probNoRecomb += correctLog(1.0 + exp(-2.0 * dist));
      /*if(probRecomb < log(ZERO)){
	probRecomb = log(ZERO);
	probNoRecomb = log(1.0 - ZERO);
      }
      else if(probNoRecomb < log(ZERO)){
	probNoRecomb = log(ZERO);
	probRecomb = log(1.0 - ZERO);
	}*/
    }
    logRecomb[i].resize(maxGens + 1, 0.0);
    logNoRecomb[i].resize(maxGens + 1, 0.0);
    for(int j = 1; j <= maxGens; ++j){
      logRecomb[i][j] = j * probRecomb;
      logNoRecomb[i][j] = j * probNoRecomb;
    }
  }
}

/*
 * Return value:
 * the letter at position (SNP) pos of the haploIdx-th haplotype of the 
 * SNPData object.
 */
int getAllele(SNPData& data, int haploIdx, int pos){
  stringstream stream;
  stream << "Invalid haplo index " << haploIdx;
  AssertMsg(haploIdx < data.num_haplos, stream.str());
  stream.str("");
  stream << "Invalid index " << pos;
  AssertMsg(pos < data.seq_len, stream.str());
  int8 returnValue = data.haplos[haploIdx].data[pos];
  Assert(returnValue >= 0 && returnValue < 5);
  return (int)returnValue;
}

/*
 * Fills in a the matrix letterFreq with the frequencies of letters at each SNP
 * of the data object.
 * letterFreq[i][c] will be the frequency of letter c at position i.
 * A pseudocount of 1 is added to all cells of the matrix.
 */
void computeLetterFreq(SNPData &data, double count, 
		       vector< vector<double> > &letterFreq,
		       int size){
  letterFreq.clear();
  letterFreq.resize(data.seq_len);
  double denominator =  data.num_haplos + size * count;
  for(int i = 0; i < data.seq_len; ++i){
    letterFreq[i].resize(size, count / denominator);
    for(int j = 0; j < data.num_haplos; ++j){
      Assert(getAllele(data, j, i) < size);
      letterFreq[i][getAllele(data, j, i)] += 1.0 / denominator;
    }
  }
}

void initErrorFunction(vector< vector<double> > &errorFunction){
  errorFunction.clear();
  errorFunction.resize(5);
  for(int i = 0; i < 4; ++i){
    errorFunction[i].resize(5, 0.00333);
    errorFunction[i][4] = 0.05001;
    errorFunction[i][i] = 0.94;
  }
  errorFunction[4].resize(5, 0.05001);
  errorFunction[4][4] = 0.79996;
  /*
  for(int i = 0; i < 5; ++i){
    errorFunction[i].resize(5, 0.0);
    errorFunction[i][i] = 1.0;
  }
  */
}

/*
 * Inputs:
 * - predictions: predicted log-probabilities of IBD.
 * - trueIBD: vector of indicators of the true IBD state.
 * Return value:
 * The average squared difference between the true and predicted values.
 */
double squareErrorIBD(vector<double> &predictions, vector<bool> &trueIBD){
  AssertMsg(predictions.size() == trueIBD.size(), 
	    "Input vectors have incompatible sizes.");
  double error = 0.0;
  for(int i = 0; i < predictions.size(); ++i){
    double diff = (double)trueIBD[i] - exp(predictions[i]);
    error += diff * diff;
  }
  return error / (double)predictions.size();
}

bool areAllMissing(SNPData &data, int snpIdx){
  AssertMsg(snpIdx >= 0 && snpIdx < data.seq_len,
	    "SNP index out of range.");
  for(int j = 0; j < data.num_haplos; ++j){
    if((int)getAllele(data, j, snpIdx) != 4) return false;
  }
  return true;
}

void storeIndividual(Simulate::Diplo &diplo, double switchProb, string label,
		     ofstream &outfile){
  diplo.WriteBin(outfile);
  Write(outfile, switchProb);
  Write(outfile, diplo.pop);
  Write(outfile, label);
}

int restoreIndividual(SNPData &data, Simulate::Diplo &diplo, ifstream &infile){
  //cerr << "restoring" << endl;
  diplo.ReadBin(infile);
  diplo.idx = data.num_diplos - 1;
  double switchProb;
  Read(infile, switchProb);
  int pop;
  Read(infile, pop);
  string label;
  Read(infile, label);
  //cerr << diplo.haplos[0].sources.size() << endl;
  vector < vector < int8 > > childHaplos(2);
  //Generate the haplos based on the segments of the Simulate::Diplo object.
  diplo.Generate(data, childHaplos[0], childHaplos[1]);
  //cerr << "childHaplo_0  size" << childHaplos[0].size() << endl; 
  //Add the individual as the last individual of data.
  data.AddDiplo(childHaplos[0], childHaplos[1], switchProb, pop, label);
  data.ComputeDerived();
  return data.num_diplos - 1;
}

/*
 * The haplotype of the individual is restored based on data,
 * but then the individual is added to addToData.
 */
int restoreIndividual(SNPData &data, SNPData &addToData,
		      Simulate::Diplo &diplo, ifstream &infile){
  //cerr << "restoring" << endl;
  diplo.ReadBin(infile);
  diplo.idx = data.num_diplos - 1;
  double switchProb;
  Read(infile, switchProb);
  int pop;
  Read(infile, pop);
  string label;
  Read(infile, label);
  //cerr << diplo.haplos[0].sources.size() << endl;
  vector < vector < int8 > > childHaplos(2);
  //Generate the haplos based on the segments of the Simulate::Diplo object.
  diplo.Generate(data, childHaplos[0], childHaplos[1]);
  //cerr << "childHaplo_0  size" << childHaplos[0].size() << endl; 
  //Add the individual as the last individual of data.
  addToData.AddDiplo(childHaplos[0], childHaplos[1], switchProb, pop, label);
  addToData.ComputeDerived();
  return addToData.num_diplos - 1;
}

void restoreIndividual(Simulate::Diplo &diplo, 
		       ifstream &infile, double &switchProb, int &pop, string &label){
  diplo.ReadBin(infile);
  Read(infile, switchProb);
  Read(infile, pop);
  Read(infile, label);
}
