
#include "System.h"
#include "Utility.h"
#include "SNPData.h"


#define AssertNotEmpty(x) Assert(!(x).empty())
#define AssertEqual(x, y) AssertMsg((x) == (y), ToStr(x) + " != " + ToStr(y))
#define AssertNotEqual(x, y) AssertMsg((x) != (y), ToStr(x) + " == " + ToStr(y))
#define AssertLess(x, y) AssertMsg((x) < (y), ToStr(x) + " >= " + ToStr(y))
#define AssertLessEqual(x, y) AssertMsg((x) <= (y), ToStr(x) + " > " + ToStr(y))
#define AssertGreater(x, y) AssertMsg((x) > (y), ToStr(x) + " <= " + ToStr(y))
#define AssertGreaterEqual(x, y) AssertMsg((x) >= (y), ToStr(x) + " < " + ToStr(y))
#define AssertRange(x, y, z) AssertMsg(((x) >= (y)) && ((x) < (z)), ToStr(x) + " !in [" + ToStr(y) + ", " + ToStr(z) + ")")
#define AssertSegment(x, y, z) AssertMsg(((x).left() >= (y)) && ((x).left() <= (x).right()) && ((x).right() <= (z)), "[" + ToStr((x).left()) + ", " + ToStr((x).right()) + ") !in [" + ToStr(y) + ", " + ToStr(z) + ")")

#define ROL(x, bits) (((x) << (bits)) | ((x) >> (8 * sizeof(x) - (bits))))

bool SNPProfile::status_output = true;
uint32 SNPProfile::magic = 0xcafebeef;


SNPProfile::SNPProfile()
{
}

SNPProfile::SNPProfile(const string &filename)
{
  ReadBin(filename);
}

void SNPProfile::Copy(const SNPProfile &init)
{
  loci = init.loci;
  sry_genomic_pos = init.sry_genomic_pos;
  x_y_par2_genomic_pos_delta = init.x_y_par2_genomic_pos_delta;

  chroms = init.chroms;
  par1_range = init.par1_range;
  par2_range = init.par2_range;
  x_nonpar_range = init.x_nonpar_range;
  y_nonpar_range = init.y_nonpar_range;
  mito_range = init.mito_range;

  ComputeDerived();
}

void SNPProfile::Clear()
{
  loci.clear();
  sry_genomic_pos = -1;
  x_y_par2_genomic_pos_delta = 0;

  chroms.clear();
  par1_range = Segment();
  par2_range = Segment();
  x_nonpar_range = Segment();
  y_nonpar_range = Segment();
  mito_range = Segment();

  seq_len = 0;
  num_chroms = 0;
  sry_pos = -1;
  locus_chrom.clear();
  locus_genomic_pos.clear();
  locus_genomic_pos.clear();
  locus_genetic_pos.clear();
  gender_locus_genetic_pos.clear();

  ComputeDerived();
}

void SNPProfile::Swap(SNPProfile &other)
{
  loci.swap(other.loci);
  swap(sry_genomic_pos, other.sry_genomic_pos);
  swap(x_y_par2_genomic_pos_delta, other.x_y_par2_genomic_pos_delta);

  chroms.swap(other.chroms);
  par1_range.swap(other.par1_range);
  par2_range.swap(other.par2_range);
  x_nonpar_range.swap(other.x_nonpar_range);
  y_nonpar_range.swap(other.y_nonpar_range);
  mito_range.swap(other.mito_range);

  ComputeDerived();
  other.ComputeDerived();
}

void SNPProfile::Validate() const
{
  AssertNotEmpty(loci);
  AssertNotEmpty(chroms);

  if (status_output)
    cerr << "Validating SNP profile: " << loci.size() << " SNPs, " << chroms.size() << " chromosomes, " << (loci.back().genetic_pos - loci.front().genetic_pos) << " Morgans" << flush;

  AssertGreater(seq_len, 0);
  AssertLessEqual(seq_len, loci.size());
  AssertEqual(num_chroms, chroms.size());

  for (int chrom = 0; chrom < num_chroms; ++chrom)
    AssertSegment(chroms[chrom].range, 0, seq_len);
  AssertSegment(par1_range, 0, seq_len);
  AssertSegment(par2_range, 0, seq_len);
  AssertSegment(x_nonpar_range, 0, seq_len);
  AssertSegment(y_nonpar_range, 0, seq_len);
  AssertSegment(mito_range, 0, seq_len);
  AssertSegment(sex_range, 0, seq_len);
  AssertRange(sex_chrom, -1, num_chroms);

  AssertEqual(par1_range.right(), y_nonpar_range.left());
  AssertEqual(y_nonpar_range.right(), x_nonpar_range.left());
  AssertEqual(x_nonpar_range.right(), par2_range.left());

  AssertRange(sry_pos, -1, seq_len + 1);
  AssertGreaterEqual(locus_chrom.size(), seq_len);
  AssertGreaterEqual(locus_genomic_pos.size(), seq_len);
  AssertGreaterEqual(locus_genetic_pos.size(), seq_len);
  AssertGreaterEqual(gender_locus_genetic_pos.size(), 2);
  AssertGreaterEqual(gender_locus_genetic_pos[0].size(), seq_len);
  AssertGreaterEqual(gender_locus_genetic_pos[1].size(), seq_len);

  for (int i = 0; i < seq_len; ++i)
  {
    AssertEqual(locus_chrom[i], loci[i].chrom);
    AssertEqual(locus_genomic_pos[i], loci[i].genomic_pos);
    AssertEqual(locus_genetic_pos[i], loci[i].genetic_pos);
    AssertEqual(gender_locus_genetic_pos[0][i], loci[i].gender_genetic_pos[0]);
    AssertEqual(gender_locus_genetic_pos[1][i], loci[i].gender_genetic_pos[1]);
  }

  for (int i = 0; i < seq_len; ++i)
    AssertRange(loci[i].chrom, 0, num_chroms);

  for (int i = 1; i < seq_len; ++i)
  {
    if ((loci[i - 1].chrom == loci[i].chrom) && (i != x_nonpar_range.left()))
    {
      if (loci[i].genomic_pos <= loci[i - 1].genomic_pos)
      {
        cerr << endl;
        cerr << par1_range.left() << ", " << par1_range.right() << endl;
        cerr << par2_range.left() << ", " << par2_range.right() << endl;
        cerr << x_nonpar_range.left() << ", " << x_nonpar_range.right() << endl;
        cerr << y_nonpar_range.left() << ", " << y_nonpar_range.right() << endl;
        cerr << i << endl;
        cerr << loci[i - 1].chrom << ", " << loci[i - 1].genomic_pos << endl;
        cerr << loci[i].chrom << ", " << loci[i].genomic_pos << endl;
      }
      AssertGreater(loci[i].genomic_pos, loci[i - 1].genomic_pos);
    }

    AssertGreaterEqual(loci[i].genetic_pos, loci[i - 1].genetic_pos);
    AssertGreaterEqual(loci[i].gender_genetic_pos[0], loci[i - 1].gender_genetic_pos[0]);
    AssertGreaterEqual(loci[i].gender_genetic_pos[1], loci[i - 1].gender_genetic_pos[1]);
  }

  if (status_output)
    cerr << ": Valid!" << endl;
}

void SNPProfile::ComputeDerived()
{
  seq_len = loci.size();
  num_chroms = chroms.size();

  sry_pos = y_nonpar_range.left();
  for (int i = y_nonpar_range.left() + 1; i < y_nonpar_range.right(); ++i)
  {
    if ((loci[i].genomic_pos >= sry_genomic_pos) &&
        (loci[i].genomic_pos < loci[sry_pos].genomic_pos))
      sry_pos = i;
  }

  locus_chrom.resize(seq_len);
  locus_genomic_pos.resize(seq_len);
  locus_genetic_pos.resize(seq_len);
  gender_locus_genetic_pos.resize(2);
  gender_locus_genetic_pos[0].resize(seq_len);
  gender_locus_genetic_pos[1].resize(seq_len);
  for (int i = 0; i < seq_len; ++i)
  {
    locus_chrom[i] = loci[i].chrom;
    locus_genomic_pos[i] = loci[i].genomic_pos;
    locus_genetic_pos[i] = loci[i].genetic_pos;
    gender_locus_genetic_pos[0][i] = loci[i].gender_genetic_pos[0];
    gender_locus_genetic_pos[1][i] = loci[i].gender_genetic_pos[1];
  }

  Union(par1_range, par2_range, sex_range);
  sex_chrom = -1;
  for (int chrom = 0; chrom < num_chroms; ++chrom)
    if (Contains(chroms[chrom].range, sex_range))
    {
      sex_chrom = chrom;
      break;
    }
}

uint32 SNPProfile::ComputeHash() const
{
  uint32 hash = uint32(loci.size());
  hash = ROL(hash, 7) + sry_genomic_pos;
  hash = ROL(hash, 7) + x_y_par2_genomic_pos_delta;

  hash = ROL(hash, 7) + uint32(chroms.size());
  for (int chrom = 0; chrom < chroms.size(); ++chrom)
  {
    hash = ROL(hash, 7) + uint32(chroms[chrom].label.length());
    for (int i = 0; i < chroms[chrom].label.length(); ++i)
      hash = ROL(hash, 7) + uint32(chroms[chrom].label[i]);
    hash = ROL(hash, 7) + uint32(chroms[chrom].range.left());
    hash = ROL(hash, 7) + uint32(chroms[chrom].range.right());
  }

  hash = ROL(hash, 7) + uint32(par1_range.left());
  hash = ROL(hash, 7) + uint32(par1_range.right());
  hash = ROL(hash, 7) + uint32(par2_range.left());
  hash = ROL(hash, 7) + uint32(par2_range.right());
  hash = ROL(hash, 7) + uint32(x_nonpar_range.left());
  hash = ROL(hash, 7) + uint32(x_nonpar_range.right());
  hash = ROL(hash, 7) + uint32(y_nonpar_range.left());
  hash = ROL(hash, 7) + uint32(y_nonpar_range.right());
  hash = ROL(hash, 7) + uint32(mito_range.left());
  hash = ROL(hash, 7) + uint32(mito_range.right());

  return hash;
}

bool SNPProfile::Compatible(const SNPProfile &other) const
{
  return (ComputeHash() == other.ComputeHash());
}

void SNPProfile::ReadBin(istream &in)
{
  int32 size;
  pair<int, int> range;

  ::ReadBin(in, size);
  loci.resize(size);
  for (int i = 0; i < size; ++i)
  {
    ::Read(in, loci[i].label);
    ::ReadBin(in, loci[i].strand);
    ::ReadBin(in, loci[i].chrom);
    ::ReadBin(in, loci[i].genomic_pos);
    ::ReadBin(in, loci[i].genetic_pos);
    ::ReadBin(in, loci[i].gender_genetic_pos[0]);
    ::ReadBin(in, loci[i].gender_genetic_pos[1]);
  }

  ::ReadBin(in, sry_genomic_pos);
  ::ReadBin(in, x_y_par2_genomic_pos_delta);

  ::ReadBin(in, size);
  chroms.resize(size);
  for (int chrom = 0; chrom < size; ++chrom)
  {
    ::Read(in, chroms[chrom].label);
    ::ReadBin(in, range);
    chroms[chrom].range = Segment(range.first, range.second, true);
  }

  ::ReadBin(in, range);
  par1_range = Segment(range.first, range.second, true);
  ::ReadBin(in, range);
  par2_range = Segment(range.first, range.second, true);
  ::ReadBin(in, range);
  x_nonpar_range = Segment(range.first, range.second, true);
  ::ReadBin(in, range);
  y_nonpar_range = Segment(range.first, range.second, true);
  ::ReadBin(in, range);
  mito_range = Segment(range.first, range.second, true);

  uint32 read_magic;
  ::ReadBin(in, read_magic);
  AssertMsg(read_magic == magic, ToHex(read_magic, true) + " != " + ToHex(magic, true));

  ComputeDerived();
}

void SNPProfile::ReadBin(const string &filename)
{
  istream *in = InFileStream(filename);
  AssertMsg(in, filename);
  ReadBin(*in);
  delete in;
}

void SNPProfile::WriteBin(ostream &out) const
{
  int32 size;
  pair<int, int> range;

  size = loci.size();
  ::WriteBin(out, size);
  for (int i = 0; i < size; ++i)
  {
    ::Write(out, loci[i].label);
    ::WriteBin(out, loci[i].strand);
    ::WriteBin(out, loci[i].chrom);
    ::WriteBin(out, loci[i].genomic_pos);
    ::WriteBin(out, loci[i].genetic_pos);
    ::WriteBin(out, loci[i].gender_genetic_pos[0]);
    ::WriteBin(out, loci[i].gender_genetic_pos[1]);
  }

  ::WriteBin(out, sry_genomic_pos);
  ::WriteBin(out, x_y_par2_genomic_pos_delta);

  size = chroms.size();
  ::WriteBin(out, size);
  for (int chrom = 0; chrom < size; ++chrom)
  {
    ::Write(out, chroms[chrom].label);
    range = pair<int, int>(chroms[chrom].range.left(), chroms[chrom].range.right());
    ::WriteBin(out, range);
  }

  range = pair<int, int>(par1_range.left(), par1_range.right());
  ::WriteBin(out, range);
  range = pair<int, int>(par2_range.left(), par2_range.right());
  ::WriteBin(out, range);
  range = pair<int, int>(x_nonpar_range.left(), x_nonpar_range.right());
  ::WriteBin(out, range);
  range = pair<int, int>(y_nonpar_range.left(), y_nonpar_range.right());
  ::WriteBin(out, range);
  range = pair<int, int>(mito_range.left(), mito_range.right());
  ::WriteBin(out, range);

  uint32 write_magic = magic;;
  ::WriteBin(out, write_magic);
}

void SNPProfile::WriteBin(const string &filename) const
{
  ostream *out = OutFileStream(filename);
  AssertMsg(out, filename);
  WriteBin(*out);
  delete out;
}

void SNPProfile::CopyRestrictLoci(const SNPProfile &init, 
                                  const vector<int> &select)
{
  AssertEqual(select.size(), init.loci.size());
  
  int new_seq_len = 0;
  for (int i = 0; i < init.loci.size(); ++i)
    if (select[i])
      ++new_seq_len;

  vector<int> index_map;
  index_map.reserve(new_seq_len);
  vector<int> index_unmap;
  index_unmap.reserve(init.loci.size() + 1);

  for (int i = 0; i < init.loci.size(); ++i)
  {
    index_unmap.push_back(index_map.size());
    if (select[i])
      index_map.push_back(i);
  }
  index_map.push_back(init.loci.size());
  index_unmap.push_back(new_seq_len);

  Clear();

  vector<int> chrom_map;
  for (int chrom = 0; chrom < init.chroms.size(); ++chrom)
  {
    if (index_unmap[init.chroms[chrom].range.left()] < index_unmap[init.chroms[chrom].range.right()])
    {
      chrom_map.push_back(chroms.size());
      chroms.push_back(Chrom(init.chroms[chrom].label, Segment(index_unmap[init.chroms[chrom].range.left()], index_unmap[init.chroms[chrom].range.right()], true)));
    }
    else
      chrom_map.push_back(-1);
  }

  loci.resize(new_seq_len);
  for (int i = 0; i < new_seq_len; ++i)
  {
    loci[i] = init.loci[index_map[i]];
    loci[i].chrom = chrom_map[loci[i].chrom];
    AssertRange(loci[i].chrom, 0, chroms.size());
  }
  
  sry_genomic_pos = init.sry_genomic_pos;
  x_y_par2_genomic_pos_delta = init.x_y_par2_genomic_pos_delta;

  par1_range.SetLeft(index_unmap[init.par1_range.left()]);
  par1_range.SetRight(index_unmap[init.par1_range.right()]);
  par2_range.SetLeft(index_unmap[init.par2_range.left()]);
  par2_range.SetRight(index_unmap[init.par2_range.right()]);
  x_nonpar_range.SetLeft(index_unmap[init.x_nonpar_range.left()]);
  x_nonpar_range.SetRight(index_unmap[init.x_nonpar_range.right()]);
  y_nonpar_range.SetLeft(index_unmap[init.y_nonpar_range.left()]);
  y_nonpar_range.SetRight(index_unmap[init.y_nonpar_range.right()]);
  mito_range.SetLeft(index_unmap[init.mito_range.left()]);
  mito_range.SetRight(index_unmap[init.mito_range.right()]);

  ComputeDerived();
}

void SNPProfile::CopyRestrictChrom(const SNPProfile &init, int chrom)
{
  AssertRange(chrom, 0, init.chroms.size());

  vector<int> select(init.chroms[chrom].range.left(), 0);
  select.resize(init.chroms[chrom].range.right(), 1);
  select.resize(init.loci.size(), 0);

  CopyRestrictLoci(init, select);
}


template <>
void Write<SNPProfile>(ostream &o, const SNPProfile &x)
{
  x.WriteBin(o);
}

template <>
void Read<SNPProfile>(istream &i, SNPProfile &x)
{
  x.ReadBin(i);
}

//////////////////////////////////////////////////////////////////////////////


bool SNPData::status_output = true;
uint32 SNPData::magic = 0xcafebeef;
vector<double> SNPData::switch_code;


void SNPData::Pop::WriteBin(ostream &out) const
{
  ::Write(out, label);
}

void SNPData::Pop::ReadBin(istream &in)
{
  ::Read(in, label);
}

int SNPData::Pop::haplo_index(int k) const
{
  vector<int>::const_iterator find = lower_bound(haplos.begin(), haplos.end(), k);
  if ((find == haplos.end()) || (*find != k))
    return -1;
  else
    return (find - haplos.begin());
}

int SNPData::Pop::diplo_index(int k) const
{
  vector<int>::const_iterator find = lower_bound(diplos.begin(), diplos.end(), k);
  if ((find == diplos.end()) || (*find != k))
    return -1;
  else
    return (find - diplos.begin());
}

void SNPData::Haplo::WriteBin(ostream &out) const
{
  ::Write(out, label);
  ::WriteBin(out, pop);
  ::WriteBin(out, sry);
  ::WriteBin(out, data);
}

void SNPData::Haplo::ReadBin(istream &in)
{
  ::Read(in, label);
  ::ReadBin(in, pop);
  ::ReadBin(in, sry);
  ::ReadBin(in, data);
}

void SNPData::Diplo::WriteBin(ostream &out) const
{
  ::Write(out, label);
  ::WriteBin(out, haplos);
  ::WriteBin(out, switches);
}

void SNPData::Diplo::ReadBin(istream &in)
{
  ::Read(in, label);
  ::ReadBin(in, haplos);
  ::ReadBin(in, switches);
}


void SNPData::Init()
{
  if (switch_code.empty())
  {
    switch_code.resize(256, 0.0);
    switch_code[0] = 0.0;
    switch_code[255] = 0.5;
    for (int c = 1; c <= 254; ++c)
      switch_code[c] = 0.5 * exp(-0.1 * (255 - c));
  }
}

void SNPData::SetProfile(const SNPProfile *profile_)
{
  profile = profile_;
  if (!profile)
  {
    loci = NULL;
    chroms = NULL;
    locus_chrom = NULL;
    locus_genomic_pos = NULL;
    locus_genetic_pos = NULL;
    gender_locus_genetic_pos = NULL;
    sry_pos = -1;
  }
  else
  {
    loci = &(profile->loci[0]);
    chroms = &(profile->chroms[0]);
    locus_chrom = &(profile->locus_chrom[0]);
    locus_genomic_pos = &(profile->locus_genomic_pos[0]);
    locus_genetic_pos = &(profile->locus_genetic_pos[0]);
    gender_locus_genetic_pos = &(profile->gender_locus_genetic_pos[0]);
    sry_pos = profile->sry_pos;
    seq_len = profile->seq_len;
    num_chroms = profile->num_chroms;
  }
}

SNPData::SNPData()
{
  Init();
  SetProfile(NULL);
}

SNPData::SNPData(const SNPProfile &profile_)
{
  Init();
  SetProfile(profile_);
}

SNPData::SNPData(const SNPProfile &profile_, const string &filename)
{
  Init();
  SetProfile(profile_);
  ReadBin(filename);
}

SNPData::SNPData(const SNPData &init, bool copy_data)
{
  Init();
  Copy(init, copy_data);
}

void SNPData::SetProfile(const SNPProfile &profile_)
{
  SetProfile(&profile_);
}

void SNPData::Copy(const SNPData &init, bool copy_data)
{
  SetProfile(init.profile);
  Clear();
  if (copy_data)
  {
    pops = init.pops;
    haplos = init.haplos;
    diplos = init.diplos;
  }

  ComputeDerived();
}


void SNPData::Clear()
{
  pops.clear();
  haplos.clear();
  diplos.clear();

  ComputeDerived();
}

void SNPData::Swap(SNPData &other)
{
  swap(profile, other.profile);

  pops.swap(other.pops);
  haplos.swap(other.haplos);
  diplos.swap(other.diplos);
  
  ComputeDerived();
  other.ComputeDerived();
}    

void SNPData::ClearDiploPairing()
{
  diplos.clear();
  ComputeDerived();
}

void SNPData::Validate() const
{
  AssertNotEmpty(switch_code);
  Assert(profile);
  Assert(loci);
  Assert(chroms);
  Assert(locus_chrom);
  Assert(locus_genomic_pos);
  Assert(locus_genetic_pos);
  Assert(gender_locus_genetic_pos);
  AssertRange(sry_pos, -1, profile->seq_len + 1);
  
  if (status_output)
    cerr << "Validating SNP data: " << seq_len << " SNPs, " << num_chroms << " chromosomes, " << num_pops << " populations, " << num_haplos << " haplotypes, " << num_diplos << " diploid" << flush;

  AssertGreater(seq_len, 0);
  AssertLessEqual(seq_len, profile->seq_len);
  AssertEqual(num_chroms, profile->num_chroms);
  AssertEqual(num_pops, pops.size());
  AssertEqual(num_haplos, haplos.size());
  AssertEqual(num_diplos, diplos.size());
  AssertRange(num_pops, 1, 65);
  AssertGreater(num_haplos, 0);
  AssertGreaterEqual(num_diplos, 0);

  for (int pop = 0; pop < num_pops; ++pop)
  {
    for (int i = 0; i < pops[pop].haplos.size(); ++i)
    {
      AssertRange(pops[pop].haplos[i], 0, num_haplos);
      AssertEqual(haplos[pops[pop].haplos[i]].pop, pop);
      if (i > 0)
        AssertGreater(pops[pop].haplos[i], pops[pop].haplos[i - 1]);
    }

    for (int i = 0; i < pops[pop].diplos.size(); ++i)
    {
      AssertRange(pops[pop].diplos[i], 0, num_diplos);
      AssertEqual(diplos[pops[pop].diplos[i]].pop, pop);
      if (i > 0)
        AssertGreater(pops[pop].diplos[i], pops[pop].diplos[i - 1]);
    }
  }

  for (int k = 0; k < num_haplos; ++k)
  {
    AssertRange(haplos[k].pop, 0, num_pops);
    AssertGreaterEqual(pops[haplos[k].pop].haplo_index(k), 0);
    AssertRange(haplos[k].sry, 0, 2);
    AssertGreaterEqual(haplos[k].data.size(), seq_len);
    for (int i = 0; i < seq_len; ++i)
      AssertRange(haplos[k][i], 0, 5);

    int d = haplos[k].diplo;
    int kp = haplos[k].diplo_pair;
    if (d == -1)
      AssertEqual(kp, -1);
    else
    {
      AssertRange(d, 0, num_diplos);
      AssertRange(kp, 0, num_haplos);
      Assert(((diplos[d].haplos[0] == k) && (diplos[d].haplos[1] == kp)) ||
             ((diplos[d].haplos[0] == kp) && (diplos[d].haplos[1] == k)));
    }
  }

  {
    vector<int> referenced(num_haplos, 0);
    for (int d = 0; d < num_diplos; ++d)
    {
      for (int p = 0; p < 2; ++p)
      {
        int k = diplos[d].haplos[p];
        AssertRange(k, 0, num_haplos);
        Assert(!referenced[k]);
        AssertEqual(haplos[k].diplo, d);
        AssertEqual(haplos[k].diplo_pair, diplos[d].haplos[1 - p]);
        referenced[k] = 1;

        AssertEqual(haplos[k].pop, diplos[d].pop);
      }

      AssertGreaterEqual(diplos[d].switches.size(), seq_len - 1);
      AssertRange(diplos[d].gender, 0, 2);
      Assert(!(haplos[diplos[d].haplos[0]].sry && haplos[diplos[d].haplos[1]].sry));
      if (haplos[diplos[d].haplos[0]].sry || haplos[diplos[d].haplos[1]].sry)
        AssertEqual(diplos[d].gender, 1);
      else
        AssertEqual(diplos[d].gender, 0);
    }
  }

  AssertEqual(haplo_pop.size(), num_haplos);
  AssertEqual(haplo_data.size(), num_haplos);
  for (int k = 0; k < num_haplos; ++k)
  {
    AssertEqual(haplo_pop[k], haplos[k].pop);
    AssertEqual(haplo_data[k], &(haplos[k].data[0]));
  }

  AssertEqual(diplo_pop.size(), num_diplos);
  AssertEqual(diplo_switches.size(), num_diplos);
  for (int d = 0; d < num_diplos; ++d)
  {
    AssertEqual(diplo_pop[d], diplos[d].pop);
    AssertEqual(diplo_switches[d], &(diplos[d].switches[0]));
  }

  if (status_output)
    cerr << ": Valid!" << endl;  
};


void SNPData::ComputeDerived()
{
  num_pops = pops.size();
  num_haplos = haplos.size();
  num_diplos = diplos.size();

  for (int k = 0; k < num_haplos; ++k)
  {
    haplos[k].diplo = -1;
    haplos[k].diplo_pair = -1;
  }

  for (int d = 0; d < num_diplos; ++d)
  {
    diplos[d].pop = haplos[diplos[d].haplos[0]].pop;
    AssertEqual(diplos[d].pop, haplos[diplos[d].haplos[1]].pop);
    
    diplos[d].gender = (haplos[diplos[d].haplos[0]].sry || haplos[diplos[d].haplos[1]].sry) ? 1 : 0;

    for (int p = 0; p < 2; ++p)
    {
      haplos[diplos[d].haplos[p]].diplo = d;
      haplos[diplos[d].haplos[p]].diplo_pair = diplos[d].haplos[1 - p];
    }
  }
  
  for (int pop = 0; pop < num_pops; ++pop)
  {
    pops[pop].haplos.clear();
    for (int k = 0; k < num_haplos; ++k)
      if (haplos[k].pop == pop)
        pops[pop].haplos.push_back(k);

    pops[pop].diplos.clear();
    for (int d = 0; d < num_diplos; ++d)
      if (diplos[d].pop == pop)
        pops[pop].diplos.push_back(d);
  }

  haplo_pop.resize(num_haplos);
  haplo_data.resize(num_haplos);
  for (int k = 0; k < num_haplos; ++k)
  {
    haplo_pop[k] = haplos[k].pop;
    haplo_data[k] = &(haplos[k].data[0]);
  }

  diplo_pop.resize(num_diplos);
  diplo_switches.resize(num_diplos);
  for (int d = 0; d < num_diplos; ++d)
  {
    diplo_pop[d] = diplos[d].pop;
    diplo_switches[d] = &(diplos[d].switches[0]);
  }
}

bool SNPData::Compatible(const SNPData &other) const
{
  return profile->Compatible(*other.profile);
}

void SNPData::ReadBin(istream &in)
{
  uint32 profile_hash;
  ::ReadBin(in, profile_hash);
  AssertMsg(profile_hash == profile->ComputeHash(), "Incompatible SNP profiles: " + ToHex(profile_hash, true) + " != " + ToHex(profile->ComputeHash(), true));

  int32 size;
  ::ReadBin(in, size);
  pops.resize(size);
  for (int pop = 0; pop < size; ++pop)
    pops[pop].ReadBin(in);

  ::ReadBin(in, size);
  haplos.resize(size);
  for (int k = 0; k < size; ++k)
    haplos[k].ReadBin(in);

  ::ReadBin(in, size);
  diplos.resize(size);
  for (int d = 0; d < size; ++d)
    diplos[d].ReadBin(in);

  uint32 read_magic;
  ::ReadBin(in, read_magic);
  AssertMsg(read_magic == magic, ToHex(read_magic, true) + " != " + ToHex(magic, true));

  ComputeDerived();
}

void SNPData::ReadBin(const string &filename)
{
  istream *in = InFileStream(filename);
  AssertMsg(in, filename);
  ReadBin(*in);
  delete in;
}

void SNPData::WriteBin(ostream &out) const
{
  uint32 profile_hash = profile->ComputeHash();
  ::WriteBin(out, profile_hash);

  int32 size;
  size = pops.size();
  ::WriteBin(out, size);
  for (int pop = 0; pop < size; ++pop)
    pops[pop].WriteBin(out);

  size = haplos.size();
  ::WriteBin(out, size);
  for (int k = 0; k < size; ++k)
    haplos[k].WriteBin(out);

  size = diplos.size();
  ::WriteBin(out, size);
  for (int d = 0; d < size; ++d)
    diplos[d].WriteBin(out);

  uint32 write_magic = magic;;
  ::WriteBin(out, write_magic);
}

void SNPData::WriteBin(const string &filename) const
{
  ostream *out = OutFileStream(filename);
  AssertMsg(out, filename);
  WriteBin(*out);
  delete out;
}


void SNPData::SetDiploSwitch(const vector<int8> &haplo0, 
                             const vector<int8> &haplo1, int gender,
                             vector<uint8> &switches, uint8 value) const
{
  Assert(seq_len > 0);
  Assert(haplo0.size() >= seq_len);
  Assert(haplo1.size() >= seq_len);
  switches.resize(seq_len - 1);

  const double *genetic_pos = (gender < 0) ? locus_genetic_pos : &(gender_locus_genetic_pos[gender][0]);

  uint8 zero = switch_encode(0.0);
  uint8 half = switch_encode(0.5);

  for (int i = 0; i < (seq_len - 1); ++i)
  {
    if (profile->loci[i].chrom != profile->loci[i + 1].chrom)
      switches[i] = half;
    else
    {
      if ((haplo0[i] == haplo1[i]) && (haplo0[i] != 4))
        switches[i] = zero;
      else
      {
        int j = i + 1;
        while ((j < (seq_len - 1)) && (haplo0[j] == haplo1[j]) && (haplo0[j] != 4))
          ++j;
    
        if ((j >= (seq_len - 1)) || 
            (profile->loci[i].chrom != profile->loci[j].chrom) ||
            (genetic_pos[i] == genetic_pos[j]))
          switches[i] = zero;
        else
          switches[i] = value;
      }
    }
  }
}



void SNPData::SetDiploSwitch(int d, uint8 value)
{
  int h0 = diplos[d].haplos[0];
  int h1 = diplos[d].haplos[1];
  int gender = (haplos[h0].sry || haplos[h1].sry) ? 1 : 0;
  SetDiploSwitch(haplos[h0].data, haplos[h1].data, gender, 
                 diplos[d].switches, value);
}

void SNPData::SetAllDiploSwitch(uint8 value)
{
  for (int d = 0; d < num_diplos; ++d)
    SetDiploSwitch(d, value);
}


void SNPData::RestrictHaplos(const vector<int> &select)
{
  num_haplos = haplos.size();
  num_diplos = diplos.size();
  Assert(select.size() == num_haplos);

  vector<int> haplo_map;
  vector<int> new_haplo_map(num_haplos, -1);
  int new_num_haplos = 0;
  for (int k = 0; k < num_haplos; ++k)
    if (select[k])
    {
      haplo_map.push_back(k);
      new_haplo_map[k] = new_num_haplos++;
    }

  {
    vector<Haplo> new_haplos(new_num_haplos);
    for (int k = 0; k < new_num_haplos; ++k)
    {
      int mk = haplo_map[k];
      new_haplos[k].label = haplos[mk].label;
      new_haplos[k].pop = haplos[mk].pop;
      new_haplos[k].sry = haplos[mk].sry;
      new_haplos[k].data.swap(haplos[mk].data);
    }

    haplos.swap(new_haplos);
  }

  {
    vector<Diplo> new_diplos;
    for (int d = 0; d < num_diplos; ++d)
      if (select[diplos[d].haplos[0]] && select[diplos[d].haplos[1]])
      {
        new_diplos.push_back(Diplo(diplos[d].label, new_haplo_map[diplos[d].haplos[0]], new_haplo_map[diplos[d].haplos[1]], vector<uint8>()));
        new_diplos.back().switches.swap(diplos[d].switches);
      }

    diplos.swap(new_diplos);
  }

  ComputeDerived();
}


void SNPData::RestrictDiplos(const vector<int> &select)
{
  num_haplos = haplos.size();
  num_diplos = diplos.size();
  Assert(select.size() == num_diplos);
  
  vector<int> select_haplo(num_haplos, 0);
  for (int d = 0; d < num_diplos; ++d)
    if (select[d])
    {
      select_haplo[diplos[d].haplos[0]] = 1;
      select_haplo[diplos[d].haplos[1]] = 1;
    }

  RestrictHaplos(select_haplo);
}

void SNPData::RestrictPops(const vector<int> &select)
{
  num_haplos = haplos.size();
  num_pops = pops.size();
  Assert(select.size() == num_pops);

  vector<int> select_haplo(num_haplos, 0);
  for (int k = 0; k < num_haplos; ++k)
    if (select[haplos[k].pop])
      select_haplo[k] = 1;

  RestrictHaplos(select_haplo);
}

void SNPData::RestrictPops(uint64 pop_mask)
{
  num_pops = pops.size();
  vector<int> select(num_pops, 0);
  for (int pop = 0; pop < num_pops; ++pop)
    if (pop_mask & (uint64(1) << pop))
      select[pop] = 1;
  RestrictPops(select);
}

void SNPData::RestrictPop(int pop)
{
  num_pops = pops.size();
  Assert((pop >= 0) && (pop < num_pops));
  vector<int> select(num_pops, 0);
  select[pop] = 1;
  RestrictPops(select);
}


void SNPData::MapPops(const vector<int> &pop_map)
{
  num_pops = pops.size();
  Assert(pop_map.size() == num_pops);
  int max_pop = 0;
  for (int pop = 0; pop < num_pops; ++pop)
  {
    Assert(pop_map[pop] >= 0);
    if (pop_map[pop] > max_pop)
      max_pop = pop_map[pop];
  }

  vector<int> map_count(max_pop + 1, 0);
  for (int pop = 0; pop < num_pops; ++pop)
    ++map_count[pop_map[pop]];

  vector<Pop> new_pops(max_pop + 1);
  for (int pop = 0; pop <= max_pop; ++pop)
    if (map_count[pop] > 1)
      new_pops[pop].label = "{";

  for (int pop = 0; pop < num_pops; ++pop)
  {
    int new_pop = pop_map[pop];
    if (map_count[new_pop] > 1)
      new_pops[new_pop].label = new_pops[new_pop].label + " " + pops[pop].label;
    else
      new_pops[new_pop].label = pops[pop].label;
  }

  for (int pop = 0; pop <= max_pop; ++pop)
    if (map_count[pop] > 1)
      new_pops[pop].label = new_pops[pop].label + " }";

  for (int k = 0; k < num_haplos; ++k)
    haplos[k].pop = pop_map[haplos[k].pop];

  pops.swap(new_pops);

  ComputeDerived();
}


int SNPData::AddHaplo(const vector<int8> &data, int pop, 
                      const string &label, int sry)
{
  haplos.push_back(Haplo(label, pop, sry, data));
  return int(haplos.size()) - 1;
}

int SNPData::AddHaplo(const Haplo &haplo, int pop, const string &label)
{
  return AddHaplo(haplo.data, (pop == INT_MIN) ? haplo.pop : pop, 
                  (label == "~") ? haplo.label : label, haplo.sry);
}

int SNPData::AddHaplo(const SNPData &data, int k, 
                       int pop, const string &label)
{
  return AddHaplo(data.haplos[k], pop,label);
}


int SNPData::AddDiplo(const vector<int8> &haplo0, 
                      const vector<int8> &haplo1,
                      const vector<uint8> &diplo_switch, int pop,
                      const string &label, int sry0, int sry1)
{
  haplos.push_back(Haplo(label + "_0", pop, sry0, haplo0));
  haplos.push_back(Haplo(label + "_1", pop, sry1, haplo1));
  diplos.push_back(Diplo(label, haplos.size() - 2, haplos.size() - 1, diplo_switch));
  return int(diplos.size()) - 1;
}

int SNPData::AddDiplo(const vector<int8> &haplo0, 
                      const vector<int8> &haplo1,
                      const vector<double> &diplo_switch, int pop,
                      const string &label, int sry0, int sry1)
{
  vector<uint8> i_diplo_switch(diplo_switch.size());
  for (int i = 0; i < diplo_switch.size(); ++i)
    i_diplo_switch[i] = switch_encode(diplo_switch[i]);
  return AddDiplo(haplo0, haplo1, i_diplo_switch, pop, label, sry0, sry1);
}

int SNPData::AddDiplo(const vector<int8> &haplo0,
                      const vector<int8> &haplo1,
                      double switch_error_rate, int pop,
                      const string &label, int sry0, int sry1)
{
  vector<uint8> diplo_switch(seq_len - 1, switch_encode(0.0));
  int d = AddDiplo(haplo0, haplo1, diplo_switch, pop, label, sry0, sry1);
  SetDiploSwitch(d, switch_error_rate);
}

int SNPData::AddDiplo(const SNPData &data, int d,
                      int pop, const string &label)
{
  int h0 = data.diplos[d].haplos[0];
  int h1 = data.diplos[d].haplos[1];
  return AddDiplo(data.haplos[h0].data, data.haplos[h1].data, 
                  data.diplos[d].switches,
                  (pop == INT_MIN) ? data.diplos[d].pop : pop,
                  (label == "~") ? data.diplos[d].label : label,
                  data.haplos[h0].sry, data.haplos[h1].sry);
}

void SNPData::Concat(const SNPData &other)
{
  Assert(Compatible(other));
  num_pops = pops.size();
  num_haplos = haplos.size();
  num_diplos = diplos.size();

  pops.insert(pops.end(), other.pops.begin(), other.pops.end());

  haplos.insert(haplos.end(), other.haplos.begin(), other.haplos.end());
  for (int k = num_haplos; k < haplos.size(); ++k)
    haplos[k].pop += num_pops;

  diplos.insert(diplos.end(), other.diplos.begin(), other.diplos.end());
  for (int d = num_diplos; d < diplos.size(); ++d)
  {
    diplos[d].haplos[0] += num_haplos;
    diplos[d].haplos[1] += num_haplos;
  }

  ComputeDerived();
}


void SNPData::MergePopIndex(const SNPData &other)
{
  Assert(Compatible(other));
  num_pops = pops.size();
  num_haplos = haplos.size();
  num_diplos = diplos.size();

  pops.resize(max(pops.size(), other.pops.size()));
  for (int pop = num_pops; pop < other.pops.size(); ++pop)
    pops[pop] = other.pops[pop];

  haplos.insert(haplos.end(), other.haplos.begin(), other.haplos.end());

  diplos.insert(diplos.end(), other.diplos.begin(), other.diplos.end());
  for (int d = num_diplos; d < diplos.size(); ++d)
  {
    diplos[d].haplos[0] += num_haplos;
    diplos[d].haplos[1] += num_haplos;
  }

  ComputeDerived();
}


void SNPData::MergePopLabel(const SNPData &other)
{
  Assert(Compatible(other));
  num_pops = pops.size();
  num_haplos = haplos.size();
  num_diplos = diplos.size();

  vector<int> pop_map(other.pops.size(), -1);
  for (int pop = 0; pop < other.pops.size(); ++pop)
  {
    int find = -1;
    for (int p = 0; p < pops.size(); ++p)
      if (pops[p].label == other.pops[pop].label)
      {
        find = p;
        break;
      }

    if (find < 0)
    {
      find = pops.size();
      pops.push_back(other.pops[pop]);
    }
    
    pop_map[pop] = find;
  }
  
  haplos.insert(haplos.end(), other.haplos.begin(), other.haplos.end());
  for (int k = num_haplos; k < haplos.size(); ++k)
    haplos[k].pop = pop_map[haplos[k].pop];

  diplos.insert(diplos.end(), other.diplos.begin(), other.diplos.end());
  for (int d = num_diplos; d < diplos.size(); ++d)
  {
    diplos[d].haplos[0] += num_haplos;
    diplos[d].haplos[1] += num_haplos;
  }

  ComputeDerived();
}


void SNPData::CopyRestrictLoci(const SNPData &init, const vector<int> &select)
{
  AssertEqual(select.size(), init.profile->loci.size());
  
  int new_seq_len = 0;
  for (int i = 0; i < init.profile->loci.size(); ++i)
    if (select[i])
      ++new_seq_len;

  vector<int> index_map;
  index_map.reserve(new_seq_len);
  for (int i = 0; i < init.profile->loci.size(); ++i)
    if (select[i])
      index_map.push_back(i);
  index_map.push_back(init.profile->loci.size());

  pops = init.pops;

  haplos.resize(init.haplos.size());
  for (int k = 0; k < haplos.size(); ++k)
  {
    haplos[k].label = init.haplos[k].label;
    haplos[k].pop = init.haplos[k].pop;
    haplos[k].sry = init.haplos[k].sry;
    haplos[k].data.resize(new_seq_len);
    for (int i = 0; i < new_seq_len; ++i)
      haplos[k][i] = init.haplos[k][index_map[i]];
  }

  diplos.resize(init.diplos.size());
  for (int d = 0; d < diplos.size(); ++d)
  {
    diplos[d].label = init.diplos[d].label;
    diplos[d].haplos[0] = init.diplos[d].haplos[0];
    diplos[d].haplos[1] = init.diplos[d].haplos[1];
    if (new_seq_len > 0)
      diplos[d].switches.resize(new_seq_len - 1);
    for (int i = 0; i < (new_seq_len - 1); ++i)
    {
      int begin = index_map[i];
      int end = index_map[i + 1];
      double total = 0.0;
      for (int j = begin; j < end; ++j)
        total += switch_decode(init.diplos[d].switches[j]);
      if (total > 0.5)
        total = 0.5;
      diplos[d].switches[i] = switch_encode(total);
    }
  }

  ComputeDerived();
}

void SNPData::CopyRestrictChrom(const SNPData &init, int chrom)
{
  AssertRange(chrom, 0, init.profile->chroms.size());

  vector<int> select(init.profile->chroms[chrom].range.left(), 0);
  select.resize(init.profile->chroms[chrom].range.right(), 1);
  select.resize(init.profile->loci.size(), 0);

  CopyRestrictLoci(init, select);
}


void SNPData::CopyRestrictHaplos(const SNPData &init, 
                                 const vector<int> &select)
{
  Assert(select.size() == init.haplos.size());

  vector<int> haplo_map;
  vector<int> new_haplo_map(init.haplos.size(), -1);
  int num_haplos = 0;
  for (int k = 0; k < init.haplos.size(); ++k)
    if (select[k])
    {
      haplo_map.push_back(k);
      new_haplo_map[k] = num_haplos++;
    }

  SetProfile(*init.profile);
  Clear();

  pops = init.pops;

  haplos.resize(num_haplos);
  for (int k = 0; k < haplos.size(); ++k)
  {
    haplos[k].label = init.haplos[haplo_map[k]].label;
    haplos[k].pop = init.haplos[haplo_map[k]].pop;
    haplos[k].sry = init.haplos[haplo_map[k]].sry;
    haplos[k].data = init.haplos[haplo_map[k]].data;
  }

  for (int d = 0; d < init.diplos.size(); ++d)
  {
    if (select[init.diplos[d].haplos[0]] && select[init.diplos[d].haplos[1]])
    {
      diplos.push_back(Diplo());
      diplos.back().label = init.diplos[d].label;
      diplos.back().haplos[0] = new_haplo_map[init.diplos[d].haplos[0]];
      diplos.back().haplos[1] = new_haplo_map[init.diplos[d].haplos[1]];
      diplos.back().switches = init.diplos[d].switches;
    }
  }

  ComputeDerived();
}
  
  
void SNPData::CopyRestrictDiplos(const SNPData &init, 
                                 const vector<int> &select)
{
  Assert(select.size() == init.diplos.size());

  vector<int> select_haplo(init.haplos.size(), 0);
  for (int d = 0; d < init.diplos.size(); ++d)
    if (select[d])
    {
      select_haplo[init.diplos[d].haplos[0]] = 1;
      select_haplo[init.diplos[d].haplos[1]] = 1;
    }

  CopyRestrictHaplos(init, select_haplo);
}

void SNPData::CopyRestrictPops(const SNPData &init, const vector<int> &select)
{
  Assert(select.size() == init.pops.size());
  
  vector<int> select_haplo(init.haplos.size(), 0);
  for (int k = 0; k < init.haplos.size(); ++k)
    if (select[init.haplos[k].pop])
      select_haplo[k] = 1;

  CopyRestrictHaplos(init, select_haplo);
}

void SNPData::CopyRestrictPops(const SNPData &init, uint64 pop_mask)
{
  vector<int> select_pop(init.pops.size(), 0);
  for (int pop = 0; pop < init.pops.size(); ++pop)
    if (pop_mask & (uint64(1) << pop))
      select_pop[pop] = 1;

  CopyRestrictPops(init, select_pop);
}

void SNPData::CopyRestrictPop(const SNPData &init, int pop)
{
  Assert((pop >= 0) && (pop < init.pops.size()));
  vector<int> select_pop(init.pops.size(), 0);
  select_pop[pop] = 1;

  CopyRestrictPops(init, select_pop);
}  


template <>
void Write<SNPData>(ostream &o, const SNPData &x)
{
  x.WriteBin(o);
}

template <>
void Read<SNPData>(istream &i, SNPData &x)
{
  x.ReadBin(i);
}


void ReadBinSNPProfileData(istream &in, SNPProfile &profile, SNPData &data)
{
  profile.ReadBin(in);
  data.SetProfile(profile);
  data.ReadBin(in);
}

void ReadBinSNPProfileData(const string &filename,
                           SNPProfile &profile, SNPData &data)
{
  istream *in = InFileStream(filename);
  AssertMsg(in, filename);
  ReadBinSNPProfileData(*in, profile, data);
  delete in;
}

void WriteBinSNPProfileData(ostream &out, const SNPData &data)
{
  data.profile->WriteBin(out);
  data.WriteBin(out);
}

void WriteBinSNPProfileData(const string &filename, const SNPData &data)
{
  ostream *out = OutFileStream(filename);
  AssertMsg(out, filename);
  WriteBinSNPProfileData(*out, data);
  delete out;
}

