
#ifndef SNPDATA_H
#define SNPDATA_H

#include "System.h"
#include "Segment.h"
#include "limits.h"

class SNPProfile
{
public:
  static bool status_output;
  static uint32 magic;

  /*** Fundamental Data ***/

  // SNP/location data

  struct Locus
  {
    string label;
    bool strand;

    int chrom;
    int32 genomic_pos;

    double genetic_pos;
    double gender_genetic_pos[2];

    Locus() { }
    Locus(const string &label_, bool strand_, int chrom_, 
          int32 genomic_pos_, double genetic_pos_, 
          double gender_genetic_pos0, double gender_genetic_pos1)
      : label(label_), strand(strand_), chrom(chrom_), 
         genomic_pos(genomic_pos_), genetic_pos(genetic_pos_)
    {
      gender_genetic_pos[0] = gender_genetic_pos0;
      gender_genetic_pos[1] = gender_genetic_pos1;
    }

    void WriteBin(ostream &out) const;
    void ReadBin(istream &in);
  };

  vector<Locus> loci;

  int32 sry_genomic_pos;
  int32 x_y_par2_genomic_pos_delta;

  // Chromosome data

  struct Chrom
  {
    string label;
    Segment range;

    Chrom() { }
    Chrom(const string &label_, const Segment &range_)
      : label(label_), range(range_) { }

    void WriteBin(ostream &out) const;
    void ReadBin(istream &in);
  };

  vector<Chrom> chroms;

  Segment par1_range;
  Segment par2_range;
  Segment x_nonpar_range;
  Segment y_nonpar_range;
  Segment mito_range;

  /*** Derived Data ***/

  // Data sizes

  int seq_len, num_chroms;

  // SNP/location data

  int sry_pos;

  vector<int> locus_chrom;
  vector<int32> locus_genomic_pos;
  vector<double> locus_genetic_pos;
  vector< vector<double> > gender_locus_genetic_pos;

  // Chromosome data

  int sex_chrom;
  Segment sex_range;
  
  /*** Methods ***/

  SNPProfile();
  SNPProfile(const string &filename);
  void Copy(const SNPProfile &init);
  void CopyRestrictLoci(const SNPProfile &init, const vector<int> &select);
  void CopyRestrictChrom(const SNPProfile &init, int chrom);
  void Clear();
  void Swap(SNPProfile &other);

  void Validate() const;
  void ComputeDerived();
  uint32 ComputeHash() const;
  bool Compatible(const SNPProfile &other) const;

  void ReadBin(istream &in);
  void ReadBin(const string &filename);
  void WriteBin(ostream &out) const;
  void WriteBin(const string &filename) const;
};


template <>
void Write<SNPProfile>(ostream &o, const SNPProfile &x);

template <>
void Read<SNPProfile>(istream &i, SNPProfile &x);


class SNPData
{
public:
  static bool status_output;
  static uint32 magic;
  static vector<double> switch_code;

  // SNP Profile

  const SNPProfile *profile;
  int seq_len, num_chroms;

  const SNPProfile::Locus *loci;
  const SNPProfile::Chrom *chroms;
  const int *locus_chrom;
  const int32 *locus_genomic_pos;
  const double *locus_genetic_pos;
  const vector<double> *gender_locus_genetic_pos;
  int sry_pos;

  /*** Fundamental Data ***/

  // Population data
  
  struct Pop
  {
    string label;

    vector<int> haplos; //Derived
    vector<int> diplos; //Derived

    Pop() { }
    Pop(const string &label_) : label(label_) { }

    void WriteBin(ostream &out) const;
    void ReadBin(istream &in);

    int haplo_index(int k) const;
    int diplo_index(int d) const;
  };

  vector<Pop> pops;

  // Haplotype data

  struct Haplo
  {
    string label;
    int pop;
    int sry;
    vector<int8> data;

    int diplo; //Derived
    int diplo_pair; //Derived

    Haplo() { }
    Haplo(const string &label_, int pop_, int sry_, const vector<int8> &data_)
      : label(label_), pop(pop_), sry(sry_), data(data_) { }

    const int8 &operator[](int i) const { return data[i]; }
    int8 &operator[](int i) { return data[i]; }

    void WriteBin(ostream &out) const;
    void ReadBin(istream &in);
  };

  vector<Haplo> haplos;

  // Diploid data

  struct Diplo
  {
    string label;
    int haplos[2];
    vector<uint8> switches;

    int pop; //Derived
    int gender; //Derived

    Diplo() { }
    Diplo(const string &label_, int haplo0, int haplo1, const vector<uint8> &switches_)
      : label(label_), switches(switches_)
    {
      haplos[0] = haplo0; haplos[1] = haplo1;
    }

    void WriteBin(ostream &out) const;
    void ReadBin(istream &in);
  };

  vector<Diplo> diplos;

  /*** Derived Data ***/

  // Data sizes

  int num_pops, num_haplos, num_diplos;

  // Haplotype data

  vector<int> haplo_pop;
  vector<const int8 *> haplo_data;

  // Diplod data

  vector<int> diplo_pop;
  vector<const uint8 *> diplo_switches;

  /*** Methods ***/

  SNPData();
  SNPData(const SNPProfile &profile_);
  SNPData(const SNPProfile &profile_, const string &filename);
  SNPData(const SNPData &init, bool copy_data = true);
  void SetProfile(const SNPProfile &profile_);
  operator const SNPProfile &() const { return *profile; }
  void Copy(const SNPData &init, bool copy_data = true);
  void CopyRestrictLoci(const SNPData &init, const vector<int> &select);
  void CopyRestrictChrom(const SNPData &init, int chrom);
  void CopyRestrictHaplos(const SNPData &init, const vector<int> &select);
  void CopyRestrictDiplos(const SNPData &init, const vector<int> &select);
  void CopyRestrictPops(const SNPData &init, const vector<int> &select);
  void CopyRestrictPops(const SNPData &init, uint64 pop_mask);
  void CopyRestrictPop(const SNPData &init, int pop);

  void Clear();
  void Swap(SNPData &other);
  void ClearDiploPairing();
  void ReserveHaplos(int num_haplos) { haplos.reserve(num_haplos); }
  void ReserveDiplos(int num_diplos) { diplos.reserve(num_diplos); }

  void Validate() const;
  void ComputeDerived();
  bool Compatible(const SNPData &other) const;

  void ReadBin(istream &in);
  void ReadBin(const string &filename);
  void WriteBin(ostream &out) const;
  void WriteBin(const string &filename) const;

  static double switch_decode(uint8 value)
  { 
    return switch_code[value];
  }

  static uint8 switch_encode(double value)
  {
    AssertMsg(value >= 0.0, ToStr(value));
    if (value == 0.0)
      return 0;
    if (value >= 0.5)
      return uint8(255);
    int v = 255 - int(ceil(-10.0 * log(2.0 * value)));
    if (fabs(switch_code[v] - value) < fabs(switch_code[v + 1] - value))
      return v;
    else
      return v + 1;
  }

  void SetDiploSwitch(const vector<int8> &haplo0, const vector<int8> &haplo1,
                      int gender, vector<uint8> &switches, uint8 value) const;
  void SetDiploSwitch(const vector<int8> &haplo0, const vector<int8> &haplo1,
                      int gender, vector<uint8> &switches, 
                      double switch_error_rate = 0.5) const
  {
    SetDiploSwitch(haplo0, haplo1, gender, switches, 
                   switch_encode(switch_error_rate));
  }

  void SetDiploSwitch(int d, uint8 value);
  void SetDiploSwitch(int d, double switch_error_rate = 0.5)
  {
    SetDiploSwitch(d, switch_encode(switch_error_rate));
  }

  void SetAllDiploSwitch(uint8 value);
  void SetAllDiploSwitch(double switch_error_rate = 0.5)
  {
    SetAllDiploSwitch(switch_encode(switch_error_rate));
  }

  void RestrictHaplos(const vector<int> &select);
  void RestrictDiplos(const vector<int> &select);
  void RestrictPops(const vector<int> &select);
  void RestrictPops(uint64 pop_mask);
  void RestrictPop(int pop);
  void MapPops(const vector<int> &pop_map);

  // call ComputeDerived() afterward:
  int AddHaplo(const vector<int8> &data, int pop, 
               const string &label = "", int sry = 0);
  int AddHaplo(const Haplo &haplo, 
               int pop = INT_MIN, const string &label = "~");
  int AddHaplo(const SNPData &data, int k, 
               int pop = INT_MIN, const string &label = "~");
  int AddDiplo(const vector<int8> &haplo0, const vector<int8> &haplo1,
               const vector<uint8> &diplo_switch, int pop,
               const string &label = "", int sry0 = 0, int sry1 = 0);
  int AddDiplo(const vector<int8> &haplo0, const vector<int8> &haplo1,
               const vector<double> &diplo_switch, int pop,
               const string &label = "", int sry0 = 0, int sry1 = 0);
  int AddDiplo(const vector<int8> &haplo0, const vector<int8> &haplo1,
               double switch_error_rate, int pop, 
               const string &label = "", int sry0 = 0, int sry1 = 0);
  int AddDiplo(const SNPData &data, int d,
               int pop = INT_MIN, const string &label = "~");

  void Concat(const SNPData &other);
  void MergePopIndex(const SNPData &other);
  void MergePopLabel(const SNPData &other);

private:
  void Init();
  void SetProfile(const SNPProfile *profile_);
};


template <>
void Write<SNPData>(ostream &o, const SNPData &x);

template <>
void Read<SNPData>(istream &i, SNPData &x);


void ReadBinSNPProfileData(istream &in, SNPProfile &profile, SNPData &data);
void ReadBinSNPProfileData(const string &filename, 
                           SNPProfile &profile, SNPData &data);
void WriteBinSNPProfileData(ostream &out, const SNPData &data);
void WriteBinSNPProfileData(const string &filename, const SNPData &data);


#endif
