
#ifndef SIMULATE_H
#define SIMULATE_H

#include "System.h"
#include "Segment.h"
#include "SNPData.h"

namespace Simulate
{

  class HaploPop;

  struct Haplo
  {
    vector< pair<Segment, int> > sources;

    Haplo() { }
    Haplo(const SNPData &data, int k) { Set(data, k); }

    void Set(const SNPData &data, int k);
    void Sample(const SNPData &data);
    void Sample(const SNPData &data, const vector<int> &select);
    void SamplePop(const SNPData &data, int pop);
    void SamplePops(const SNPData &data, const vector<int> &select);
    void SamplePops(const SNPData &data, uint64 pop_mask);

    void SetFullTree(const SNPProfile &profile, int num_gens);
    /* SetFullTree: resulting indices go from 0 ... 2^num_gens - 1 */

    void Composite(const HaploPop &pop, const Haplo &haplo);
    void Composite(const vector<HaploPop> &pops, const Haplo &haplo);
    void MapSources(const vector<int> &mapping);
    void MapSources(const map<int, int> &mapping);

    bool empty() const { return sources.empty(); }
    int size() const { return int(sources.size()); }
    int left() const { Assert(!sources.empty()); return sources.front().first.left(); }
    int right() const { Assert(!sources.empty()); return sources.back().first.right(); }
    int length() const { return right() - left(); }
    void clear() { sources.clear(); }
    void Free() { vector< pair<Segment, int> > empty; sources.swap(empty); }
    void swap(Haplo &other) { sources.swap(other.sources); }
    void WriteBin(ostream &o) const;
    void ReadBin(istream &i);
    void DebugDisplay(ostream &o) const;
    void RestrictHaplo(const vector<int> &select, vector<int> &result_map);

    const pair<Segment, int> &operator[](int i) const { return sources[i]; }

    pair<Segment, int> &operator[](int i) { return sources[i]; }

    template <typename VT, typename T>
    void Generate(const vector<VT> &templates, vector<T> &result) const
    {
      if (sources.empty() || (sources[0].first.left() != 0))
      {
        cerr << "Sources:" << endl;
        for (int i = 0; i < sources.size(); ++i)
          cerr << "[" << sources[i].first.left() << ", " << sources[i].first.right() << ") => " << sources[i].second << endl;
      }
      Assert(!sources.empty() & (sources[0].first.left() == 0));
      result.resize(sources.back().first.right());
      for (int i = 0; i < sources.size(); ++i)
        for (int p = sources[i].first.left(); p < sources[i].first.right(); ++p)
          result[p] = T(templates[sources[i].second][p]);
    }

    template <typename T>
    void Generate(const SNPData &data, vector<T> &result) const
    {
      Generate(data.haplo_data, result);
      AssertMsg(sources.back().first.right() == data.seq_len, ToStr(sources.back().first.right()) + " != " + ToStr(data.seq_len));
    }

    template <typename T>
    void Label(vector<T> &result) const
    {
      Assert(!sources.empty() && (sources[0].first.left() == 0));
      result.resize(sources.back().first.right());
      for (int i = 0; i < sources.size(); ++i)
        for (int p = sources[i].first.left(); p < sources[i].first.right(); ++p)
          result[p] = sources[i].second;
    }

    template <typename VT, typename T>
    void Label(const vector<VT> &mapping, vector<T> &result) const
    {
      Assert(!sources.empty() && (sources[0].first.left() == 0));
      result.resize(sources.back().first.right());
      for (int i = 0; i < sources.size(); ++i)
        for (int p = sources[i].first.left(); p < sources[i].first.right(); ++p)
          result[p] = T(mapping[sources[i].second]);
    }

    template <typename T>
    void LabelPop(vector<T> &result) const
    {
      Assert(!sources.empty() && (sources[0].first.left() == 0));
      result.resize(sources.back().first.right());
      for (int i = 0; i < sources.size(); ++i)
      {
        T value = T(sources[i].second >> 16);
        for (int p = sources[i].first.left(); p < sources[i].first.right(); ++p)
          result[p] = value;
      }
    }      

    template <typename T>
    void LabelPop(const SNPData &data, vector<T> &result) const
    {
      Label(data.haplo_pop, result);
      AssertMsg(sources.back().first.right() == data.seq_len, ToStr(sources.back().first.right()) + " != " + ToStr(data.seq_len));
    }

    template <typename T>
    void Set(const vector<T> &sources_)
    {
      sources.clear();
      if (sources_.empty())
        return;
      
      int last_source = int(sources_[0]);
      int last_start = 0;
      for (int i = 1; i < sources_.size(); ++i)
      {
        if (sources_[i] != last_source)
        {
          sources.push_back(pair<Segment, int>(Segment(last_start, i, true), last_source));
          last_source = int(sources_[i]);
          last_start = i;
        }
      }
      
      sources.push_back(pair<Segment, int>(Segment(last_start, sources_.size(), true), last_source));
    }

  private:
    void SetFullTreeRecurse(const SNPProfile &profile, int gens,
                            const vector<Segment> &restrict,
                            vector< pair<Segment, int> > &sources,
                            int index = 0);
  };


  class DiploPop;

  struct Diplo
  {
    Haplo haplos[2];
    int haplo_sry[2];
    int gender;
    int pop;
    int idx;  //ID in SNPData
    Diplo() { haplo_sry[0] = haplo_sry[1] = 0; SetGender(); }

    Diplo(const SNPData &data, int d)
    { 
      Set(data, d);
    }

    void SetGender();
    void Set(const SNPData &data, int d);
    void Sample(const SNPData &data);
    void Sample(const SNPData &data, const vector<int> &select);
    void SamplePop(const SNPData &data, int pop);
    void SamplePops(const SNPData &data, const vector<int> &select);
    void SamplePops(const SNPData &data, uint64 pop_mask);

    void SetFullTree(const SNPProfile &profile, int num_gens, int gender = -1);
    /* SetFullTree: resulting indices go from 0 .. 2^(num_gens + 1)-1 
       and are laid out as 
       <female0_haplo0_mito, female0_haplo1, male1_haploX, male1_haploY,
        female2_haplo0, female2_haplo1, male3_haploX, male3_haploY,
        ...
        male(2^num_gens-1)_haploX, male(2^num_gens-1)_haploY_transmitted> */

    void Composite(const DiploPop &pop, const Diplo &diplo);
    void Composite(const vector<DiploPop> &pops, const Diplo &diplo);
    void MapSources(const vector<int> &mapping);
    void MapSources(const map<int, int> &mapping);
	 void RestrictDiplo(const vector<int> &select, vector<int> &result_map);
    bool empty() const { return haplos[0].empty() && haplos[1].empty(); }
    void clear();
    void Free();
    void swap(Diplo &other);
    void WriteBin(ostream &o) const;
    void ReadBin(istream &i);

    const Haplo &operator[](int index) const
    { 
      Assert((index >= 0) && (index <= 1));
      return haplos[index];
    }

    Haplo &operator[](int index) 
    { 
      Assert((index >= 0) && (index <= 1));
      return haplos[index];
    }

    template <typename VT, typename T>
    void Generate(const vector<VT> &templates, 
                  vector<T> &result0, vector<T> &result1) const
    {
      haplos[0].Generate(templates, result0);
      haplos[1].Generate(templates, result1);
    }

    template <typename T>
    void Generate(const SNPData &data, 
                  vector<T> &result0, vector<T> &result1) const
    {
      haplos[0].Generate(data, result0);
      haplos[1].Generate(data, result1);
    }

    template <typename T>
    void Label(vector<T> &result0, vector<T> &result1) const
    {
      haplos[0].Label(result0);
      haplos[1].Label(result1);
    }

    template <typename VT, typename T>
    void Label(const vector<VT> &mapping, 
               vector<T> &result0, vector<T> &result1) const
    {
      haplos[0].Label(mapping, result0);
      haplos[1].Label(mapping, result1);
    }

    template <typename T>
    void LabelPop(vector<T> &result0, vector<T> &result1) const
    {
      haplos[0].LabelPop(result0);
      haplos[1].LabelPop(result1);
    }

    template <typename T>
    void LabelPop(const SNPData &data, 
                  vector<T> &result0, vector<T> &result1) const
    {
      haplos[0].LabelPop(data, result0);
      haplos[1].LabelPop(data, result1);
    }

  private:
    void SetFullTreeRecurse(const SNPProfile &profile, int gens,
                            const vector<Segment> &restrict,
                            vector< pair<Segment, int> > &sources,
                            int gender, int index);
  };


  void GenerateRecombs(const SNPProfile &profile, vector<int> &recombs, 
                       int gender = -1);
  int GenerateRecombs(const SNPProfile &profile, 
                      vector<Segment> &x, vector<Segment> &y, int gender = -1);


  void RecombineHaplos(const SNPProfile &profile, 
                       const Haplo &x, const Haplo &y, Haplo &result);

  void MateDiplos(const SNPProfile &profile, const Diplo &x, const Diplo &y, 
                  Diplo &result, int gender = -1);


  class HaploPop
  {
  public:
    typedef Haplo SampleType;

    const SNPData *data;
    vector<Haplo> haplos;

    HaploPop() { data = NULL; }

    HaploPop(const SNPData &data_)
    { 
      SetSNPData(data_); 
    }

    void clear() { haplos.clear(); }
    int size() const { return int(haplos.size()); }
    void resize(int size) { haplos.resize(size); }
    void reserve(int size) { haplos.reserve(size); }
    void swap(HaploPop &other);

    const Haplo &operator[](int i) const { return haplos[i]; }
    Haplo &operator[](int i) { return haplos[i]; }

    void SetSNPData(const SNPData &data_);
    void SetAll();
    void SetPop(int pop);
    void SetPops(const vector<int> &select);
    void SetPops(uint64 pop_mask);
    void Set(const vector<int> &select);

    void Add(const Haplo &haplo) { haplos.push_back(haplo); }

    void Derive(Haplo &result) const;
    void DerivePop(const HaploPop &sources, int num_samples);

    void MapSources(const vector<int> &mapping);
    void MapSources(const map<int, int> &mapping);

    int CompactSources();

    void CreateSNPData(SNPData &result);

    void WriteBin(ostream &o) const;
    void WriteBin(const string &filename) const;
    void ReadBin(istream &i);
    void ReadBin(const string &filename);
  };

  class DiploPop
  {
  public:
    typedef Diplo SampleType;

    const SNPData *data;
    vector<Diplo> diplos;

    DiploPop() { data = NULL; }

    DiploPop(const SNPData &data_)
    {
      SetSNPData(data_);
    }

    void clear() { diplos.clear(); }
    int size() const { return int(diplos.size()); }
    void resize(int size) { diplos.resize(size); }
    void reserve(int size) { diplos.reserve(size); }
    void swap(DiploPop &other);

    const Diplo &operator[](int i) const { return diplos[i]; }
    Diplo &operator[](int i) { return diplos[i]; }

    void SetSNPData(const SNPData &data_);
    void SetAll();
    void SetPop(int pop);
    void SetPops(const vector<int> &select);
    void SetPops(uint64 pop_mask);
    void Set(const vector<int> &select);

    void Add(const Diplo &diplo) { diplos.push_back(diplo); }

    void Derive(Diplo &result) const;
    void DerivePop(const DiploPop &sources, int num_samples);

    void MapSources(const vector<int> &mapping);
    void MapSources(const map<int, int> &mapping);

    int CompactSources();

    void CreateSNPData(SNPData &result);
    void CreateHaploPop(HaploPop &result);

    void WriteBin(ostream &o) const;
    void WriteBin(const string &filename) const;
    void ReadBin(istream &i);
    void ReadBin(const string &filename);
  };



  class GraphColoring
  {
  public:
    struct Vertex
    {
      int color;
      vector<int> adj;
    };

    struct VertexDegreeDecreasing
    {
      bool operator()(const Vertex &x, const Vertex &y) const
      {
        return x.adj.size() > y.adj.size();
      }
    };

    vector<Vertex> vertices;

    void Set(const vector<Haplo> &haplos, map<int, int> &source_to_vertex);

    int GreedyColor();
  };

};


template <>
void Write<Simulate::Haplo>(ostream &o, const Simulate::Haplo &x);
template <>
void Read<Simulate::Haplo>(istream &i, Simulate::Haplo &x);

template <>
void Write<Simulate::Diplo>(ostream &o, const Simulate::Diplo &x);
template <>
void Read<Simulate::Diplo>(istream &i, Simulate::Diplo &x);

template <>
void Write<Simulate::HaploPop>(ostream &o, const Simulate::HaploPop &x);
template <>
void Read<Simulate::HaploPop>(istream &i, Simulate::HaploPop &x);

template <>
void Write<Simulate::DiploPop>(ostream &o, const Simulate::DiploPop &x);
template <>
void Read<Simulate::DiploPop>(istream &i, Simulate::DiploPop &x);


#endif
