#include "RelationshipModel.h"

TwoMRCAState::TwoMRCAState(bool inSharedMRCA0, bool inSharedMRCA1, 
			   bool inPassedA, bool inPassedB,
			   bool inDonationsA, bool inDonationsB,
			   bool inPhaseA, bool inPhaseB) : 
  RelationshipHMMState::RelationshipHMMState(), 
  sharedMRCA0(inSharedMRCA0), sharedMRCA1(inSharedMRCA1), 
  passedA(inPassedA), passedB(inPassedB), 
  phaseA(inPhaseA), phaseB(inPhaseB),
  donationsA(inDonationsA), donationsB(inDonationsB), idx(-1),
  variables(NUM_VAR_2MRCA){}

TwoMRCAState::TwoMRCAState(const TwoMRCAState &copy) : 
  RelationshipHMMState::RelationshipHMMState(static_cast<RelationshipHMMState>(copy)),
  sharedMRCA0(copy.sharedMRCA0), sharedMRCA1(copy.sharedMRCA1),
  passedA(copy.passedA), passedB(copy.passedB), 
  phaseA(copy.phaseA), phaseB(copy.phaseB),
  donationsA(copy.donationsA), donationsB(copy.donationsB), 
  idx(copy.idx), variables(NUM_VAR_2MRCA){
  //I won't update the variables until I need them.
  //updateVariables();
}

TwoMRCAState& TwoMRCAState::operator=(const TwoMRCAState &that){
  if(this != &that){
    RelationshipHMMState::operator=(that);
    sharedMRCA0 = that.sharedMRCA0;
    sharedMRCA1 = that.sharedMRCA1;
    passedA = that.passedA;
    passedB = that.passedB;
    phaseA = that.phaseA;
    phaseB = that.phaseB;
    donationsA = that.donationsA;
    donationsB = that.donationsB;
    idx = that.idx;
    variables.resize(NUM_VAR_2MRCA);
    updateVariables();
  }
  return *this;
}

TwoMRCAState::~TwoMRCAState(){
  TwoMRCAState::clearObj();
}

void TwoMRCAState::updateVariables(){
  variables[SHARED0_IDX] = &sharedMRCA0;
  variables[SHARED1_IDX] = &sharedMRCA1;
  variables[PASSEDA_IDX] = &passedA;
  variables[PASSEDB_IDX] = &passedB;
  variables[PHASEA_IDX] = &phaseA;
  variables[PHASEB_IDX] = &phaseB;
  variables[DONA_IDX] = &donationsA;
  variables[DONB_IDX] = &donationsB;
}

string TwoMRCAState::getBinaryString(){
  stringstream stream("");
  for(int i = 0; i < NUM_VAR_2MRCA; ++i)
    stream << (int)(*variables[i]);
  return stream.str();
}

void TwoMRCAState::print(ostream &os) const{
  os << "State " << idx << " (shared0 = " << sharedMRCA0 << ", ";
  os << "shared1 = " << sharedMRCA1 << ", ";
  os << "passA = " << passedA << ", ";
  os << "passB = " << passedB << ", ";
  os << "donA = " << donationsA << ", ";
  os << "donB = " << donationsB << ", ";
  os << "phaseA = " << phaseA << ", ";
  os << "phaseB = " << phaseB << ", ";
  os << "numCorr = " << numCorr << ", ";
  os << "corrA = " << corrA << ", ";
  os << "corrB = " << corrB << ", ";
  os << "logInitPr = " << logInitialProb << ")" << endl;
}

TwoMRCAModel::TwoMRCAModel(SNPData &inData, int inIdxA, int inIdxB,
			   vector< vector<double> > &inLetterFreq, 
			   vector< vector<double> > &inErrorFunction,
			   vector< vector<int> > &inPascalTri,
			   vector< vector<double> > &inLogRecomb,
			   vector< vector<double> > &inLogNoRecomb,  
			   int inNumGensA, int inNumGensB) :
  RelationshipHMM::RelationshipHMM(inData, inIdxA, inIdxB, inLetterFreq, 
				   inErrorFunction, inPascalTri,
				   inLogRecomb, inLogNoRecomb, 
				   inNumGensA, inNumGensB),
  maxValues(NUM_VAR_2MRCA, 1){
  validateNumGens(numGensA, numGensB);
  //passedA is redundant and is always set to false if numGensA < 1.
  maxValues[PASSEDA_IDX] = (int)(numGensA > 0);
  maxValues[PASSEDB_IDX] = (int)(numGensB > 0);
  //donationsA is always set to false if numGensA < 2.
  maxValues[DONA_IDX] = (int)(numGensA > 1);
  maxValues[DONB_IDX] = (int)(numGensB > 1);
  initStates();
}

/*
 * Creates shallow copies of the pointers (e.g. data), but deep
 * copies of the non-pointer vectors (e.g. logRecomb). 
 */
TwoMRCAModel::TwoMRCAModel(const TwoMRCAModel &copy) : 
  RelationshipHMM::RelationshipHMM(static_cast<RelationshipHMM>(copy)),
  maxValues(copy.maxValues), states(copy.states){
  //cerr << "In TwoMRCAModel copy constructor" << endl;
  validateNumGens(numGensA, numGensB);
}

TwoMRCAModel::~TwoMRCAModel(){
  TwoMRCAModel::clearObj();
}

/*
 * Like the copy constructor, the assignment operator returns shallow 
 * copies of the pointers (e.g. data), but deep
 * copies of the non-pointer vectors (e.g. logRecomb). 
 */
TwoMRCAModel& TwoMRCAModel::operator=(const TwoMRCAModel &that){
  if(this != &that){
    RelationshipHMM::operator=(that);
    maxValues = that.maxValues;
    states = that.states;
  }
  return *this;
}

/*
 * Return value:
 * An index (or ID) for the given state.
 * The states for a given level will be assigned consecutive integer
 * indices starting from 0.
 * The value returned by this function is also the index of the state
 * in the model's "states" vector. 
 * Since each variable is a bool (i.e. a bit), the state index
 * is in fact the decimal corresponding to the binary number
 * represented by the state's variables.
 */
int TwoMRCAModel::getStateIdx(const TwoMRCAState &state){
  int pos = 0;
  int idx = 0;
  for(int var = 0; var < NUM_VAR_2MRCA; ++var){
    //This will be either 0 (if that variable has the value 0)
    //or 2^pos (if that variable has the value 1).
    idx += (1 << pos) * (*(state.variables[var]));
    pos += maxValues[var];
  }
  return idx;
}

/*
 * Returns the index of the state that is similar to the 
 * given state, but has the value of the "var" variable
 * reversed.
 * var is the index of the variable (from 0 to NUM_VAR_2MRCA - 1)
 */
int TwoMRCAModel::getInverseStateIdx(const TwoMRCAState& state, int var){
  int val = *(state.variables[var]);
  //The position of the var-th variable in the binary representation
  //is not fixed, but depends on how many of the previous variables
  //are redundant.
  int sumMax = 0;
  for(int v = 0; v < var; ++v)
    sumMax += maxValues[v];
  int multiplier = 1 << sumMax;
  //mult * (1 - val) - mult * val = mult - mult * val - mult * val = 
  //mult - 2 * mult * val = mult * (1 - 2 * val) = mult * (1 - (val << 1))
  return state.idx + multiplier * (1 - (val << 1));
}

/*
 * Return value:
 * the model's total number of states.
 */
int TwoMRCAModel::getNumStates(){
  int sumMax = 0;
  for(int var = 0; var < NUM_VAR_2MRCA; ++var)
    sumMax += maxValues[var];
  return 1 << sumMax;
}

void TwoMRCAModel::getBinaryStates(vector<string> &binStr){
  binStr.clear();
  //I know that the states vector has state with index idx
  //at the idx-th position.
  for(int i = 0; i < states.size(); ++i){
    binStr.push_back(states[i].getBinaryString());
  }
}

/*
 * Initializes the model's vector of states. It also computes the 
 * derived variables of the states.
 */
void TwoMRCAModel::initStates(){
  states.clear();
  states.resize(getNumStates());
  for(int m0 = 0; m0 < 2; ++m0){
    for(int m1 = 0; m1 < 2; ++m1){
      for(int pA = 0; pA < 2; ++pA){
	for(int pB = 0; pB < 2; ++pB){
	  for(int passedA = 0; passedA <= maxValues[PASSEDA_IDX]; ++passedA){
	    for(int passedB = 0; passedB <= maxValues[PASSEDB_IDX]; ++passedB){
	      for(int donA = 0; donA <= maxValues[DONA_IDX]; ++donA){
		for(int donB = 0; donB <= maxValues[DONB_IDX]; ++donB){
		  TwoMRCAState state((bool)m0, (bool)m1, 
				     (bool)passedA, (bool)passedB, 
				     (bool)donA, (bool)donB, 
				     (bool)pA, (bool)pB);
		  state.updateVariables();
		  computeCorrelation(state);
		  computeLogInitialProb(state);
		  int stateIdx = getStateIdx(state);
		  state.idx = stateIdx;
		  states[stateIdx] = state;
		}
	      }
	    }
	  }
	}
      }
    }
  }
}

/*
 * Computes the variables numCorr, corrA, and corrB of the given state,
 * based on the state's "core" variables.
 */
void TwoMRCAModel::computeCorrelation(TwoMRCAState &state){
  validateState(state);
  state.numCorr = 0;
  //Default value for corr variables when there are no correlated alleles is 0.
  state.corrA = 0;
  state.corrB = 0;
  if(!state.donationsA && !state.donationsB){
    state.numCorr += (int)(state.sharedMRCA0 && (!state.passedA || numGensA == 0) &&
			   (!state.passedB || numGensB == 0));
    state.numCorr += (int)(state.sharedMRCA1 && (state.passedA || numGensA == 0) && 
			   (state.passedB || numGensB == 0));
    //If there are correlated alleles...
    if(state.numCorr > 0){
      state.corrA = state.phaseA;
      state.corrB = state.phaseB;
      if(state.numCorr == 1 && state.sharedMRCA1){
	if(numGensA == 0 && (state.passedB || numGensB == 0))
	  state.corrA = 1 - state.phaseA;
	if(numGensB == 0 && (state.passedA || numGensA == 0))
	  state.corrB = 1 - state.phaseB;
      }
    }
  }
}

void TwoMRCAModel::validateState(const TwoMRCAState &state) const{
  AssertMsg(!state.passedA || numGensA > 0, "Invalid state");
  AssertMsg(!state.passedB || numGensB > 0, "Invalid state");
  AssertMsg(!state.donationsA || numGensA > 1, "Invalid state");
  AssertMsg(!state.donationsB || numGensB > 1, "Invalid state");
}

/*
 * Computes the log of the initial probability of the state.
 * The probability can be computed as:
 * P(this state is the initial state) = 
 * P(phaseA)*P(phaseB)*P(sharedMRCA0)*P(sharedMRCA1)*P(passedA)*
 * P(passedB)*P(donationsA)*P(donationsB) = 
 * (1/16)*P(passedA)*P(passedB)*P(donationsA)*P(donationsB)
 * The probabilities P(donationsA) and P(donationsB) depend on the 
 * number of generations in the lineage of A and B respectively.
 * 
 * The probabilities are computed in log space.
 */
void TwoMRCAModel::computeLogInitialProb(TwoMRCAState &state){
  validateState(state);

  double logPassedA = (numGensA > 0)? -log(2.0): 0.0;
  double logPassedB = (numGensB > 0)? -log(2.0): 0.0;
  double logDonA = 0.0, logDonB = 0.0;
  if(numGensA > 1){
    //numGensA is the total number of generations between the MRCAs and A,
    //but donationsA only refers to the generations after the grandchildren of 
    //the MRCAs, i.e. one generation less than numGensA.
    logDonA = (!state.donationsA)? log(2.0) * (-numGensA + 1): log(1.0 - pow(2.0, -numGensA + 1)); 
  }
  if(numGensB > 1)
    logDonB = (!state.donationsB)? log(2.0) * (-numGensB + 1): log(1.0 - pow(2.0, -numGensB + 1)); 
  state.logInitialProb = -log(16.0) + logPassedA + logPassedB + logDonA + logDonB;
}

void TwoMRCAModel::print(ostream &os){
  os << "TwoMRCAModel: numGensA = " << numGensA << ", numGensB = " << 
    numGensB << ", idxA = " << idxA << ", idxB = " << idxB << endl;
  os << "Model's pascal triangle:" << endl;
  printGenericMatrix(*pascalTri, os);
  os << "Model's 10 first log-recomb probabilities:" << endl;
  printGenericMatrix(*logRecomb, os, 10);
  os << "Model's 10 first log-no-recomb probabilities:" << endl;
  printGenericMatrix(*logNoRecomb, os, 10);

  os << "Model's states (" << states.size() << "): " << endl;
  for(int i = 0; i < states.size(); ++i){
    os << i << "\t";
    states[i].print(os);
  }
}

/*
 * Return value:
 * the log-probability of transitioning from state currentState at position pos
 * to state nextState at position pos + 1, under the current model.
 * 
 * If the transition is "impossible," log(ZERO) is returned.
 */
double TwoMRCAModel::transitionProb(const TwoMRCAState& currentState, 
				    const TwoMRCAState& nextState, 
				    int pos, bool debug){
  validateState(currentState);
  validateState(nextState);
  AssertMsg(pos >= 0 && pos < (*data).seq_len - 1, "pos out of data range");
  AssertMsg(currentState.numCorr > -1, "uninitialized state");
  AssertMsg(nextState.numCorr > -1, "uninitialized state");

  if((*data).locus_chrom[pos] != (*data).locus_chrom[pos + 1]){
    if(debug)
      cerr << "In transitionProb: transitioning between states, log-prob = "
	   << nextState.logInitialProb << endl;
    return nextState.logInitialProb;//(nextState.logInitialProb < log(ZERO))? log(ZERO): nextState.logInitialProb;
  }

  double logProb = 0.0;  
  //The probability of a switch between positions pos and pos + 1 in individual A.
  double switchProb = SNPData::switch_decode((*data).diplos[idxA].switches[pos]);
  //if(switchProb < ZERO) switchProb = ZERO;
  if(debug){
    cerr << "In transitionProb: switchesA = " << (int)(*data).diplos[idxA].switches[pos] << 
      ", switchProbA = " << switchProb;
  }
  logProb += (currentState.phaseA == nextState.phaseA)? 
    correctLog(1.0 - switchProb): correctLog(switchProb);

  switchProb = SNPData::switch_decode((*data).diplos[idxB].switches[pos]);
  //if(switchProb < ZERO) switchProb = ZERO;
  if(debug){
    cerr << ", switchesB = " << (int)(*data).diplos[idxB].switches[pos] << 
      ", switchProbB = " << switchProb << endl;
  }
  logProb += (currentState.phaseB == nextState.phaseB)? 
    correctLog(1.0 - switchProb): correctLog(switchProb);

  if(debug){
    cerr << "In transitionProb: logRecomb[pos][1] = " <<  (*logRecomb)[pos][1] << 
      ", logRecomb[pos][2] = " << (*logRecomb)[pos][2] << endl;
    cerr << "In transitionProb: logNoRecomb[pos][1] = " <<  (*logNoRecomb)[pos][1] << 
      ", logNoRecomb[pos][2] = " << (*logNoRecomb)[pos][2] << endl;
  }

  //There are two ways in which the immediate descendants of the MRCAs (call them 
  //individual C and individual D) can share the same haplotype from MRCA0 at 
  //both position pos and position pos + 1 (or can have a different haplotype from
  //MRCA0 both at pos and at pos + 1): Either no recombination occured
  //between pos and pos + 1 in MRCA0's gametes for both C and D (eg. both C and D 
  //got the maternal haplotype of MRCA0 at pos and at pos + 1), or a recombination occured
  //in both C and D (eg. both C and D got the maternal haplotype of MRCA0 at pos
  //and the paternal haplotype of MRCA0 at pos + 1). Same for MRCA1.
  //See Durbin et al. p. 78 for the log of a sum of logs.
  double logShared0Prob = (currentState.sharedMRCA0 == nextState.sharedMRCA0)?
    computeLogSumOfTwo((*logRecomb)[pos][2], (*logNoRecomb)[pos][2]):
    log(2.0) + (*logRecomb)[pos][1] + (*logNoRecomb)[pos][1];
  double logShared1Prob = (currentState.sharedMRCA1 == nextState.sharedMRCA1)?
    computeLogSumOfTwo((*logRecomb)[pos][2], (*logNoRecomb)[pos][2]):
    log(2.0) + (*logRecomb)[pos][1] + (*logNoRecomb)[pos][1];
  logProb += (logShared0Prob + logShared1Prob);

  if(debug)
    cerr << "In transitionProb: log-prob for sharedMRCA0 = " << logShared0Prob
	 << ", log-prob for sharedMRCA1 = " << logShared1Prob << endl;

  double logPassedA = 0.0, logPassedB = 0.0;
  if(numGensA > 0)
    logPassedA = (currentState.passedA == nextState.passedA)? 
      (*logNoRecomb)[pos][1]: (*logRecomb)[pos][1];
  if(numGensB > 0)
    logPassedB = (currentState.passedB == nextState.passedB)? 
      (*logNoRecomb)[pos][1]: (*logRecomb)[pos][1];
  if(debug)
    cerr << "In transitionProb: log-prob for passedA = " << logPassedA
	 << ", log-prob for passedB = " << logPassedB << endl;

  logProb += (logPassedA + logPassedB);

  double logDonA = donationsTransitionProb(currentState.donationsA, 
					   nextState.donationsA, 
					   pos, numGensA - 1, debug);
  double logDonB = donationsTransitionProb(currentState.donationsB, 
					   nextState.donationsB, 
					   pos, numGensB - 1, debug);
  logProb += (logDonA + logDonB);

  if(debug)
    cerr << "In transitionProb: log-prob for donationsA = " << logDonA 
	 << ", log-prob for donationsB = " << logDonB << endl
	 << "In transitionProb: Final log-prob = " << logProb << endl; 
  
  return logProb;//((logProb < log(ZERO))? log(ZERO): logProb);
}

void TwoMRCAModel::sharedMRCAprobs(int pos, vector< vector< double> > &probs){
  probs[0][0] = probs[1][1] = computeLogSumOfTwo((*logRecomb)[pos][2], (*logNoRecomb)[pos][2]);
  probs[0][1] = probs[1][0] = log(2.0) + (*logRecomb)[pos][1] + (*logNoRecomb)[pos][1];
}

void TwoMRCAModel::passedProbs(int numGens, int pos, vector< vector<double> > &probs){
  if(numGens > 0){
    probs[0][0] = probs[1][1] = (*logNoRecomb)[pos][1];
    probs[0][1] = probs[1][0] = (*logRecomb)[pos][1];
  }
  else{
    //The passed variable is redundant in this case.
    probs[0][0] = probs[0][1] = probs[1][0] = probs[1][1] = 0.0;
  }
}

/*
 * Fills in the matrix probs with the transition probabilities for all
 * the model variables. The first dimension is always the variable
 * index. probs[idx] is a 2x2 matrix with the transition probabilities
 * between 0 and 1 for the variable with index idx.
 */
void TwoMRCAModel::computeAllTransProbs(int pos, vector< vector< vector<double> > > &probs){
  donationsTransProbs(numGensA - 1, pos, probs[DONA_IDX]);
  donationsTransProbs(numGensB - 1, pos, probs[DONB_IDX]);
  switchProbs(idxA, pos, probs[PHASEA_IDX]);
  switchProbs(idxB, pos, probs[PHASEB_IDX]);
  sharedMRCAprobs(pos, probs[SHARED0_IDX]);
  sharedMRCAprobs(pos, probs[SHARED1_IDX]);
  passedProbs(numGensA, pos, probs[PASSEDA_IDX]);
  passedProbs(numGensB, pos, probs[PASSEDB_IDX]);
}

bool TwoMRCAModel::isTransitionPossible(const TwoMRCAState& currentState, 
					const TwoMRCAState& nextState, int pos){
  double switchProb = SNPData::switch_decode((*data).diplos[idxA].switches[pos]);
  if((switchProb == 0.0) && (currentState.phaseA != nextState.phaseA)) return false;
  switchProb = SNPData::switch_decode((*data).diplos[idxB].switches[pos]);
  if((switchProb == 0.0) && (currentState.phaseB != nextState.phaseB)) return false;
  return true;
}

double TwoMRCAModel::forward(){
  vector<double> forwardProbs(states.size());
  vector<double> nextForwardProbs(states.size());
  for(int k = 0; k < states.size(); ++k){
    forwardProbs[k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  vector< vector< vector<double> > > transProbs(NUM_VAR_2MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	transProb += transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED0_IDX][states[s].sharedMRCA0][states[k].sharedMRCA0];
	transProb += transProbs[SHARED1_IDX][states[s].sharedMRCA1][states[k].sharedMRCA1];
	transProb += transProbs[PASSEDA_IDX][states[s].passedA][states[k].passedA];
	transProb += transProbs[PASSEDB_IDX][states[s].passedB][states[k].passedB];
	logs.push_back(forwardProbs[s] + transProb);
      }
      nextForwardProbs[k] = emissionProb(states[k], i) + computeLogSum(logs);
    }
    forwardProbs.swap(nextForwardProbs);
  }
  return computeLogSum(forwardProbs);
}

double TwoMRCAModel::forward(vector< vector<double> > &forwardProbs){
  forwardProbs.clear();
  forwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));

  for(int k = 0; k < states.size(); ++k){
    forwardProbs[k].resize((*data).seq_len, 0.0);
    //P(starting at state k) * P(emitting first set of symbols from state k)
    forwardProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  vector< vector< vector<double> > > transProbs(NUM_VAR_2MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));

  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	transProb += transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED0_IDX][states[s].sharedMRCA0][states[k].sharedMRCA0];
	transProb += transProbs[SHARED1_IDX][states[s].sharedMRCA1][states[k].sharedMRCA1];
	transProb += transProbs[PASSEDA_IDX][states[s].passedA][states[k].passedA];
	transProb += transProbs[PASSEDB_IDX][states[s].passedB][states[k].passedB];
	logs.push_back(forwardProbs[i - 1][s] + transProb);
      }
      forwardProbs[i][k] = emissionProb(states[k], i) + computeLogSum(logs);
    }
  }
  return computeLogSum(forwardProbs.back());
}

/*
 * Staightforward implementation of the forward algorithm in log-space.
 * Computes log-forward probabilities for the current model,
 * and stores them in forwardProbs, such that forwardsProbs[k][i] is 
 * log(f_k(i)) (see Durbin et al. p. 58)
 * Return value:
 * the probability of the whole sequence log(sum_k(f_k(seq_len))).
 */
/*
double TwoMRCAModel::forward(vector< vector<double> > &forwardProbs){
  forwardProbs.clear();
  forwardProbs.resize(states.size());
  for(int k = 0; k < states.size(); ++k){
    forwardProbs[k].resize((*data).seq_len, 0.0);
    //P(starting at state k) * P(emitting first set of symbols from state k)
    forwardProbs[k][0] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  for(int i = 1; i < (*data).seq_len; ++i){
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	logs.push_back(forwardProbs[s][i - 1] + transitionProb(states[s], states[k], i - 1));
      }
      forwardProbs[k][i] = emissionProb(states[k], i) + computeLogSum(logs);
    }
  }
  vector<double> lastColForward;
  for(int k = 0; k < states.size(); ++k){
    lastColForward.push_back(forwardProbs[k].back());
  }
  return computeLogSum(lastColForward);
}
*/

/*
 * Similar to forward(vector< vector<double> > &), but using the
 * forward algorithm for factorial HMMs.
 */
double TwoMRCAModel::factorialForward(vector< vector<double> > &forwardProbs){
  forwardProbs.clear();
  forwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));

  for(int k = 0; k < states.size(); ++k){
    //cerr << states[k].getBinaryString();
    //states[k].print();
    //P(starting at state k) * P(emitting first set of symbols from state k)
    forwardProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }

  vector< vector< vector<double> > > transProbs(NUM_VAR_2MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);
    vector<double> prevForwardProbs(forwardProbs[i - 1]);
    for(int var = 0; var < NUM_VAR_2MRCA; ++var){
      for(int k = 0; k < states.size(); ++k){
	int val = *(states[k].variables[var]);
	int zeroIdx = (val == 0)? states[k].idx: getInverseStateIdx(states[k], var);
	int oneIdx = (val == 1)? states[k].idx: getInverseStateIdx(states[k], var);
	forwardProbs[i][k] = (maxValues[var] == 0)? prevForwardProbs[k]:
	  computeLogSumOfTwo(transProbs[var][0][val] + prevForwardProbs[zeroIdx],
			     transProbs[var][1][val] + prevForwardProbs[oneIdx]);
      }
      swap(forwardProbs[i], prevForwardProbs);
    }
    for(int k = 0; k < states.size(); ++k){
      forwardProbs[i][k] = emissionProb(states[k], i) + prevForwardProbs[k];
    }
  }
  return computeLogSum(forwardProbs.back());
}

double TwoMRCAModel::backward(vector< vector<double> > &backwardProbs){
  backwardProbs.clear();
  backwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));

  vector< vector< vector<double> > > transProbs(NUM_VAR_2MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    computeAllTransProbs(i, transProbs);
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	transProb += transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED0_IDX][states[s].sharedMRCA0][states[k].sharedMRCA0];
	//I only use SHARED0_IDX, because the transition matrix is the same 
	//for sharedMRCA0 and sharedMRCA1.
	transProb += transProbs[SHARED0_IDX][states[s].sharedMRCA1][states[k].sharedMRCA1];
	transProb += transProbs[PASSEDA_IDX][states[s].passedA][states[k].passedA];
	transProb += transProbs[PASSEDB_IDX][states[s].passedB][states[k].passedB];
	logs.push_back(backwardProbs[i + 1][s] + transProb + emissionProb(states[s], i + 1));
      }
      backwardProbs[i][k] = computeLogSum(logs);
    }
  }
  
  vector<double> firstCol(states.size(), 0.0);
  for(int k = 0; k < states.size(); ++k){
    firstCol[k] = backwardProbs[0][k] + emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  return computeLogSum(firstCol);
}

/*
 * Staightforward implementation of the backward algorithm in log-space.
 * Computes log-backward probabilities for the current model,
 * and returns them in a matrix M, such that M[k][i] is 
 * log(b_k(i)) (see Durbin et al. p. 59)
 */
/*
double TwoMRCAModel::backward(vector< vector<double> > &backwardProbs){
  backwardProbs.clear();
  backwardProbs.resize(states.size());
  for(int k = 0; k < states.size(); ++k){
    backwardProbs[k].resize((*data).seq_len, 0.0);
  }
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    for(int k = 0; k < states.size(); ++k){
      vector<double> logs;
      for(int s = 0; s < states.size(); ++s)
	logs.push_back(transitionProb(states[k], states[s], i) + 
		       emissionProb(states[s], i + 1) + backwardProbs[s][i + 1]);
      backwardProbs[k][i] = computeLogSum(logs);
    }
  }
    
  vector<double> firstCol(states.size(), 0.0);
  for(int k = 0; k < states.size(); ++k){
    firstCol[k] = backwardProbs[k][0] + emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  return computeLogSum(firstCol);
}
*/

double TwoMRCAModel::factorialBackward(vector< vector<double> > &backwardProbs){
  backwardProbs.clear();
  backwardProbs.resize((*data).seq_len, vector<double>(states.size(), 0.0));
  
  vector< vector< vector<double> > > transProbs(NUM_VAR_2MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  vector<double> prevBackwardProbs(states.size(), 0.0);
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    for(int k = 0; k < states.size(); ++k){
      prevBackwardProbs[k] = emissionProb(states[k], i + 1) + backwardProbs[i + 1][k];
    }
    computeAllTransProbs(i, transProbs);
    for(int var = NUM_VAR_2MRCA - 1; var >= 0; --var){
      for(int k = 0; k < states.size(); ++k){
	int val = *(states[k].variables[var]);
	int zeroIdx = (val == 0)? states[k].idx: getInverseStateIdx(states[k], var);
	int oneIdx = (val == 1)? states[k].idx: getInverseStateIdx(states[k], var);
	backwardProbs[i][k] = (maxValues[var] == 0)? prevBackwardProbs[k]:
	  computeLogSumOfTwo(transProbs[var][val][0] + prevBackwardProbs[zeroIdx],
			     transProbs[var][val][1] + prevBackwardProbs[oneIdx]);
      }
      if(var > 0)
	swap(backwardProbs[i], prevBackwardProbs);
    }
  }
  
  for(int k = 0; k < states.size(); ++k){
    prevBackwardProbs[k] = backwardProbs[0][k] + emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  return computeLogSum(prevBackwardProbs);
}

double TwoMRCAModel::viterbi(vector<int> &stateIdx){
  vector< vector<double> > viterbiProbs((*data).seq_len, vector<double>(states.size(), 0.0));
  vector< vector<int> > maxStates((*data).seq_len, vector<int>(states.size(), -1));
  stateIdx.resize((*data).seq_len);

  for(int k = 0; k < states.size(); ++k){
    //P(starting at state k) * P(emitting first set of symbols from state k)
    viterbiProbs[0][k] = emissionProb(states[k], 0) + states[k].logInitialProb;
  }
  vector< vector< vector<double> > > transProbs(NUM_VAR_2MRCA, 
						vector< vector< double> >(2, 
									  vector<double>(2)));
  for(int i = 1; i < (*data).seq_len; ++i){
    computeAllTransProbs(i - 1, transProbs);
    for(int k = 0; k < states.size(); ++k){
      double maxVal = correctLog(0.0);
      int maxState = -1;
      for(int s = 0; s < states.size(); ++s){
	double transProb = transProbs[DONA_IDX][states[s].donationsA][states[k].donationsA];
	transProb += transProbs[DONB_IDX][states[s].donationsB][states[k].donationsB];
	transProb += transProbs[PHASEA_IDX][states[s].phaseA][states[k].phaseA];
	transProb += transProbs[PHASEB_IDX][states[s].phaseB][states[k].phaseB];
	transProb += transProbs[SHARED0_IDX][states[s].sharedMRCA0][states[k].sharedMRCA0];
	transProb += transProbs[SHARED1_IDX][states[s].sharedMRCA1][states[k].sharedMRCA1];
	transProb += transProbs[PASSEDA_IDX][states[s].passedA][states[k].passedA];
	transProb += transProbs[PASSEDB_IDX][states[s].passedB][states[k].passedB];
	transProb += viterbiProbs[i - 1][s];
	if(transProb > maxVal){
	  maxVal = transProb;
	  maxState = s;
	}
      }
      viterbiProbs[i][k] = emissionProb(states[k], i) + maxVal;
      maxStates[i][k] = maxState;
    }
  }
  
  vector<double> lastProbs = viterbiProbs.back();
  double maxVal = lastProbs[0];
  int maxState = 0;
  for(int k = 1; k < states.size(); ++k){
    if(lastProbs[k] > maxVal){
      maxVal = lastProbs[k];
      maxState = k;
    }
  }
  stateIdx[(*data).seq_len - 1] = maxState;
  for(int i = (*data).seq_len - 2; i >= 0; --i){
    stateIdx[i] = maxStates[i + 1][stateIdx[i + 1]];
  }
  return maxVal;
}

void TwoMRCAModel::simpleDiscretePosteriorIBD(vector< vector<double> > &forwardProbs, 
					      vector< vector<double> > &backwardProbs,
					      double seqProb, double nonIBDcutoff, double IBDcutoff, 
					      vector< vector< double> > &posteriorLogProb){
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len);
  //double logNonIBDcutoff = correctLog(nonIBDcutoff);
  //double logIBDcutoff = correctLog(IBDcutoff);
  for(int i = 0; i < (*data).seq_len; ++i){
    posteriorLogProb[i].resize(3, 0.0);
    vector<double> IBDlogs1, IBDlogs2, corrAlogs, corrBlogs;
    for(int k = 0; k < states.size(); ++k){
      AssertMsg(states[k].numCorr > -1, 
		"Attempt to compute posterior IBD probabilities with uninitialized states.");
      double stateProb = forwardProbs[k][i] + backwardProbs[k][i] - seqProb;
      if(states[k].corrA == 1) corrAlogs.push_back(stateProb);
      if(states[k].corrB == 1) corrBlogs.push_back(stateProb);
      if(states[k].numCorr > 0) IBDlogs1.push_back(stateProb);
      //if(states[k].numCorr == 1) IBDlogs1.push_back(stateProb);
      //else if(states[k].numCorr == 2) IBDlogs2.push_back(stateProb);
    }
    double postProbIBD1 = IBDlogs1.empty()? correctLog(0.0): computeLogSum(IBDlogs1);
    posteriorLogProb[i][0] = postProbIBD1;
    //double postProbIBD2 = IBDlogs2.empty()? correctLog(0.0): computeLogSum(IBDlogs2);
    //if(postProbIBD1 > logIBDcutoff) posteriorLogProb[i][0] = 1;
    //else if(postProbIBD2 > logIBDcutoff) posteriorLogProb[i][0] = 2;
    //else if(postProbIBD1 < logNonIBDcutoff && postProbIBD2 < logNonIBDcutoff)
    //posteriorLogProb[i][0] = 0;
    //else posteriorLogProb[i][0] = -1;
    posteriorLogProb[i][1] = corrAlogs.empty()? correctLog(0.0): computeLogSum(corrAlogs);
    posteriorLogProb[i][2] = corrBlogs.empty()? correctLog(0.0): computeLogSum(corrBlogs);
  }
}

/*
 * After this method returns, posteriorLogProbs.size() will be equal to the
 * number of SNPs. posteriorLogProb[i] will be a vector of size 4.
 * posteriorLogProb[i][0] will be the posterior probability that numCorr at 
 * position i is 1.
 * posteriorLogProb[i][1] will be the posterior probability that numCorr at
 * position i is 2 (this is meaningless in most cases).
 * posteriorLogProb[i][2] will be the posterior probability that corrA at
 * position i is 1 (i.e. the correlated allele of A is 1. Notice that 
 * when numCorr is 0, that is, when there are no correlated alleles, 
 * corrA and corrB are always 0).
 * posteriorLogProb[i][3] will be the posterior probability that corrB at
 * position i is 1.
 */
void TwoMRCAModel::simplePosteriorIBD(vector< vector<double> > &forwardProbs, 
				      vector< vector<double> > &backwardProbs,
				      double seqProb,
				      vector< vector<double> > &posteriorLogProb){
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len);
  for(int i = 0; i < (*data).seq_len; ++i){
    if(numGensA == 0 && numGensB == 0)
      posteriorLogProb[i].resize(6, 0.0);
    else
      posteriorLogProb[i].resize(5, 0.0);
    vector<double> IBDlogs1, IBDlogs2, phaseAlogs, phaseBlogs, corrAlogs, corrBlogs;
    for(int k = 0; k < states.size(); ++k){
      //Posterior probability of that state.
      double stateProb = forwardProbs[i][k] + backwardProbs[i][k] - seqProb;
      if(states[k].numCorr == 1) IBDlogs1.push_back(stateProb);
      if(states[k].numCorr == 2) IBDlogs2.push_back(stateProb);
      if(states[k].phaseA == 1) phaseAlogs.push_back(stateProb);
      if(states[k].phaseB == 1) phaseBlogs.push_back(stateProb);
      if(states[k].numCorr > 0 && states[k].corrA == 1) corrAlogs.push_back(stateProb);
      if(states[k].numCorr > 0 && states[k].corrB == 1) corrBlogs.push_back(stateProb);
    }
    int idx = 0;
    posteriorLogProb[i][idx++] = IBDlogs1.empty()? correctLog(0.0): computeLogSum(IBDlogs1);
    if(numGensA == 0 && numGensB == 0)
      posteriorLogProb[i][idx++] = IBDlogs2.empty()? correctLog(0.0): computeLogSum(IBDlogs2);
    posteriorLogProb[i][idx++] = phaseAlogs.empty()? correctLog(0.0): computeLogSum(phaseAlogs);
    posteriorLogProb[i][idx++] = phaseBlogs.empty()? correctLog(0.0): computeLogSum(phaseBlogs);
    posteriorLogProb[i][idx++] = corrAlogs.empty()? correctLog(0.0): computeLogSum(corrAlogs);
    posteriorLogProb[i][idx++] = corrBlogs.empty()? correctLog(0.0): computeLogSum(corrBlogs);
  }
}

void TwoMRCAModel::simplePosteriorIBDAll(vector< vector<double> > &forwardProbs, 
					 vector< vector<double> > &backwardProbs,
					 double seqProb,
					 vector< vector<double> > &posteriorLogProb){
  posteriorLogProb.clear();
  posteriorLogProb.resize((*data).seq_len);
  for(int i = 0; i < (*data).seq_len; ++i){
    posteriorLogProb[i].resize(8, 0.0);
    
    vector<double> IBDlogs1, IBDlogs2, passedLogsA, passedLogsB,
      ancLogs0, ancLogs1, donLogsA, donLogsB;
    for(int k = 0; k < states.size(); ++k){
      //Posterior probability of that state.
      double stateProb = forwardProbs[i][k] + backwardProbs[i][k] - seqProb;
      if(states[k].numCorr == 1) IBDlogs1.push_back(stateProb);
      if(states[k].numCorr == 2) IBDlogs2.push_back(stateProb);
      if(states[k].sharedMRCA0 == 1) ancLogs0.push_back(stateProb);
      if(states[k].sharedMRCA1 == 1) ancLogs1.push_back(stateProb);
      if(states[k].passedA == 1) passedLogsA.push_back(stateProb);
      if(states[k].passedB == 1) passedLogsB.push_back(stateProb);
      if(states[k].donationsA == 1) donLogsA.push_back(stateProb);
      if(states[k].donationsB == 1) donLogsB.push_back(stateProb);
    }
    int idx = 0;
    posteriorLogProb[i][idx++] = IBDlogs1.empty()? correctLog(0.0): computeLogSum(IBDlogs1);
    posteriorLogProb[i][idx++] = IBDlogs2.empty()? correctLog(0.0): computeLogSum(IBDlogs2);
    posteriorLogProb[i][idx++] = ancLogs0.empty()? correctLog(0.0): computeLogSum(ancLogs0);
    posteriorLogProb[i][idx++] = ancLogs1.empty()? correctLog(0.0): computeLogSum(ancLogs1);
    posteriorLogProb[i][idx++] = passedLogsA.empty()? correctLog(0.0): computeLogSum(passedLogsA);
    posteriorLogProb[i][idx++] = passedLogsB.empty()? correctLog(0.0): computeLogSum(passedLogsB);
    posteriorLogProb[i][idx++] = donLogsA.empty()? correctLog(0.0): computeLogSum(donLogsA);
    posteriorLogProb[i][idx++] = donLogsB.empty()? correctLog(0.0): computeLogSum(donLogsB);
  }
}

/*
 * After this method returns, posteriorLogProbs[k][i] will be the posterior
 * probability of the k-th state at the i-th position.
 */
void TwoMRCAModel::computePostLogProb(vector< vector<double> > &forwardProbs, 
				      vector< vector<double> > &backwardProbs,
				      vector< vector<double> > &posteriorLogProbs){
  vector<double> lastColForward;
  for(int k = 0; k < states.size(); ++k){
    lastColForward.push_back(forwardProbs[k].back());
  }
  double seqLogProb = computeLogSum(lastColForward);

  posteriorLogProbs.clear();
  posteriorLogProbs.resize(states.size());
  for(int k = 0; k < states.size(); ++k){
    posteriorLogProbs[k].resize((*data).seq_len, 0.0);
    for(int i = 0; i < (*data).seq_len; ++i)
      posteriorLogProbs[k][i] = forwardProbs[k][i] + backwardProbs[k][i] - seqLogProb;
  }
}
