% data: 3-dimensional matrix, rels x individuals x features
% numFolds: number of folds for cross-validation
% classifyRels: not all the rels in the first dimension of data will be used. 
% Those not in classifyRels will be ignored. In other words, only
% individuals from the rels in classifyRels will be used for training (and
% testing).
% assumedRels: this determines what the features in the third dimension of 
% data are.
% classifyRelIdx: vector with length length(classifyRels).
% classifyRelIdx(i) is j, if classifyRels(i) corresponds to the j-th
% relationship in data.
% relAssign: vector with length length(classifyRels). relAssign(i) is the
% group of the i-th relationship in classifyRels. 
% selCols: selected features from data used for classification.
function [confSum] = classifyPost2(data, numFolds, classifyRels, assumedRels, classifyRelIdx, relAssign, selCols)

% determine the number of classes. Classes must be consecutive integers
% starting from 1.
mappedClasses = remapClasses(relAssign);
numClasses = length(unique(mappedClasses));
% determine how many relationships there are in each group
classCounts = zeros(numClasses, 1);
for c = 1:numClasses,
    classCounts(c) = sum(mappedClasses == c);
end

% Total number of individuals per relationship.
numIndivTot = size(data, 2);
% number of individuals used in each fold of cross-validation
numTest = ceil(numIndivTot / numFolds)
numTrain = numIndivTot - numTest

numCounts = 4; % haplocount features
numBlockStats = 1; % block statistics
numCols = numCounts * length(assumedRels) + numBlockStats * length(assumedRels) + length(assumedRels);
blockStartIdx = numCounts * length(assumedRels) + 1;
likeStartIdx = blockStartIdx + numBlockStats * length(assumedRels);

% average confusion matrix
confSum = zeros(numClasses);

for f = 1:numFolds,
    % Indices of individuals to use for testing.
    testInd = ((f - 1) * numTest + 1):(min(f * numTest, numIndivTot));
    trainInd = setdiff(1:numIndivTot, testInd);
    
    % trainData(i, j) is the j-th feature of the i-th training pair
    % trainGroups(i) is the true relationship (i.e. class) of the i-th
    % training pair.
    trainData = zeros(length(classifyRels) * length(trainInd), numCols);
    trainGroups = zeros(length(classifyRels) * length(trainInd), 1);
    testData = zeros(length(classifyRels) * length(testInd), numCols);
    testGroups = zeros(length(classifyRels) * length(testInd), 1);
    
    for r = 1:length(classifyRels),
        newTrainInd = ((r - 1) * numTrain + 1):((r - 1) * numTrain + length(trainInd));
        trainData(newTrainInd, :) = data(classifyRelIdx(r), trainInd, :);
        trainGroups(newTrainInd) = mappedClasses(r);
        newTestInd = ((r - 1) * numTest + 1):((r - 1) * numTest + length(testInd));
        testData(newTestInd, :) = data(classifyRelIdx(r), testInd, :);
        testGroups(newTestInd) = mappedClasses(r);
    end
%     trainData(:, likeStartIdx + 1) = trainData(:, likeStartIdx + 1) - trainData(:, likeStartIdx);
%     trainData(:, likeStartIdx + 2) = trainData(:, likeStartIdx + 2) - trainData(:, likeStartIdx);
%     trainData(:, likeStartIdx + 3) = trainData(:, likeStartIdx + 3) - trainData(:, likeStartIdx);
%     trainData(:, likeStartIdx + 4) = trainData(:, likeStartIdx + 4) - trainData(:, likeStartIdx);
%     temp = trainData(:, likeStartIdx + 5);
%     trainData(:, likeStartIdx + 5) = trainData(:, likeStartIdx + 5) - trainData(:, likeStartIdx);
%     trainData(:, likeStartIdx + 6) = trainData(:, likeStartIdx + 6) - temp;
%     
%     testData(:, likeStartIdx + 1) = testData(:, likeStartIdx + 1) - testData(:, likeStartIdx);
%     testData(:, likeStartIdx + 2) = testData(:, likeStartIdx + 2) - testData(:, likeStartIdx);
%     testData(:, likeStartIdx + 3) = testData(:, likeStartIdx + 3) - testData(:, likeStartIdx);
%     testData(:, likeStartIdx + 4) = testData(:, likeStartIdx + 4) - testData(:, likeStartIdx);
%     temp = testData(:, likeStartIdx + 5);
%     testData(:, likeStartIdx + 5) = testData(:, likeStartIdx + 5) - testData(:, likeStartIdx);
%     testData(:, likeStartIdx + 6) = testData(:, likeStartIdx + 6) - temp;
    
    % Rescale based on the training data.
%     trainMean = mean(trainData, 1);
%     trainStd = std(trainData, 1);
%     
%     for c = 1:size(trainData, 2),
%         trainData(:, c) = (trainData(:, c) - trainMean(c)) / trainStd(c);
%         testData(:, c) = (testData(:, c) - trainMean(c)) / trainStd(c);
%     end
    
    % linear discriminant analysis
    [pred, err, post] = classify(testData(:, selCols), trainData(:, selCols), trainGroups, 'linear');
    
    conf = confusionmat(testGroups, pred);
    confSum = confSum + conf;
end

prec = sum(diag(confSum)) / (numIndivTot * length(classifyRels));
fprintf('Confusion matrix\n');
for c = 1:numClasses,
    confSum(c, :) = confSum(c, :) / (classCounts(c) * numIndivTot);
end

disp(confSum);
fprintf('prec = %.4f\n', prec);
end

% if exist('keepData', 'var') ~= 1 || keepData ~= 1,
%     clear all
% end
% 
% pairPref = 'CEU_TSI';
% parDir = '/media/My Passport/RelationshipInfer';
% countsPref = strcat(parDir, '/PostCounts/hapmap3_', pairPref, '_cleared/');
% likePref = strcat(parDir, '/Likelihoods/hapmap3_', pairPref, '/');
% likeSuf = '_allChr_pruned';
% distPref = '../BlockGeneticDist/hapmap3_CEU/Pruned/Pair_';
% sizesPath = '../../hapmap/hapmap3/CEU_TSI/CEU_TSI_pruned.chrSizes';
% 
% rels1 = {'2_0_0', '2_0_1', '2_0_2', '2_0_3', '2_1_1', '2_1_2', ...
%     '2_1_3', '2_2_2', '2_2_3', '2_3_3', ...
%     '1_0_0', '1_0_1', '1_0_2', '1_0_3', '1_1_1', '1_1_2', ...
%     '1_1_3', '1_2_2', '1_2_3', '1_3_3', ...
%     '1_0_n1', '1_n1_0', '1_n1_1', '1_n1_2', '1_n1_3'};
% 
% rels2 = {'2_0_3', '2_1_2', '1_n1_3', '1_0_2', '1_1_1'};
% 
% rels3 = {'2_0_3', '2_1_2', '1_n1_3', '1_0_2', '1_1_1', '2_0_1', '2_1_1', '2_2_2'};
% 
% rels4 = {'2_0_0', '2_0_1', '2_0_2', '2_1_1', '1_n1_2', '1_0_1', ...
%     '2_0_3', '2_1_2', '1_n1_3', '1_0_2', '1_1_1', ...
%     '2_2_2', '2_1_3', '1_1_2', '1_0_3'};
% 
% rels5 = {'2_0_1', '2_1_1', '2_1_2', '2_2_2'};
% rels6 = {'2_2_2', '2_1_3', '1_1_2', '1_0_3'};
% 
% % Balanced vs unbalanced
% %relAssign = [1, 2, 3, 2];
% 
% trueRels = rels4;
% assumedRels = rels4;
% % The loaded data has examples from all classes in trueRels, but we'll only
% % use examples from the classes in classifyRels.
% classifyRels = rels4;
% % classifyRelIdx(i) is the index j of the relationship in trueRels such that
% % classifyRels(i) == trueRels(j).
% classifyRelIdx = zeros(length(classifyRels), 1);
% for i = 1:length(classifyRels),
%     for j = 1:length(trueRels),
%         if strcmp(classifyRels(i), trueRels(j)),
%             classifyRelIdx(i) = j;
%             break;
%         end
%     end
% end
% sets = [1];
% 
% % Balanced vs unbalanced
% % relAssign = [2, 2, 2, 2, 1, 2, 2, 1];
% % Balanced, a little imbalanced, very imbalanced.
% % relAssign = [3, 2, 3, 3, 1, 2, 1, 1];
% % Degree
% % relAssign = [3, 3, 3, 3, 3, 1, 2, 4];
% % Exact rel
% relAssign = 1:length(classifyRels);
% classes = unique(relAssign);
% classCounts = zeros(length(classes), 1);
% for c = classes,
%     classCounts(c) = sum(relAssign == classes(c));
% end
% 
% minIndiv = 1;
% maxIndiv = 100;
% numIndiv = maxIndiv - minIndiv + 1;
% numIndivTot = numIndiv * length(sets);
% 
% win = 20;
% expIbdCutoff = 0.9;
% expNonIbdCutoff = 0.2;
% expCutoff = 0.5;
% countsSuf = strcat('_w', num2str(win), '_i', num2str(expIbdCutoff), ...
%     '_n', num2str(expNonIbdCutoff), '_c', num2str(expCutoff));
% percent = 10;
% 
% numCounts = 4; % haplocount features
% numBlockStats = 1; % block statistics
% numCols = numCounts * length(assumedRels) + numBlockStats * length(assumedRels) + length(assumedRels);
% blockStartIdx = numCounts * length(assumedRels) + 1;
% likeStartIdx = blockStartIdx + numBlockStats * length(assumedRels);
% 
% if exist('keepData', 'var') ~= 1 || keepData ~= 1,
%     data = loadData(pairPref, countsPref, countsSuf, likePref, likeSuf, ...
%         distPref, sizesPath, trueRels, assumedRels, sets, minIndiv, maxIndiv, percent);
% end
% 
% % Permute all the data now
% for r = 1:length(trueRels),
%     data(r, :, :) = data(r, randperm(numIndivTot), :);
% end
% 
% numFolds = 10;
% % num of testing examples (per set per relationship) in each CV fold
% numTest = ceil(numIndivTot / numFolds);
% numTrain = numIndivTot - numTest;
% 
% blockIndices = (blockStartIdx:(likeStartIdx - 1));
% newIndices = numCols + 1;
% % selected features
% % What features will we take from each assumed model?
% selLikelihoods = 1:length(assumedRels);
% selBlocks = 1:length(assumedRels) %[4];
% selPostcounts = [2];
% selCols = [];
% for s = selPostcounts,
%     selCols = [selCols, (s - 1) * numCounts + [1:numCounts]];
% end
% for s = selBlocks,
%     selCols = [selCols, (s - 1) * numBlockStats + blockStartIdx - 1 + [1:numBlockStats]];
% end
% selCols = [selCols, likeStartIdx - 1 + selLikelihoods];