% Loads data from several mat files and combines them into a single matrix.
% filenames: names of the mat files to load
% trueRels: true rels for the data matrix returned (the files loaded may
% have additional rels which will be ignored).
% assumedRels: assumed rels for the data matrix retured
% numIndivPerSet: number of individuals that will be loaded from each set
% NO CHECK is performed that the loaded files actually contain all the
% necessary information, that is, they do contain all the rels in trueRels
% and assumedRels and all the individuals specified by numIndivPerSet.
% Return value:
% data: tree-dimensional matrix with size 
% length(trueRels) * sum(numIndivPerSet) * numFeatures, where numFeatures
% is the number of classifier features.
function data = loadClassifierData(filenames, trueRels, assumedRels, numIndivPerSet)

numCounts = 4;
numBlockCounts = 1;
% Number of features
numCols = (numCounts + numBlockCounts + 1) * length(assumedRels);
blockStartIdx = numCounts * length(assumedRels) + 1;
likeStartIdx = blockStartIdx + numBlockCounts * length(assumedRels);

data = ones(length(trueRels), sum(numIndivPerSet), numCols) * (-100);
% The individuals loaded from set i will be stored in data(:,
% startSetIdx(i):(startSetIdx(i) + numIndivPerSet(i) - 1), :).
startSetIdx = ones(length(numIndivPerSet), 1);
for s = 2:length(numIndivPerSet),
    startSetIdx(s) = startSetIdx(s - 1) + numIndivPerSet(s - 1);
end

% Min and max individual read for each set and each rel in trueRels and
% assumedRels.
% minIndivReadTrue = ones(length(numIndivPerSet), length(trueRels)) * sum(numIndivPerSet);
% maxIndivReadTrue = ones(length(numIndivPerSet), length(trueRels)) * sum(numIndivPerSet);
% minIndivReadAssume = ones(length(numIndivPerSet), length(assumedRels)) * sum(numIndivPerSet);
% maxIndivReadAssume = ones(length(numIndivPerSet), length(assumedRels)) * sum(numIndivPerSet);

for filename = filenames,
    restored = load(char(filename));
    s = restored.set;
    indivIndices = startSetIdx(s) - 1 + [restored.minIndiv:min(restored.maxIndiv, numIndivPerSet(s))];
    
    if isempty(indivIndices)
        continue;
    end
    % relIndices(i) is 0 if restored.trueRels is not in trueRels.
    % Otherwise, relIndices(i) = j means that restore.trueRels(i) ==
    % trueRels(j).
    relIndices = zeros(length(restored.trueRels), 1);
    for i = 1:length(restored.trueRels),
        for j = 1:length(trueRels),
            if strcmp(restored.trueRels{i}, trueRels{j}),
                relIndices(i) = j;
%                 minIndivReadTrue(s, j) = min(minIndivReadTrue(s, j), restored.minIndiv);
%                 maxIndivReadTrue(s, j) = max(maxIndivReadTrue(s, j), restored.maxIndiv);
                break;
            end
        end
    end
    
    % The features at indices originalIndices in restore.data will be
    % copied to the indices newIndices in data.
    originalIndices = [];
    newIndices = [];
    for i = 1:length(restored.assumedRels),
        for j = 1:length(assumedRels),
            if strcmp(restored.assumedRels{i}, assumedRels{j}),
                restoreCountIndices = numCounts * (i - 1) + [1:numCounts];
                countIndices = numCounts * (j - 1) + [1:numCounts];
                restoreBlockIndices = numCounts * length(restored.assumedRels) + numBlockCounts * (i - 1) + [1:numBlockCounts];
                blockIndices = numCounts * length(assumedRels) + numBlockCounts * (j - 1) + [1:numBlockCounts];
                restoreLikeIndices = (numCounts + numBlockCounts) * length(restored.assumedRels) + i;
                likeIndices = (numCounts + numBlockCounts) * length(assumedRels) + j;
                originalIndices = [originalIndices, restoreCountIndices, restoreBlockIndices, restoreLikeIndices];
                newIndices = [newIndices, countIndices, blockIndices, likeIndices];
%                 minIndivReadAssume(s, j) = min(minIndivReadAssume(s, j), restored.minIndiv);
%                 maxIndivReadAssume(s, j) = max(maxIndivReadAssume(s, j), restored.maxIndiv);
            end
        end
    end
    
    % For each true rel in the restored data that does correspond to some
    % rel in data (i.e. for i such that relIndices(i) > 0)
    for i = 1:length(relIndices),
        if relIndices(i) > 0,
            data(relIndices(i), indivIndices, newIndices) = restored.data(i, :, originalIndices);
        end
    end
end

for r = 1:size(data, 1),
    for i = 1:size(data, 2),
        % This means that no data was loaded for this true rel and this individual. 
        if any(data(r, i, blockStartIdx:(likeStartIdx - 1)) == -100),
            err = MException('', 'Not all data loaded for relationship %s, indiv %d', trueRels{r}, i);
            throw(err)
        end
    end
end
end