% Reads the postcounts for all individuals from minIndiv to maxIndiv.
% trueRel is assumed to be the true relationship of the individuals and
% assumedRel is assumed to be the relationship for which the postcounts
% were computed.
% The input files are assumed to be
% <countPref><trueRel>/CEU_Pair_<trueRel>_<i>/CEU_Pair_<trueRel>_<i>_w20_i
% <expIbdCutoff>_n<expNonIBDCutoff>_c<cutoff>_<assumedRel>.logProb
% chromSizes(i) must be the genetic position of the last SNP in the i-th
% chromosome
function counts = loadPostCounts2(minIndiv, maxIndiv, countPref, distPref, ...
        trueRel, assumedRel, expIbdCutoff, expNonIbdCutoff, expCutoff, ...
        chromSizes)

    numIndiv = maxIndiv - minIndiv + 1;
    % Genetic distances cutoffs for each chromosome under the assumed
    % relationship. IBD blocks with smaller genetic distances will be
    % ignored.
    cutoffs = zeros(22, 1);
    for c = 1:22,
        dists = importdata(strcat(distPref, assumedRel, '/Pair_', assumedRel, ...
            '_chr', int2str(c), '_Merged_10000'));
        % Use 5% percentile as the cutoff.
        cutoffs(c) = prctile(dists, 10);
    end
    
    chrCol = 4;
    startPos = 11;
    endPos = 12;
    freqAcol = 15;
    freqBcol = 16;
    
    counts = zeros(numIndiv, 4);
    for i = minIndiv:maxIndiv,
        countFile = strcat(countPref, trueRel, ...
            '/CEU_Pair_', trueRel, '_', num2str(i), '/CEU_Pair_', trueRel, ...
            '_', num2str(i), '_w20', ...
            '_i', num2str(expIbdCutoff), '_n', num2str(expNonIbdCutoff), ...
            '_c', num2str(expCutoff), '_', assumedRel, '.postcounts');
        tempStruct = importdata(countFile, '\t', 1);
        temp = tempStruct.data;
        chroms = temp(:, chrCol);
        selectors = zeros(size(temp, 1), 1);
        last01 = 0;
        last10 = 0;
        for j = 1:size(temp, 1),
            if temp(j, 1) == 0 && temp(j, 2) == 1,
                selectors(j) = 1;
                last
            % The following code might not filter correctly blocks at the 
            % end of the chromosome.
            % If this is a transition 1->0, then startPos is the start of 
            % the IBD block and endPos is the end of the IBD block.
            if temp(j, 1) == 1 && temp(j, 2) == 0,
                startPos = temp(j, startPos);
                if last01 == 0 || temp(last01, chrCol) ~= temp(j, chrCol),
                    startPos = temp(last01, endPos);
                end
                temp(j, endPos) - temp(j, startPos) > cutoffs(chroms(j)),
                selectors(j) = 1;
                %blocks(j, :) = [1, temp(j, startPos), temp(j, endPos)]; 
            % If this is a transition 0->1, then the following transition
            % is probably 1->0, so we look at the following transition to
            % see if the block passed the cutoff. If the following
            % transition is not a 1->0 transition on the same chromosome,
            % we just pretend that the block passed the cutoff.
            elseif (temp(j, 1) == 0 && temp(j, 2) == 1),
                cut = cutoffs(chroms(j));
                if j == size(temp, 1) || chroms(j) ~= chroms(j + 1),
                    if chromSizes(chroms(j)) -  temp(j, endPos) > cut
                        selectors(j) = 1;
                        blocks(j, :) = [0, temp(j, endPos), chromSizes(chroms(j))];
                    end
                elseif temp(j + 1, 1) == 1 && chroms(j) == chroms(j + 1) && ...
                        temp(j, endPos) == temp(j + 1, startPos) && ...
                        temp(j + 1, endPos) - temp(j + 1, startPos) > cut,
                    selectors(j) = 1;
                    blocks(j, :) = [0, temp(j + 1, endPos), chromSizes(chroms(j))];
                end
            end
        end
        sel = temp(selectors == 1, :);
        idx = i - minIndiv + 1;
        if ~isempty(sel),
            % mean log ratio of freqA and freqB over all transitions
            counts(idx, 1) = mean(sel(:, freqAcol) - sel(:, freqBcol));
            counts(idx, 2) = sum((sel(:, freqAcol) - sel(:, freqBcol)) < 0) / size(sel, 1);
            counts(idx, 3) = sum((sel(:, freqAcol) - sel(:, freqBcol)) > 0) / size(sel, 1);
        else
            counts(idx, 1:3) = 0.0;
        end
        % number of transitions.
        counts(idx, 4) = size(sel, 1);
    end
    
end