% Discretizes the posteriors and returns a set of IBD, non-IBD and
% transition blocks.
% postIbd: posterior probabilities of IBD for ***ONE CHROMOSOME***.
% map(i, 1) is the index of the SNP in the pruned data corresponding to the
% i-th SNP of the full data, or -1 if there's no such SNP.
% map can have other columns corresponding for example to genomic or
% genetic positions.
% prunedMap is the "inverse" of map. prunedMap(i) is the index of the SNP
% of the full data corresponding to the i-th SNP in the pruned data.
% ibdCut: if posterior prob > ibdCut, then call IBD
% nonIbdCut: if posterior prob < nonIbdCut, then call non-IBD
% The positions returned are in the pruned space.
% sizeCol: the column of map that will be used to get positions for the
% SNPs in order to compare the sizes of blocks to minSize.
% minSize: IBD blocks with size less than that size (where "size" is
% determined by sizeCol and map) will be merged with surrounding blocks. 
function blocks = post2Blocks(postIbd, map, prunedMap, sizeCol, minSize, ibdCut, nonIbdCut)
% I assume I rarely have more than 40 blocks on the same chrom.
temp = ones(40, 3) * (-2);
% Discretize the posteriors.
discretePost = ones(size(postIbd, 1), 1) * (-1);
discretePost(postIbd(:, 1) > ibdCut) = 1;
discretePost(postIbd(:, 1) < nonIbdCut) = 0;
numBlocks = 0;
if ~isempty(postIbd),
    % First, I will chunk up the indicators into blocks. Consecutive 1's
    % (probably with some -1's in between) are blocked together,
    % consecutive 0's are blocked together, and consecutive -1's between
    % blocks of 1's and 0's are blocked together. (-1's within a block of
    % 1's will be combined with the 1's into a single block, same for -1's
    % within 0's.)
    
    pos = 1;
    % Find the first position that is not -1.
    lastPos = -1;
    while pos < size(discretePost, 1),
        if discretePost(pos, 1) > -1,
            lastPos = pos;
            break;
        end
        pos = pos + 1;
    end
    if lastPos < 0,
        % All positions are unknown, we'll return no blocks at all.
        blocks = temp(temp(:, 1) > -2, :);
        return;
    end
    
    firstKnown = lastPos;
    % Starting position of an IBD or non-IBD block.
    blockStart = 1;
    for i = (firstKnown + 1):size(discretePost, 1),
        ibd = discretePost(i, 1);
        if ibd < 0,
            continue;
        end
        % Transitions in corr alleles are ignored.
        %corrA = discretePost(i, 2);
        %prevCorrA = discretePost(lastPos, 2);
        %corrB = discretePost(i, 3);
        %prevCorrB = discretePost(lastPos, 3);
        
        % lastPos is the last position before the current position that is
        % not -1.
        %                     i   
        % ????11111111????????00000
        %     b      l
        % b is the block start
        % l is lastPos
        % I only look at transitions between 0 and 1.
        prevIbd = discretePost(lastPos, 1);
        if ibd ~= prevIbd,
            numBlocks = numBlocks + 1;
            temp(numBlocks, :) = [prevIbd, blockStart, lastPos];
            if lastPos < i - 1,
                numBlocks = numBlocks + 1;
                temp(numBlocks, :) = [-1, lastPos + 1, i - 1];
            end
            blockStart = i;
        end 
        lastPos = i;
    end
    
    % The previous for loop might miss the last block.
    if numBlocks > 0 && temp(numBlocks, 3) < size(discretePost, 1),
        numBlocks = numBlocks + 1;
        temp(numBlocks, :) = [discretePost(lastPos, 1), blockStart, size(discretePost, 1)];
    end
    %temp(1:numBlocks, :)
    % Now filter out small IBD blocks. 
    for i = 1:numBlocks,
        % If the block is IBD and the size is less than minSize.
        mappedStartSnp = prunedMap(temp(i, 2), 2);
        mappedEndSnp = prunedMap(temp(i, 3), 2);
        if temp(i, 1) == 1 && map(mappedEndSnp, sizeCol) - map(mappedStartSnp, sizeCol) < minSize,
            % Merge this block with the preceding and following ? blocks.
            blockStartIdx = i;
            while blockStartIdx > 1 && temp(blockStartIdx - 1, 1) < 0, 
                blockStartIdx = blockStartIdx - 1;
            end
            blockStartIdx = max(1, blockStartIdx);
            blockEndIdx = i;
            while blockEndIdx < numBlocks && temp(blockEndIdx + 1, 1) < 0,
                blockEndIdx = blockEndIdx + 1;
            end
            blockEndIdx = min(numBlocks, blockEndIdx);
%             if i < numBlocks,
%                 blockEndIdx = i + 1;
%                 if i < numBlocks - 1 && temp(i + 1, 1) < 0,
%                     blockEndIdx = i + 2;
%                 end
%             end
            blockStart = temp(blockStartIdx, 2);
            blockEnd = temp(blockEndIdx, 3);
            % Replace the block at the position blockStartIdx with the
            % merged block, and invalidate the other blocks of the merged
            % block.
            temp(blockStartIdx, :) = [-1, blockStart, blockEnd];
            temp((blockStartIdx + 1:blockEndIdx), 1) = -2;
        end
    end
    %temp(1:numBlocks, :)
    % The previous step might have created ? blocks between two blocks of
    % the same kind.
    
%     last0 = 0;
%     last1 = 0;
%     for i = 1:(size(discetePost, 1) - 1),
%         if discretePost(i + 1) > -1 && discretePost(i + 1, 1) ~= discretePost(i, 1),
%             numTrans = numTrans + 1;
%             blockEnd = min(map(i) + win - 1, totalSnps);
%             if discretePost(i + 1, 1) == 0 && last1 > 0 && last0  < last1,
%                 blockStart = max(map(last1) - win + 1, 0);
%                 temp(numTrans, :) = [blockStart, blockEnd];
%             elseif discretePost(i + 1, 1) == 1 && last0 > 0 && last1  < last0,
%                 blockStart = max(map(last0) - win + 1, 0);
%                 temp(numTrans, :) = [blockStart, blockEnd];
%             end
%         end
%         if discretePost(i, 1) == 0,
%             last0 = i;
%         elseif discretePost(i, i) == 1,
%             last1 = i;
%         end
%     end
end
blocks = temp(temp(:, 1) > -2, :);
end