% Converts a set of transitions into a set of IBD and non-IBD blocks, 
% filtering out IBD blocks (and the corresponding transitions)
% that are smaller than a threshold. The transitions are assumed to be for
% a SINGLE CHROMOSOME.
%
% Arguments:
% trans: matrix of transitions, one line per transition. May have several
% columns, as long as, col 1 is IBD status before the transition, col 2 is
% IBD status after the transition, and the columns start/end[Size]Col are
% as described below.
% startCol: col of trans with the starting position of the block out of
% which we are transitioning. This is assumed to be a SNP index (0-based)
% endCol: col of trans with the ending position of the block out of which
% we are transitioning. 
% startSizeCol, endSizeCol: columns of trans that will be used to get the
% size of the block out of which we are transitioning.
% cutoff: IBD blocks with size smaller than that will be filtered out. 
% chrSize: position of the last SNP of the chromosome. Assumed to be in the
% same units as startSizeCol, endSizeCol
% numSNPs: number of SNPs in the chromosome. This is assumed to be 1 plus
% the largest value that startCol and endCol can take.
%
% Return values:
% selectors(i) is 1 if the i-th transition was not filtered out
% blocks is an m x 4 matrix, one row per block
% blocks(i, 1) is 0 or 1, determining whether this is an IBD or a non-IBD
% block. blocks(i, 2:3) are the start and end positions of the block.
% blocks(i, 4) is the index of the transition (in trans) out of the block.
%
% Transitions 1->1 are ignored if the blocks are consecutive, so the
% indicators 0000????111|111???000 where there is a change in correlated
% allele at the position of the bar, will be grouped into 5 blocks 0?1?0 
% (assuming the two 1 blocks together pass the cutoff).
% Similarly transitions between 2 and 1 IBD are ignored.
function [selectors, blocks] = trans2Blocks(trans, startCol, endCol, ...
    startSizeCol, endSizeCol, cutoff, chrSize, numSnps)
selectors = zeros(size(trans, 1), 1);
temp = ones(40, 4) * (-2);
numBlocks = 0;

% If there are no transitions, return no blocks.
if isempty(trans),
    blocks = temp((temp(:, 1) > -2), :);
    return
end

% Index of the last transition considered. This is used to find the last block which
% will be extended to the end of the chromosome.
lastTrans = 0;
for j = 1:size(trans, 1),
    % Transition IBD -> non-IBD
    if trans(j, 1) > 0 && trans(j, 2) == 0,
        % Traverse the transitions backwards to see where the IBD block
        % started.
        startTransIdx = j;
        while startTransIdx > 1 && ...
                trans(startTransIdx - 1, 1) > 0 && trans(startTransIdx - 1, 2) > 0% && ...
                %trans(startTransIdx - 1, endCol) == trans(startTransIdx, startCol) - 1,
            startTransIdx = startTransIdx - 1;
        end
        startTransIdx = max(startTransIdx, 1);
        % Size computed based on the "sizeCols", but block start and end
        % points are based on startCol and endCol.
        blockSize = trans(j, endSizeCol) - trans(startTransIdx, startSizeCol);
        if blockSize > cutoff,
            selectors(j) = 1;
            numBlocks = numBlocks + 1;
            temp(numBlocks, :) = [1, trans(startTransIdx, startCol), trans(j, endCol), j];
            %temp(numBlocks, :) = [1, trans(startTransIdx, startSizeCol), trans(j, endSizeCol), j];
            lastTrans = j;
        end
    elseif trans(j, 1) == 0 && trans(j, 2) > 0,
        % Transition from non-IBD to IBD. Skip all subsequent IBD blocks
        % until you reach a non-IBD block.
        endTransIdx = j + 1;
        while endTransIdx <= size(trans, 1) && ...
                trans(endTransIdx, 2) > 0,% && ...
                %trans(endTransIdx + 1, startCol) == trans(endTransIdx, endCol) + 1,
            endTransIdx = endTransIdx + 1;
        end
        blockEnd = chrSize;
        if endTransIdx <= size(trans, 1),
            blockEnd = trans(endTransIdx, endSizeCol);
        end
        %endTransIdx = min(endTransIdx, size(trans, 1));
        % trans(j, endSizeCol) is where the IBD block into which we are
        % transitioning starts
        blockSize = blockEnd - trans(j, endSizeCol);
        if blockSize > cutoff,
            selectors(j) = 1;
            numBlocks = numBlocks + 1;
            % When transitioning from a non-IBD block into an IBD block, I
            % merge unknown positions and "false-positive" IBD blocks
            % preceding the non-IBD block into that non-IBD block, so I
            % take all positions after lastTrans. lastTrans has to be a
            % transition into a non-IBD block, because if there was a
            % transition into an IBD block that I trusted, then there
            % should have been a transition out of that block that I also
            % trusted before the next transition into an IBD block.
            % 0-based positions
            %startPos = chrStart;
            startPos = 0;
            if lastTrans > 0,
                startPos = trans(lastTrans, endCol);
                %startPos = trans(lastTrans, endSizeCol);
            end
            temp(numBlocks, :) = [0, startPos, trans(j, endCol), j];
            %temp(numBlocks, :) = [0, startPos, trans(j, endSizeCol), j];
            lastTrans = j;
        end
    end
end

% Conventionally the last block ends at the end of the chromosome. I do not
% extend the first block to the start though.
if lastTrans > 0,
    numBlocks = numBlocks + 1;
    temp(numBlocks, :) = [trans(lastTrans, 2) > 0, trans(lastTrans, endCol), numSnps - 1, -1];
    %temp(numBlocks, :) = [trans(lastTrans, 2) > 0, trans(lastTrans, endSizeCol), chrSize, -1];
end
blocks = temp((temp(:, 1) > -2), :);

% % Index of the last transition considered. This is used to find the last block which
% % will be extended to the end of the chromosome.
% lastTrans = 0;
% for j = 1:size(trans, 1),
%     % Transition IBD -> non-IBD
%     if trans(j, 1) > 0 && trans(j, 2) == 0,
%         % Traverse the transitions backwards to see where the IBD block
%         % started.
%         startTransIdx = j;
%         while startTransIdx > 1 && ...
%                 trans(startTransIdx - 1, 1) > 0 && trans(startTransIdx - 1, 2) > 0% && ...
%                 %trans(startTransIdx - 1, endCol) == trans(startTransIdx, startCol) - 1,
%             startTransIdx = startTransIdx - 1;
%         end
%         startTransIdx = max(startTransIdx, 1);
%         % Size computed based on the "sizeCols", but block start and end
%         % points are based on startCol and endCol.
%         blockSize = trans(j, endSizeCol) - trans(startTransIdx, startSizeCol);
%         if blockSize > cutoff,
%             selectors(j) = 1;
%             numBlocks = numBlocks + 1;
%             temp(numBlocks, :) = [1, trans(startTransIdx, startCol), trans(j, endCol), j];
%             % If the previous IBD block (the one I just added) and the
%             % non-IBD block are not adjacent, add an "unknown" block
%             % between them.
%             if trans(j, endCol) + 1 < trans(j, transPosCol),
%                 numBlocks = numBlocks + 1;
%                 temp(numBlocks, :) = [-1, trans(j, endCol) + 1, trans(j, transPosCol) - 1, -1];
%             end
%             lastTrans = j;
%         end
%     elseif trans(j, 1) == 0 && trans(j, 2) > 0,
%         % Transition from non-IBD to IBD. Skip all subsequent IBD blocks
%         % until you reach a non-IBD block.
%         endTransIdx = j + 1;
%         while endTransIdx <= size(trans, 1) && ...
%                 trans(endTransIdx, 2) > 0,% && ...
%                 %trans(endTransIdx + 1, startCol) == trans(endTransIdx, endCol) + 1,
%             endTransIdx = endTransIdx + 1;
%         end
%         blockEnd = chrSize - 1;
%         if endTransIdx <= size(trans, 1),
%             blockEnd = trans(endTransIdx - 1, endCol);
%         end
%         %endTransIdx = min(endTransIdx, size(trans, 1));
%         blockSize = blockEnd - trans(j, transSizeCol);
%         if blockSize > cutoff,
%             selectors(j) = 1;
%             numBlocks = numBlocks + 1;
%             % When transitioning from a non-IBD block into an IBD block, I
%             % merge unknown positions and "false-positive" IBD blocks
%             % preceding the non-IBD block into that non-IBD block, so I
%             % take all positions after lastTrans. lastTrans has to be a
%             % transition into a non-IBD block, because if there was a
%             % transition into an IBD block that I trusted, then there
%             % should have been a transition out of that block that I also
%             % trusted before the next transition into an IBD block.
%             % 0-based positions
%             startPos = 0;
%             if lastTrans > 0,
%                 startPos = trans(lastTrans, transPosCol);
%             end
%             temp(numBlocks, :) = [0, startPos, trans(j, endCol), j];
%             if trans(j, endCol) + 1 < trans(j, transPosCol),
%                 numBlocks = numBlocks + 1;
%                 temp(numBlocks, :) = [-1, trans(j, endCol) + 1, trans(j, transPosCol) - 1, -1];
%             end
%             lastTrans = j;
%         end
%     end
% end
% 
% % Conventionally the last block ends at the end of the chromosome. I do not
% % extend the first block to the start though.
% if lastTrans > 0,
%     numBlocks = numBlocks + 1;
%     temp(numBlocks, :) = [trans(lastTrans, 2) > 0, trans(lastTrans, transPosCol), numSnps - 1, -1];
% end
% blocks = temp((temp(:, 1) > -2), :);
end