% Finds the overlap between a set of transitions and a set of blocks.
% trans(i) is the position right before the i-th transition. trans is
% assumed to be in increasing order. 
% blocks is an n x 2 matrix where the i-th row has the start and end of
% the i-th block. Blocks might be overlapping and are assumed to be in
% increasing start position.
% Returns:
% transOverlaps(i, 1) is the number of blocks that overlap the i-th
% transition. transOverlaps(i, 2) is the average distance of the center of
% the blocks that overlap the i-th transition from the i-th transition.
% blockOverlaps(i) is the number of transitions that overlap the i-th
% block.
function [transOverlaps, blockOverlaps] = transOverlapCountsAndDist(trans, blocks)
transOverlaps = zeros(size(trans, 1), 2);
blockOverlaps = zeros(size(blocks, 1), 1);
if isempty(blocks) || isempty(trans)
    return;
end
blockIdx1 = 1;
for transIdx = 1:size(trans, 1),
    % Forward the block index as you read blocks before the current
    % transition.
    while blockIdx1 <= size(blocks, 1) && blocks(blockIdx1, 2) < trans(transIdx),
        blockIdx1 = blockIdx1 + 1;
    end
    % At this point, we reached either the last block, or a block that
    % overlaps the current transition, or a block after the current
    % transition. The seach of blocks for the next transition will start
    % from here.
    blockIdx2 = blockIdx1;
    % Update the counts as long as the block overlaps the transition point
    % AND the position right after the transition.
    while blockIdx1 <= size(blocks, 1) && blocks(blockIdx1, 1) <= trans(transIdx) && blocks(blockIdx1, 2) > trans(transIdx),
        blockCenter = blocks(blockIdx1, 1) + ceil((blocks(blockIdx1, 2) - blocks(blockIdx1, 1) + 1) / 2) - 1;
        transOverlaps(transIdx, :) = [transOverlaps(transIdx, 1) + 1, blockCenter - trans(transIdx)];
        blockOverlaps(blockIdx1, 1) = blockOverlaps(blockIdx1, 1) + 1;
        blockIdx1 = blockIdx1 + 1;
    end
    blockIdx1 = blockIdx2;
end
overlaps = find(transOverlaps(:, 1));
% Find the mean distance.
transOverlaps(overlaps, 2) = transOverlaps(overlaps, 2) ./ transOverlaps(overlaps, 1);
end